#! /usr/bin/env perl
#
# word_xtract.pl
#
# Extract high mutual-info word singles, pairs, triples and quads
# from DCAPS VA data.
#
# Disclaimer: This is a hack, and not a robust system. The
# Durkehim API supports a full featured version.  It is meant to
# provide a quick and simple access to a specific file, to
# provide some basic statistics about word counts, and to extract
# ad-hoc data for machine learning experiments.
#
# Usage: cat records.xml | ./word_xtract.pl > some_file.csv
#
# This script reads from standard input and writes to standard
# output. The format of standard input is presumed to be of the
# following form:
#
# <?xml version="1.0" encoding="ISO-8859-1" ?>
# <records>
#   <record>
#     <group>2</group>
#     <study_id>23</study_id>
#     <index_day>348</index_day>
#     <visit_comment>GENERAL HOSPITAL OUTPATIENT NOTE</visit_comment>
#     <notes>Clinical Reminders Activity
#   Influenza Vaccine:
#     Patient received . ml influenza vaccine per order.
#       Primary Diagnosis
#       Site of injection:  Left deltoid intramuscularly
#       Lot number/manufacturer:  Fluviron: Novartis Lot#  P  Exp /
#       Patient received Vaccine Information Statement (VIS) about the
#       influenza vaccine, dated // Centers for Disease Control and
#       Prevention.
#         Level of Understanding: Good
#       Temperature:
#         Temp: . F (. C)</notes>
#   </record>
#   <record>
#     <group>3</group>
#     <study_id>42</study_id>
#     <index_day>127</index_day>
#     <visit_comment>ADMINISTRATIVE NOTE</visit_comment>
#     <notes>Contacted patient regarding recent hospitalization.
# Date of discharge: //
# Date of contact:  //
#
# How are you doing to ? "I am feeling great"
#
# What questions do you have regarding your condition?  None
#
# What questions do you have regarding your treatment? None
#
# What questions do you have regarding your medication(s)? None
#
# What questions or problems do you have with equipment? N/A
#
# What questions or problems do you have with supplies? N/A
#
# Confirm the patient knows when and whom to contact if symptoms deteriorate. Yes,
# patient to contact  .  Contact information given.
#
# Reviewed plan of care:
# __X___ Future appointments at
#         // :  RHEUMATOLOGY WKAM
#         // :  NHB V
# _____ Tests
# _____ Consults
#
# Patient had complaints regarding hospitalization
# __X___ No
# _____ Yes:  referred to</notes>
#   </record>
# </records>
#
# Given the above input, this script will create either a bag-of-words
# dataset, or one of several kinds of bag-of-phrases datasets, depending
# on how it has been configured.  All configuration of this script is
# purely manual, done by editing this script, by hand, before using it.
# There is no provision for automation: this is meant to be a POC
# script only.  To truly understand what it does, you should read
# the script source.
#
# Several notable flags/paramters to set/unset or modify:
# circa line 262: $do_triples  set to 1 to print out 3-grams
# circa line 290: $do_quads    set to 1 to print out 4-grams
# circa line 366-367: modify as appropriate, per above, to print debug stats.
#
# The output data can be cut in various ways by applying thresholds:
# circa lines 484-491:
#	$min_word_cnt = word must appear this many times to be output.
#	$min_pair_cnt = n-gram must appear this many times to be output.
#	$min_pair_mi = mutual information must be this big to be output.
#
# circa lines 430-460:
# Uncomment one of these lists of allowed words. An n-gram will be output
# only if it contains one of these 'allowed words'.  The list of allowed
# words was obtained from the raw dataset, using any one of several
# different techniques. In all cases, the allowed words correlate strongly
# with the patient classification.
#
# circa line 499-500: Currently, this script will only print out records
# for groups 2 vs. 3.  Uncomment/modify as appropriate to print other groups.
#
# This script outputs data in a generalized CSV format.  Fixed columns
# have column labels.  Variable columns are in the form of key:value pairs.
# This format is used to handle sparse data efficiently.  The poses subsystem
# is able to natively read and write this format.
#
# Linas Vepstas December 2012-Feburary 2013

use strict;

# Database record fields
my $group = -1;
my $study_id = -1;
my $index_day = -1;

# start of text flag
my $have_text = 0;

# records to be printed out.
my $records_words = {};
my $records_pairs = {};

# Global word and word-pair counts
my %glo_word_cnt = ();
my %glo_pair_cnt = ();
my $glo_cnt = 0;
my $bogus_word = "XYZZY";
my $prev_word = $bogus_word;
my $prepre_word = $bogus_word;
my $preprepre_word = $bogus_word;
my $prepreprepre_word = $bogus_word;

sub get_recname
{
	my $grp = $_[0];
	my $sid = $_[1];
	my $day = $_[2];

	my $str = "";
	$str .= $grp;
	$str .= "#";
	$str .= $sid;
	return $str;
}


while (<>) {
	if (/<?xml version/) { next; }
	if (/<records>/) { next; }
	if (/<\/records>/) { next; }

	# Clear the counters.
	if (/<record>/ or /<\/record>/) {
		$group = -1;
		$study_id = -1;
		$index_day = -1;
		$have_text = 0;
		next;
	}

	if (/<group>([1-3])<\/group>/) {
		$group = $1;
		next;
	}
	if (/<study_id>([0-9]+)<\/study_id>/) {
		$study_id = $1;
		next;
	}
	if (/<index_day>([0-9]+)<\/index_day>/) {
		$index_day = $1;
		next;
	}

	if (/<visit_comment>/) {
		$have_text = 1;
		s/<visit_comment>//;
		s/<\/visit_comment>//;
	}

	if (/<notes>/) {
		$have_text = 1;
		s/<notes>//;
		s/<\/notes>//;
	}

	if ($have_text) {
		# Remove leading trailing whitespace
		chomp;
		s/\r//;
		s/^\s+//;
		s/\s+$//;
		if (/<\/notes>/) {
			s/<\/notes>//;
			$have_text = 0;
		}
		if (/<\/record>/) {
			$have_text = 0;
			next;
		}

		# quotes and ampersands turn into one word.
		# Note: attivio is using these as word-separators.
		s/\&quot;//g;
		s/\&amp;//g;

		# word-separators.
		s/\&lt;/ /g;
		s/\&gt;/ /g;

		# skip blank lines
		if ("" eq $_) {
			next;
		}

		my $recidx = &get_recname($group, $study_id, $index_day);
		my $wcnt = $records_words->{$recidx};
		my $pcnt = $records_pairs->{$recidx};

		# strip punctuation
		# Hyphenated, underscores words count as one word,
		# all other punct are word separators
		s/_//g;
		s/-//g;
		s/\W/ /g;
		my @words = split;
		foreach my $word (@words) {
			# Uppercase
			$word = uc($word);
			if ("" eq $word) { next; }

			# Attivio does not count words if they are single letters.
			if (1 == length($word)) { next; }

			# print "$recidx $word \n";

			# count how often this word occurs in this record.
			$wcnt->{$word} += 1;
			# while (my($w, $c) = each (%$wcnt)) { print "$w : $c , "; } print "\n";

			# Increment global word counts.
			$glo_cnt ++;
			$glo_word_cnt{$word} += 1;

			# Word pairs
			my $pair = $prev_word . "_" . $word;   # 10
			$glo_pair_cnt{$pair} += 1;
			$pcnt->{$pair} += 1;

			$pair = $prepre_word . "_" . $word;   # 20
			$glo_pair_cnt{$pair} += 1;
			$pcnt->{$pair} += 1;

			$pair = $preprepre_word . "_" . $word; # 30
			$glo_pair_cnt{$pair} += 1;
			$pcnt->{$pair} += 1;

			# Triples with blanks in them.
			my $do_triples = 0;
			if (0 < $do_triples) {
				$pair = $prepre_word . "_" . $prev_word . "_" . $word;  #210
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $preprepre_word . "_" . $prev_word . "_" . $word; #310
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $preprepre_word . "_" . $prepre_word . "_" . $word; #320
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $prepreprepre_word . "_" . $prev_word . "_" . $word;  #410
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $prepreprepre_word . "_" . $prepre_word . "_" . $word;  #420
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $prepreprepre_word . "_" . $preprepre_word . "_" . $word; #430
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;
			}

			# Quads with blanks in them.
			my $do_quads = 0;
			if (0 < $do_quads) {
				$pair = $preprepre_word . "_" . $prepre_word . "_" . $prev_word . "_" . $word;  # 3210
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $prepreprepre_word . "_" . $prepre_word . "_" . $prev_word . "_" . $word;  # 4210
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $prepreprepre_word . "_" . $preprepre_word . "_" . $prev_word . "_" . $word;  # 4310
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;

				$pair = $prepreprepre_word . "_" . $preprepre_word . "_" . $prepre_word . "_" . $word;  # 4320
				$glo_pair_cnt{$pair} += 1;
				$pcnt->{$pair} += 1;
			}


			# Bump them down
			$prepreprepre_word = $preprepre_word;
			$preprepre_word = $prepre_word;
			$prepre_word = $prev_word;
			$prev_word = $word;
		}
		$records_words->{$recidx} = $wcnt;
		$records_pairs->{$recidx} = $pcnt;
		next;
	}

	# Do something.
	if (/<\/record>/) {
		$have_text = 0;
		if ((-1 == $group) or (-1 == $index_day) or (-1 == $study_id)) {
			print "Error: invalid record!\n";
			exit 1;
		}
		next;
	}

	print "Error: unhandled: $_";
}

print "!  Done reading\n";

# Compute probabilities
#Hack for inital left word
$glo_word_cnt{$bogus_word} = 2;
$glo_cnt += 2;

# ------------------------------------------------------------
# Report on single-word probabilities
my %prob_word = ();
my $tot_prob_word = 0;

foreach my $word (keys %glo_word_cnt) {
	$prob_word{$word} = $glo_word_cnt{$word} / $glo_cnt;
	$tot_prob_word += $prob_word{$word};
}

print "! Total word count = $glo_cnt\n";
my @uniq_wds = keys %glo_word_cnt;
print "! Total unique words = $#uniq_wds\n";

my $err = 1.0 - $tot_prob_word;
print "! Word probability missing mass = 1 - sum_w p_w = $err\n";

# --------------------------------------------------
# Report on word pairs

my %prob_pair = ();
my $tot_prob_pair = 0;

# -2 to undo the damage of the bogus word XYZZY
# 3+6+4 because we create 3 different pairs, 6 different triples, and 4 quads
# my $pair_cnt = ($glo_cnt - 2) * (3+6+4);
my $pair_cnt = ($glo_cnt - 2) * (3);

foreach my $pair (keys %glo_pair_cnt) {
	$prob_pair{$pair} = $glo_pair_cnt{$pair} / $pair_cnt;
	$tot_prob_pair += $prob_pair{$pair};
}

print "! Total multi-word phrase count = $pair_cnt\n";
my @uniq_prs = keys %glo_pair_cnt;
print "! Total unique multi-word phrases = $#uniq_prs\n";

my $err = 1.0 - $tot_prob_pair;
print "! Phrase probability missing mass = 1 - sum_wv p_wv = $err\n";

# -------------------------------------------------
# Compute mutual information
my %pair_mi = ();

# XXX  this is incorrect for word triples, quads
foreach my $pair (keys %prob_pair) {
	my ($wd_left, $wd_right) = split (/_/, $pair);
	$pair_mi{$pair} = $prob_pair{$pair} / ($prob_word{$wd_left} * $prob_word{$wd_right});
	$pair_mi{$pair} = log($pair_mi{$pair}) / 0.693147181;
}

my @mi_dist = [];
foreach my $pair (reverse sort {$pair_mi{$a} <=> $pair_mi{$b}} keys %pair_mi) {
	my $pmi = $pair_mi{$pair};
	for (my $m= 1; $m < 10; $m++ ) {
		my $mi = 24 - 4.0* $m;
		if ($mi < $pmi) { $mi_dist[$m] += 1; }
	}
}

for (my $m= 1; $m < 10; $m++ ) {
	my $mi = 24 - 4.0* $m;
	print "! $mi_dist[$m] \t pairs with mi > $mi\n";
}

# foreach my $pair (reverse sort {$pair_mi{$a} <=> $pair_mi{$b}} keys %pair_mi) {
# 	print "$pair_mi{$pair} $pair\n";
# }

my %log_word = ();

foreach my $word (keys %prob_word) {
	$log_word{$word} = log($prob_word{$word}) / 0.693147181;
}

# foreach my $word (reverse sort {$log_word{$a} <=> $log_word{$b}} keys %log_word) {
# 	print "$log_word{$word} $word\n";
# }

my %log_pair = ();

foreach my $pair (keys %prob_pair) {
	$log_pair{$pair} = log($prob_pair{$pair}) / 0.693147181;
}

# foreach my $pair (reverse sort {$log_pair{$a} <=> $log_pair{$b}} keys %log_pair) {
# 	print "$log_pair{$pair} $pair \tMI=$pair_mi{$pair}\n";
# }

# The allowed-words list below consists of 370 words exactly, extracted
# from a prior training run on the so-called "January-2013" dataset.  The
# training run was performed on the va32 dataset, created 40 models,
# and then took the list of all of the words that appeared in any of the
# models.
my $allowed_words = "ZER YEAR THI SUPPLY TEASPOONFUL UNDERGONE STAIN YE YOUNGER VERIFIED THY SUSAN TRAZODONE VOLUME WORTHLESSNESS SSN SYMPTOMS TURMOIL EPIDERMAL ULTIMATELY STRAIGHTENED WATCHES VALLEY SUPERVISORS SUBLUXATION STRAIGHTEN TIB SPELL GEROPSYCHIATRY UNT UC SU UNSP TRAVELS WHIP NALCOHOL TOPICAL STX CONSISTANTLY SUPERVISOR STANDARDS LIPITOR WTIH PSYCHIATRICALLY WOUNDED SUBSALICYLATE VISUALIZATION TP TWAVE CAHNGE STRANGE TAMINE UNHAPPY TRIPLET UCINATIONS INTERMITTANT UES UNCHANGED TOP VIT STVHCS UCINATING VERB FRIGHTENING VOIDING WELLGROOMED WARRANT USEFULNESS CRYING APPOINT NOBODY SPARE REGAINED SWABS VTACH STREN IRRITATIVE URINATED THERAPEUTIC SUBCUTANEOUS TUSCON UNCOOPERATIVE THROUGHOUT TEACHER WHEN TOUGH THINNER UG WALKIN TEMPOPORMAND BLOCKERS DEMONSTRATING UNMARRIED KIMBERLY TOPICALLY TONOMETRY STENOSES SWINGS SULFATE STRAIN STYLES VET TISSUE WHEREABOUTS TON WAVE UNAWARE EUTHYMIC LIBIDO WILLING STRESS IENT VISIT ARMD THICKNESS UNRESECTABLE STUDIES SPOKEN SPLITTING DISHEVELED POINTING JEF DISORDER MANKOWSKI MONFRI NURS PREOCCUPIED EASIER GUNNERS MATE SOLD YR DRUSEN PERMITTED BREASTFEEDING PUNCTUM LUTS YOUR CUTTER ANESTHETIZING MAKER SODIUM RESIDENTIAL STAFFED VOICE STEATOSIS WOMAN TFTS UOP SWEATER STRETCH TELEMETRY TEMAZEPAM MEALTIME WEBB XPATIENT RET AROUSE TREATED THIGHS WORDS UNITL YEAROLD CURRRENT UPDATES DPT STRUGGLES INTRAARTICULAR UPCOMING ERVAMC VOICEMAIL WITH RE TRANS STOLE USMC DENT THANKED STERNUM WORKS LEASING OUTSTANDING VISTA SUBACUTE TUR HIGHER STATED CALM TEETH POOL VIDEOS RANGES WU STREET CHEN TETANUS WED SPREADING WILLL VINFLUENZA NOTABLY WARNING VESSEL COMPREHENDING VOLUNTEER TMD WORST WERE YOURE DIRECTORY TREECE STUDENT SQUIRMING SPECIMEN SPLINT VIRTUE TYPICALLY TATTOO TAMPA STYLE VOIDED TACH SWEETS YEARS UNPLEASANT SVN UNIQUE SPECIFIES NOCTURIA CANCERS ICDCM HC SUPERVISING STOP TWICE EMPHASIZED TUMORS TRANSFUSIONS WEAN VISITORS TABS VISA UPSETTING WELL TH TE REASONING THERE PREFILLED ANXIOLYTIC MPLS EYEBROWS SUPPLE INFLAMMATORY DAV NESTERASE SPIRITS APPENDICEAL QUADRANTS INTERESTING DESPONDENT LUQ NEUTRAL SOLAR RESTRICTED ACIPHEX PRIVATE SMELL BRONZE SIDE FEAR SOFTENER NIECE SOCIA MGOH HEMEONC ANAND GENTLE OWEEN SODA PRACTICING SOLELY ALOH METASTATIC RESTORED AEROBICS CEREBROVASCULAR BARS CONTRIBUTED SOL ROUSED GRASP ATHEROSCLEROSIS SOMEWHERE ANESTHESIA FLUVIRIN CERTIFY CREAT ANTIPSYCHOTIC PRESBYOPIA VACCINATED SOCHX AAA ONEONONE VENTILATE SCHWARTZ VS BRIEFS SOMETHING OBESITY ARTHRITIC HIPS APPLICABLE REORDER EFFECTED CONVENIENT VANESA ALOT ATTEND PUFF ACCOMPANIED SHAVE SPEAKS AGGRAVATING OFFICIAL PORTER ABNORMAL APNEA AFRAID ARTHRALGIA FORMULATION AREAS DOCUSATE NEUT RUMINATES CLICK TACTILE BETA COLLATERAL SOMEONE SPORADICALLY RETROFLEXION ADULT SALTS TENDED SOUTH ALLOWED SONS SOLUMEDROL ALIGNMENT SOUP ANT ALONG SPECIALIST SORE DIGOXIN NORMALIZING SOA ANTICOAGULANTS SSED STITCH ABOUT TEACHINGS ACHING STUDIED ADR ASAP AA MEDICINES SOAK ANTIBIOTICS";

# The allowed_words immediately below are those obtained from running
# feature selection on the va32 dataset, setting -C370 to get 370 words,
# (same as above) and using the -asimple algo. BTW, took 4 secs to run.
$allowed_words = "ABLE ABOUT ABUSE ACTIVE ACTIVITIES ACTIVITY ADDENDUM ADDITIONAL ADMISSION ADMITTED AFFECT AFTER AGE AGREE ALCOHOL ALERT ALL ALLERGIES ALSO AM AN AND ANXIETY ANY APPOINTMENT APPOINTMENTS APPT ARE AROUND AS ASKED ASSESSMENT ASSISTANCE AT ATTENDED ATTENDING AXIS BACK BE BED BEEN BEFORE BEHAVIOR BEING BETTER BLOOD BP BREATH BUT BY CALL CALLED CALLER CAME CAN CARE CASE CC CE CHEST CHIEF CLASS CLASSES CLINIC CM COMMENTS COMPLAINT COMPLAINTS COMPLETED CONSENT CONSTANT CONSULT CONTACT CONTINUE CONTINUED CONTINUES CONTROL CURRENT DAILY DATE DEC DENIES DEPRESSED DEPRESSION DIAGNOSIS DID DISCHARGE DISCUSSED DISEASE DISORDER DO DOCUMENT DOES DOING DOSE DRUG DUE DURING DX ED EDUCATION ER EVALUATION EVERY EXAM FACILITY FACTORS FAMILY FEEL FEELING FINDINGS FOLLOWING FOLLOWUP FOR FROM FULL GIVEN GO GOAL GOOD GROUP HAD HAS HAVE HCL HE HEALTH HELP HER HERE HIGH HIM HIS HISTORY HOME HOURS HOW HX IDEATION IF IMAGING IN INCLUDING INDIVIDUAL INFORMATION INFORMED INPATIENT IS ISSUES IT IV LABS LAST LB LEARNING LEFT LEVEL LIST LIVING LOCATION LOW LOWER MALE MANAGEMENT MAR MASS MAY MD MEDICAL MEDICATION MEDICATIONS MEDS MENTAL MET MG MH MHC MINUTES MONITOR MONITORING MONTH MOOD MORE MOUTH MR MUSCLE NA NAME NECK NEED NEEDED NEEDS NEW NEXT NO NONE NORMAL NOT NOTE NOTED NOW NUMBER NURSE NURSING NUTRITION OBTAINED OF OFFERED OLD ON ONE OR ORAL ORDERED ORIENTED OTHER OUT OUTPATIENT PAIN PAST PATIENT PATIENTS PCP PER PERFORMED PHONE PHYSICAL PHYSICIAN PING PLACE PLAN PLEASE PM PO POST PRESENT PRESSURE PRIMARY PRIOR PRN PROBLEM PROBLEMS PROCEDURE PROGRAM PROGRESS PROVIDED PROVIDER PSYCH PSYCHIATRY PT PTSD PULSE QUESTIONS REASON RECENT RECENTLY REFERRAL RELATED REPORT REPORTED REPORTS REQUESTED REQUESTING REQUIRED RESCHEDULE RESP RESPONSE RESULTS REVIEW REVIEWED RIGHT RISK ROOM SAFETY SAID SC SCALE SCHEDULED SEE SEEN SELF SENT SERVICE SERVICES SESSION SEVERAL SHE SHIFT SHOULD SI SIDE SINCE SITE SKILLS SKIN SO SOCIAL SOME SSN STABLE STAFF STARTING STATED STATES STATUS STILL SUICIDAL SUICIDE SUPPLIES SUPPLY SUPPORT SURGERY SYMPTOMS TAB TABLET TAKE TEAM TELEPHONE TEMPERATURE TEST THAN THAT THE THEM THEN THERAPEUTIC THERAPY THERE THEY THIS THOUGHTS TIME TO TOBACCO TOTAL TREATMENT TWO UNIT UP URINE USE USED VA VERIFIED VET VETERAN VETERANS VISIT VS WAS WE WEEK WEIGHT WELL WERE WHAT WHEN WHICH WHO WIFE WILL WITH WITHIN WORK WOULD YEAR YEARS YES YOU YOUR";

# The allowed_words immediately below are those obtained from running
# feature selection on the va32 dataset, setting -C370 to get 370 words,
# But actually got 367,  due to the -ainc algo used to get this. BTW, the
# -ainc algo took about 5 minutes.
# $allowed_words = "ABLE ABOUT ABUSE ACTIVE ACTIVITIES ACTIVITY ADDENDUM ADDITIONAL ADMISSION ADMITTED AFFECT AFTER AGE AGREE ALCOHOL ALERT ALL ALLERGIES ALSO AM AN AND ANXIETY ANY APPOINTMENT APPOINTMENTS APPT ARE AROUND AS ASKED ASSESSMENT ASSISTANCE AT ATTENDED ATTENDING AXIS BACK BE BED BEEN BEFORE BEHAVIOR BEING BETTER BLOOD BP BREATH BUT BY CALL CALLED CALLER CAME CAN CARE CASE CC CE CHEST CHIEF CLASS CLINIC CM COMMENTS COMPLAINT COMPLAINTS COMPLETED CONSENT CONSTANT CONSULT CONTACT CONTINUE CONTINUED CONTINUES CONTROL CURRENT DAILY DATE DEC DENIES DEPRESSED DEPRESSION DIAGNOSIS DID DISCHARGE DISCUSSED DISEASE DISORDER DO DOCUMENT DOES DOING DOSE DRUG DUE DURING DX ED EDUCATION ER EVALUATION EVERY EXAM FACILITY FACTORS FAMILY FEEL FEELING FINDINGS FOLLOWING FOLLOWUP FOR FROM FULL GIVEN GO GOAL GOOD GROUP HAD HAS HAVE HCL HE HEALTH HELP HER HERE HIGH HIM HIS HISTORY HOME HOURS HOW HX IDEATION IF IMAGING IN INCLUDING INDIVIDUAL INFORMATION INFORMED INPATIENT IS ISSUES IT IV LABS LAST LB LEARNING LEFT LEVEL LIST LIVING LOCATION LOW LOWER MALE MANAGEMENT MAR MASS MAY MD MEDICAL MEDICATION MEDICATIONS MEDS MENTAL MET MG MH MHC MINUTES MONITOR MONITORING MONTH MOOD MORE MOUTH MR MUSCLE NA NAME NECK NEED NEEDED NEEDS NEW NEXT NO NONE NORMAL NOT NOTE NOTED NOW NUMBER NURSE NURSING NUTRITION OBTAINED OF OFFERED OLD ON ONE OR ORAL ORDERED ORIENTED OTHER OUT OUTPATIENT PAIN PAST PATIENT PATIENTS PCP PER PERFORMED PHONE PHYSICAL PHYSICIAN PING PLACE PLAN PLEASE PM PO POST PRESENT PRESSURE PRIMARY PRIOR PRN PROBLEM PROBLEMS PROCEDURE PROGRAM PROGRESS PROVIDED PROVIDER PSYCH PSYCHIATRY PT PTSD PULSE QUESTIONS REASON RECENT RECENTLY REFERRAL RELATED REPORT REPORTED REPORTS REQUESTED REQUESTING REQUIRED RESCHEDULE RESP RESPONSE RESULTS REVIEW REVIEWED RIGHT RISK ROOM SAFETY SAID SC SCALE SCHEDULED SEE SEEN SELF SENT SERVICE SERVICES SESSION SEVERAL SHE SHOULD SI SIDE SINCE SITE SKILLS SKIN SO SOCIAL SOME SSN STABLE STAFF STARTING STATED STATES STATUS STILL SUICIDAL SUICIDE SUPPLIES SUPPLY SUPPORT SURGERY SYMPTOMS TAB TABLET TAKE TEAM TELEPHONE TEMPERATURE TEST THAN THAT THE THEM THEN THERAPEUTIC THERAPY THERE THEY THIS THOUGHTS TIME TO TOBACCO TOTAL TREATMENT TWO UNIT UP URINE USE USED VA VERIFIED VET VETERAN VETERANS VISIT VS WAS WE WEEK WEIGHT WELL WERE WHAT WHEN WHICH WHO WIFE WILL WITH WITHIN WORK WOULD YEAR YEARS YES YOU YOUR";


# Below are 370 top words from the Janaury 2013 va12 dataset
# Obtained via feature selection
# $allowed_words = "ABOUT ABUSE ACTIVE ACTIVITIES ACTIVITY ACUTE ADDENDUM ADMISSION ADMITTED ADVISED AFFECT AFTER AGAIN AGO ALCOHOL ALERT ALLERGIES ALSO AM AN AND ANSWER ANSWERED ANXIETY ANY APPOINTMENT APPT ARE AS ASSESSMENT AT AXIS BACK BE BED BEEN BEING BETTER BLOOD BOTH BP BREATH BUT BY CALL CALLED CAN CANCER CARE CASE CBOC CHANGE CHANGES CHECK CHEST CHIEF CLEAR CLICK CLINIC CLINICAL CM COMMENT COMMENTS COMPLAINT COMPLETE COMPLETED CONCERN CONDITION CONFIRMED CONSENT CONSULT CONTACT CONTINUE COULD CURRENT CURRENTLY DAILY DATE DAUGHTER DENIES DEPENDENCE DEPRESSED DEPRESSION DIAGNOSIS DID DIFFICULTY DISCHARGE DISCHARGED DISCUSSED DISORDER DO DOB DOCUMENT DOES DOSE DRUG DURING DX EACH ED EDUCATION ER EVALUATION EVERY EXAM EXCEPT FACILITY FAIR FALL FOLLOW FOLLOWING FOR FOUND FOUR FROM GI GIVEN GO GOALS GOOD GROUP HAD HAS HAVE HAVING HCL HE HEALTH HEART HERE HI HIM HIS HISTORY HOME HOSPITAL HOURS IDENTIFIED IF IN INCLUDING INDIVIDUAL INFLUENZA INFORMATION INFORMED ING INPATIENT INSTRUCTIONS INTERVENTIONS IS IT IV JUL JUST KG KNOW LABS LAST LB LEARNING LEFT LEG LEVEL LIKE LIST LIVING LOSS MADE MALE MANAGEMENT MAY ME MED MEDICAL MEDICATION MEDICATIONS MEDS MEMBER MENTAL MESSAGE MET MG MH MHC MIND MINUTES MOOD MORNING MOUTH MR NAME NASAL NECK NEED NEEDED NEEDS NEURO NEW NEXT NO NONE NONVA NORMAL NOT NOTE NOTED NOW NUMBER NURSE NURSING NUTRITION OBTAINED OF OFFERED OFFICE OLD ON ONCE ONE OR ORAL ORDERED ORDERS ORIENTED OTHER OUT OUTPATIENT PAIN PATIENT PATIENTS PC PER PHONE PHYSICAL PHYSICIAN PLAN PLEASE PO PRESCRIBED PRESENT PRESENTS PRESSURE PRIMARY PRIOR PROBLEM PROBLEMS PROCEDURE PROGRESS PROVIDED PROVIDER PSYCHIATRIC PSYCHIATRIST PT PTS PTSD PULMONARY PULSE QUITE RATE REASON RECEIVED REFERRAL REFERRED REFILL REGARDING RELATIONSHIP REPORT REPORTED REPORTS REQUEST REQUESTED REQUESTING RESP RESPIRATION RESPONSE REVIEW REVIEWED RIGHT RISK ROOM SAFETY SAYS SCAN SCHEDULE SCHEDULED SCORE SEE SEEN SELF SERVICE SESSION SEVERE SHE SHIFT SI SIGNS SINCE SITE SKIN SMOKING SO SOCIAL SOME SOUNDS SPEAK SPOKE STATED STATES STATUS STRESS SUICIDAL SUICIDE SUMMARY SUPPLIES SURGERY SW SYMPTOMS TAB TABLET TAKE TAKING TEAM TELEPHONE TEMP TEMPERATURE THAT THE THEIR THEN THERE THEY THIS TIME TO TOBACCO TOTAL TREATMENT TRIAGE TWO TYPE UNIT UNTIL UP USE USED VA VACCINE VET VETERAN VETERANS VISIT VITALS WANTS WAS WE WEEK WELL WERE WHEN WHICH WHO WILL WITH WITHIN WITHOUT WORK WOULD YES YOU YOUR";

# Below are top 370 words from December 2012 va12 dataset
# Obtained via feature selection
# $allowed_words = "ABDOMEN ABOUT ABTHE ABUSE ACTIVE ACTIVITIES ACTIVITY ACUTE ADDENDUM ADMISSION ADMITTTHE AFFECT AGE AGO ALL ALSO ALTHE AM AN AND ANSWER ANXIETY ANY APPEARS APPOINTMENT APPT AR ARE ARNING AROUND AS ASE ASSESSMENT AT ATHE ATTEMPT AXIS BACK BARRIERS BE BEEN BEHAVIOR BEING BETTER BLOOD BOTH BP BREATH BROUGHT BTHE BUT BY CALL CALTHE CAN CARE CASE CHIEF CLINIC CLINICAL CM COMPLAINT COMPTHE CONCERN CONDITION CONFIRMTHE CONSENT CONSULT CONTACT COTHE COULD CT CTHE CURRENT CURRENTLY DAILY DATE DENIES DEPENDENCE DEPRESSION DEPRESSIVE DEPRESSTHE DIAGNOSIS DID DIFFICULTY DISCHARGE DISCUSSTHE DISORDER DO DOCUMENT DOES DOSE DURATION DURING DX EACH ENCOUNTER ENTS ER EVALUATION EVERY FAMILY FOLLOW FOLLOWING FOLLOWUP FOR FROM FT GI GIVEN GO GOOD GROUP HAD HAND HAS HAVE HAVING HE HEALTH HEART HER HERE HI HIM HIS HISTORY HOME HOSPITAL HOURS ICAL ICATION ICATIONS IDENTIFITHE IF IN INCLUDING INDIVIDUAL INFLUENZA INFORMATION INFORMTHE ING INHATHE INHL INPATIENT INSTEAD INSTRUCTIONS INTERVENTION INTO IS IT IV KG KNOW LABS LAST LB LIST LIVING LOWER LY MADE MANAGEMENT MATHE MAY ME MENTAL MG MH MHC MIN MINUTES MONTHS MOOD MORE MORNING MOUTH MOVEMENT MR MS MTHE MY NAME NECK NETHE NEW NO NONE NORMAL NOT NOTE NOTTHE NOW NUMBER NURSE NURSING OBTAINTHE OF OFFERTHE OLD ON ONE OR ORAL ORDER ORDERS ORDERTHE ORIENTTHE OTHER OUT OUTPATIENT PAIN PART PAST PATIENT PATIENTS PATTERN PER PERSON PERSONAL PHONE PHONTHE PHYSICAL PHYSICIAN PING PLAN PLANS PO POSSIBTHE PRESCRIBTHE PRESENT PRESSURE PRIMARY PRIOR PRN PROBTHE PROCTHE PROGRESS PROTOCOL PROVIDER PSYCH PSYCHIATRIC PSYCHIATRIST PSYCHIATRY PT PTHE PTS PTSD PULSE RAL RATE RECEIVTHE RECOTHE REFERRTHE REGARDING RELATTHE REPORT REPORTS REPORTTHE REQUEST REQUESTING REQUESTTHE RESP RESPONSE RETURNTHE REVIEW REVIEWTHE RGIES RIGHT RISK RT SAFETY SAME SAYS SCAN SCHTHE SEE SEEN SELF SENT SERVICE SESSION SEVERE SHE SHIFT SI SIGNS SINCE SITE SKIN SOCIAL SOME SPEECH SPOKE STATES STATTHE STATUS STHE STRESS SUICIDAL SUPPLIES SUPPORT SURGERY SURGICAL SW TAB TABTHE TAKE TE TEAM TELEPHONE TEMP TETHE THAT THE THEIR THEN THERAPY THEY THIS TIME TIMES TO TOBACCO TOTAL TOTHE TREATMENT TRIAGE TTHE TWO TYPE UCATION UNDERSTANDING UNIT UNTIL UP UPPER URE USE USTHE UTHE VA VEL VERBALIZTHE VET VETERAN VETERANS VETS VISIT WAS WEEK WEIGHT WELL WERE WHEN WHICH WHO WIFE WILL WITH WITHIN WITHOUT WORK WOULD WT YEAR YES YESTERTHE YOU YOUR";

# Below are top 370 words from December 2012 va32 dataset
# Obtained via feature selection
# $allowed_words = "ABD ABILITY ABOUT ABTHE ABUSE ACTIVE ACTIVITIES ACTIVITY ADDENDUM ADDITIONAL ADMISSION AFFECT AFTER AGE ALCOHOL ALL ALSO ALTHE AM AN AND ANXIETY ANY APPETITE APPLICABTHE APPOINTMENT APPT AR ARE AREA ARNING ARRIVAL AS ASANT ASKTHE ASSESSMENT AT ATHE ATTENDING ATTENDTHE BACK BE BEEN BEFORE BEING BETTER BLOOD BP BTHE BUT BY CALL CALTHE CAN CARE CC CHEST CHIEF CLINIC CM COMPLAINT COMPLAINTS COMPTHE CONDITION CONSENT CONSULT CONTACT CONTINUE CONTINUES COTHE COUGH CTHE CURRENT CURRENTLY DAILY DATE DEC DENIES DEPRESSION DEPRESSTHE DESCRIBE DIAGNOSIS DID DIFFICULTY DISCHARGE DISCUSSTHE DISORDER DO DOCUMENT DOES DOING DRUG DUE DURING DX EFFECTS EMERGENCY ENTS EVALUATION EVERY FACTORS FALL FAMILY FEEL FEELING FEELINGS FOLLOW FOLLOWING FOLLOWTHE FOLLOWUP FOR FROM FT FULL GET GIVEN GO GOAL GOOD GROUP HAD HAS HAVE HAVING HE HEALTH HEART HELP HER HERE HIM HIS HISTORY HOME HOSPITAL HOURS HOW ICAL ICATION ICATIONS IDEATION IDENTIFITHE IDENTIFY IF IMAGING IN INCLUDES INCLUDING INDIVIDUAL INFORMATION INJURY INPATIENT INTO IS ISSUES IT IV JOB KG LAST LB LIST LIVING LOCATION LOW LOWER MADE MANAGEMENT MASS MATHE MAY ME MENTAL MG MH MHC MIN MINUTES ML MOBILITY MONTH MOOD MORE MOUTH MR MS MTHE MUCH NAME NETHE NEXT NO NONE NORMAL NOT NOTE NOTTHE NOW NUMBER NURSE NURSING OBJECTIVE OBTAINTHE OF OFF OLD ON ONE OR ORAL ORDERTHE ORIENTTHE OTHER OTHERS OUT OUTPATIENT OVER PAIN PAST PATIENT PATIENTS PC PCP PER PERSON PHONE PHYSICAL PHYSICIAN PING PLAN PM PO POSSIBTHE PRESENT PRESENTS PRESSURE PRIMARY PRIOR PRN PROBTHE PROCESS PROCTHE PROGRAM PROGRESS PROTOCOL PROVIDE PROVIDER PROVIDTHE PSYCHIATRY PT PTHE PTSD PULSE QUESTIONS RATE RECEIVTHE RELATTHE REPORT REPORTS REPORTTHE REQUESTING REQUESTTHE RESP RESULTS REVIEW REVIEWTHE RGIES RIGHT RISK ROOM RT RTHE SAID SC SCATHE SCHTHE SEE SEEN SERVICE SERVICES SESSION SEX SHE SIDE SIGNIFICANT SINCE SITE SKILLS SO SOCIAL SOME SPEECH SS STAFF STARTING STATES STATTHE STATUS STHE STILL SUICIDAL SUPPLIES SUPPORT SURGERY SYMPTOMS TAB TABTHE TAKE TEACHING TEAM TELEPHONE TEST TETHE THAN THAT THE THEM THEN THERAPY THERE THEY THIS THOUGHTS TIME TIMES TO TOTHE TREATMENT TRIAGE TTHE TWO TYPE UCATION UNABTHE UNDERSTANDING UNSPECIFITHE UP URE USE USTHE UTHE VA VEL VERY VET VETERAN VETERANS VISIT WAS WE WEEK WEIGHT WELL WERE WHAT WHEN WHICH WHO WIFE WILL WITH WITHIN WORK WOULD WRITER YEAR YEARS YES YOU YOUR";


my @allowed_list = split (/ /, $allowed_words);
my %is_allowed = ();
for (@allowed_list) { $is_allowed{$_} = 1; }

# Print only certain word phrases
# Actually, return 1 if the given word belongs to the allowed list, else return 0.
sub phrase_filter
{
	my $phr = $_[0];
	my @words = split(/_/, $phr);
	foreach my $w (@words) {
		# if ($w ~~ @allowed_list) { return 1; }  # requires perl 5.10
		if ($is_allowed{$w}) { return 1; }
	}

	return 0;
}

# Print in sorted order
sub print_records
{
	# These are the setting used in the original pairs discovery.
	my $min_word_cnt = 4;
	my $min_pair_cnt = 4;
	my $min_pair_mi = 2.0;

	$min_word_cnt = 2;
	$min_pair_cnt = 2;
	$min_pair_mi = 2.0;

	my $tot_words = 0;
	my $tot_pairs = 0;

	my %uniq_words = ();
	my %uniq_pairs = ();

	print "group, study_id\n";
	# for (my $grp = 1; $grp <=3; $grp++) {
	for (my $grp = 2; $grp <=3; $grp++) {
		for (my $sid = 1; $sid <=300; $sid++) {
			my $recid = $grp . "#" . $sid;

			# Print the words if the global count exceeds a minimum.
			my $wcnt = $records_words->{$recid};
			my @wds = keys %$wcnt;
			if ($#wds < 1) { next; }
			print "$grp, $sid";
			while (my($w, $c) = each (%$wcnt)) {
				if ($min_word_cnt < $glo_word_cnt{$w}) {
					print " , $w : $c";
					$tot_words += $c;
					$uniq_words{$w} += $c;
				}
			}

			# Print word pairs if the global count exceeds a minimum.
			my $pcnt = $records_pairs->{$recid};
			my @prs = keys %$pcnt;
			if ($#prs < 1) { next; }
			while (my($p, $c) = each (%$pcnt)) {
				if ($min_pair_cnt < $glo_pair_cnt{$p} and
				    $min_pair_mi < $pair_mi{$p} and
				    0 < phrase_filter($p)) {
					print " , $p : $c";
					$tot_pairs += $c;
					$uniq_pairs{$p} += $c;
				}
			}
			print "\n";
		}
	}

	my @wlist = keys %uniq_words;
	my @plist = keys %uniq_pairs;

	print "! Min allowed word count = $min_word_cnt\n";
	print "! Min allowed phrase count = $min_pair_cnt\n";
	print "! Min allowed MI=$min_pair_mi\n";

	print "! Selected words=$#wlist  pairs=$#plist\n";
}

&print_records();
