/** argument.h --- 
 *
 * Copyright (C) 2012 OpenCog Foundation
 *
 * Author: Nil Geisweiller, Moshe Looks
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _OPENCOG_ARGUMENT_H
#define _OPENCOG_ARGUMENT_H

#include "common_def.h"
#include <boost/operators.hpp>

namespace moses3 { namespace combo {

/**
 * class argument
 *
 *  Represents the index of an input variable of a function coded as a
 *  combo tree.  For boolean arguments, a negative value corresponds to
 *  the negated literal (i.e. not of the argument).  For example,
 *  idx == -3 represents the literal NOT($3) where $3 is the third
 *  argument of the function.
 *
 *  idx == 0 is invalid.
 */
class argument
    : boost::less_than_comparable<argument>, // generate >, <= and >= given <
      boost::equality_comparable<argument>   // generate != given ==
{
public:
    arity_t idx;

    explicit argument(arity_t i) : idx(i) {
        OC_ASSERT(idx != 0, "idx should be different than zero.");
    }

    void negate() {
        idx = -idx;
    }
    bool is_negated() const {
        return idx < 0;
    }

    bool operator<(const argument& rhs) const {
        static moses3::absolute_value_order<int> comp;
        return comp(idx, rhs.idx);
    }

    bool operator==(const argument& rhs) const {
        return idx == rhs.idx;
    }

    arity_t abs_idx() const {
        return idx < 0 ? -idx : idx;
    }
    const static arity_t idx_to_abs_idx_from_zero(arity_t other_idx) {
        return (other_idx < 0 ? -other_idx : other_idx) - 1;
    }
    const static arity_t idx_from_zero_to_idx(arity_t idx_from_zero) {
        return idx_from_zero + 1;;
    }
    /// Returns 0 for argument of idx 1 or -1, 1 for idx 2 or -2, and so on
    arity_t abs_idx_from_zero() const {
        return idx_to_abs_idx_from_zero(idx);
    }
    /// Check if the index of this arg is in the possible range given arity a.
    /// Formally, idx is in the range of a if :
    /// if idx==0 then it is not valid
    /// else if a>0 (that is, no arg_list) then idx<=a
    /// else if a<0 then idx < -a
    bool is_idx_valid(arity_t a) const {
        return (idx == 0 ? false : (a > 0 ? abs_idx() <= a : abs_idx() < -a));
    }
};

} // ~namespace combo
} // ~namespace moses3

#endif // _OPENCOG_ARGUMENT_H
