/*
 * moses3/comboreduct/combo/enum_type.h
 *
 * Copyright (C) 2002-2008, 2012 Novamente LLC
 * Copyright (C) 2012 Poulin Holdings LLC
 * All Rights Reserved
 *
 * Written by Nil Geisweiller, Linas Vepstas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef _COMBO_ENUM_TYPE_H
#define _COMBO_ENUM_TYPE_H

#include <map>
#include <ostream>
#include <string>
#include <boost/thread.hpp>
#include <boost/operators.hpp>

namespace moses3 { namespace combo {

// enum_t is essentially a string but is coded as a different type
// than definite_object (or message), because it semantically denotes
// something else.
class enum_t
    : boost::less_than_comparable<enum_t>, // generate >, <= and >= given <
      boost::equality_comparable<enum_t> // generate != given ==
{
private:
    unsigned id;
    std::string _content;

    // Global table of string-to-int, so that the operator==()
    // can run efficiently (for scoring) without a string-compare.
    static unsigned enum_issued;
    static std::map<std::string, unsigned> enum_map;
    static boost::shared_mutex id_mutex;

protected:
    // Issue a unique id number for each unique string.
    unsigned get_id(const std::string& token);

    enum_t(const std::string &m, unsigned i)
    {
        _content = m;
        id = i;
    }

public:
    enum_t(const std::string &m)
    {
        _content = m;
        id = get_id (_content);
    }

    unsigned getId() const {
        return id;
    }

    std::string getContent() const {
        return _content;
    }

    bool operator==(const enum_t& m) const {
        // Compare Id's, not strings, because that just faster.
        return id == m.id;
    }

    /// Lexicographic ordering on the enum string names.
    bool operator<(const enum_t& m) const {
        return _content < m.getContent();
    }

    /// This is used by enum_str_to_vertex() to identify enums.
    static std::string prefix() {
        return "enum:";
    }

    /// Return some random enum out of the pool of all of them.
    static enum_t get_random_enum();

    /// Return the total number of enum types that have been issued.
    static size_t size();

    /// Return an enum that is gaurenteed to be not equal to any other.
    static const enum_t& invalid_enum() {
        static enum_t bad("", -1);
        return bad;
    }
};

std::ostream& operator<<(std::ostream&, const moses3::combo::enum_t&);

} // ~namespace combo
} // ~namespace moses3

#endif

