/*
 * moses3/comboreduct/reduct/action_reduction.cc
 *
 * Copyright (C) 2002-2008 Novamente LLC
 * All Rights Reserved
 *
 * Written by Nil Geisweiller
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <mutex>

#include "reduct.h"
#include "meta_rules.h"
#include "general_rules.h"
#include "action_rules.h"

namespace moses3 { namespace reduct {

const rule& action_reduction()
{
    // A note about the locking below, and the *pr pointer.  It can (and
    // has!) happened that two different threads may enter this routine
    // simltaneously.  Because c++ will defer running static initializers
    // until they are needed, then, if we did not lock below, then both
    // threads will start running the static initializers (constructors). 
    // The faster thread would have returned a rule, while the slower 
    // thread clobbered it, causing destructors to run on that rule.
    // As a result, the faster thread was found to be accessing freed
    // memory!  Ouch.  So a lock is needed.  To avoid locking *every*
    // time, the 'static rule *pr' is used to avoid locking if the
    // initializers have run at least once.
    static rule *pr = NULL;
    if (pr != NULL) return *pr;

    static std::mutex m;
    std::lock_guard<std::mutex> static_ctor_lock(m);

    static iterative r =
        iterative(sequential(//general reduction
                             downwards(level()),
	   
                             //general action reduction
                             downwards(reduce_action_if()),
                             downwards(reduce_action_action_if()),
                             downwards(reduce_const_cond_action_if()),
                             downwards(reduce_const_action_seq()),
                             downwards(reduce_empty_arg_seq()),

                             downwards(reduce_repeat_out_action_while()),
                             downwards(reduce_repeat_in_action_while()),

                             //property action reduction
                             downwards(reduce_idempotent()),
                             downwards(reduce_opposite()),
                             downwards(reduce_additive()),
                             downwards(reduce_zero_neutral()),

                             // modular argument reduction
                             downwards(reduce_modular_argument()),
                   
                             // modular argument reduction
                             preconditions_check()

                             ));

    if (pr == NULL) pr = &r;
    return *pr;
}

} // ~namespace reduct
} // ~namespace moses3 

