/*
 * moses3/comboreduct/reduct/perception_reduction.cc
 *
 * Copyright (C) 2002-2008 Novamente LLC
 * All Rights Reserved
 *
 * Written by Nil Geisweiller
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <mutex>
#include "reduct.h"
#include "meta_rules.h"
#include "general_rules.h"
#include "logical_rules.h"
#include "mixed_rules.h"
#include "perception_rules.h"

namespace moses3 { namespace reduct {

const rule& perception_reduction()
{
    // A note about the locking below, and the *pr pointer.  It can (and
    // does!) happen that two different threads may enter this routine
    // simltaneously.  Because c++ will defer running static initializers
    // until they are needed, then, if we did not lock below, then both
    // threads will start running the static initializers (constructors). 
    // The faster thread would have returned a rule, while the slower 
    // thread clobbered it, causing destructors to run on that rule.
    // As a result, the faster thread was found to be accessing freed
    // memory!  Ouch.  So a lock is needed.  To avoid locking *every*
    // time, the 'static rule *pr' is used to avoid locking if the
    // initializers have run at least once.
    static rule *pr = NULL;
    if (pr != NULL) return *pr;

    static std::mutex m;
    std::lock_guard<std::mutex> static_ctor_lock(m);

    static assum_iterative r =
       assum_iterative(sequential(downwards(level()),
                                   //simple perception rules
                                   downwards(reduce_irreflexive()),
                                   downwards(reduce_reflexive()),
                                   downwards(reduce_identity_of_indiscernibles()),

                                   //generate assumptions
                                   downwards(reduce_and_assumptions(r)),
                                   //the following is commented because due to
                                   //the fact that there is no boolean rules here
                                   //the double negation cannot be reduced
                                   //and it leads to an infinit recursion
                                   //downwards(reduce_or_assumptions(r))
                                   downwards(reduce_ultrametric()),
                                   downwards(reduce_transitive()),
                                   downwards(reduce_symmetric()),
                                   //reduce from assumptyions
                                   downwards(reduce_from_assumptions(r)),
                                   downwards(reduce_inequality_from_assumptions())
                                   )
                        );

    if (pr == NULL) pr = &r;
    return *pr;
}

} // ~namespace reduct
} // ~namespace moses3

