/** scorers/mutual_info.h ---
 *
 * Copyright (C) 2010 OpenCog Foundation
 *
 * Author: Nil Geisweiller <nilg@laptop>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _OPENCOG_FEATURE_SCORERS_MI_H
#define _OPENCOG_FEATURE_SCORERS_MI_H

#include <moses3/util/numeric.h>
#include <moses3/comboreduct/table/table.h>

namespace moses3 {

using namespace combo;

/**
 * Compute the mutual information MI(Y;X1, ..., Xn)
 *
 * where
 *    MI(Y;X1, ..., Xn) = H(X1, ..., Xn) + H(Y) - H(X1, ..., Xn, Y)
 *
 * and H(...) are the marginal entropies.
 *
 * The actual work is done in the combo table code.
 */
template<typename FeatureSet>
struct MutualInformation : public std::unary_function<FeatureSet, double>
{

    MutualInformation(const CTable& ctable)
        : _ctable(ctable) {}

    double operator()(const FeatureSet& features) const
    {
        return mutualInformation(_ctable, features);
    }

protected:
    const CTable& _ctable;
};

/**
 * Feature set scorer combining Mutual Information and Confidence. The
 * formula is as follows
 *
 * MI(fs) * confidence
 *
 * where confidence = N / (N + exp(-confi*|fs|)), the confidence of MI.
 * This is a heuristic.  In order to measure the true confidence, one could
 * compute several MI based on subsamples the dataset and estimate the
 * confidence based on the distribution of MI obtained.  (???)
 *
 * (???) I don't get it, for two reasons: 1) feature selection works
 * independently of the overall scale of teh scorers; its only the relative
 * rankings that matter.  2) In what sense is this a "confidence"?  That is,
 * if we look at the MI of a subset of the dataset, we expect the distribution
 * to be identical to the MI, measured on the whole set.  That is, the MI
 * is independent of the sample size.  (or its supposed to be, perhaps there
 * is a bug in our MI code???)
 */
template<typename FeatureSet>
struct MICScorer : public std::unary_function<FeatureSet, double>
{
    MICScorer(const ITable& it, const OTable& ot,
              double confi = 100)
        : _it(it), _ot(ot), _confi(confi) {}

    /**
     * The feature set is represented by an eda::instance encoding a
     * field of booleans. Each boolean represents whether its
     * corresponding feature is in the feature set of not.
     */
    double operator()(const FeatureSet& fs) const
    {
        double MI = mutualInformation(_it, _ot, fs);
        // double confidence = _it.size()/(_it.size() + _confi*fs.size());
        double confidence = _it.size()/(_it.size() + exp(-_confi*fs.size()));
        return MI * confidence;
    }

    const ITable& _it;
    const OTable& _ot;
    double _confi; //  confidence intensity
};

/// like above but using CTable instead of input and output table
template<typename FeatureSet>
struct MICScorerCTable : public std::unary_function<FeatureSet, double>
{
    MICScorerCTable(const CTable& ctable_, double confi_ = 100)
        : ctable(ctable_), confi(confi_), usize(ctable.uncompressed_size()) {}

    /**
     * The feature set is represented by an eda::instance encoding a
     * field of booleans. Each boolean represents whether its
     * corresponding feature is in the feature set of not.
     */
    double operator()(const FeatureSet& fs) const
    {
        double MI = mutualInformation(ctable, fs);
        logger().fine("MI = %e", MI);
        // double confidence = usize / (usize + confi*fs.size());
        double confidence = usize / (usize + exp(-confi*fs.size()));
        logger().fine("confidence = %e", confidence);
        return MI * confidence;
    }

    const CTable& ctable;
    double confi; //  confidence intensity
    unsigned usize; // uncompressed size
};

} // ~namespace moses3

#endif // _OPENCOG_FEATURE_SCORERS_MI_H
