/*
 * moses3/learning/moses/example-progs/scoring_iterators.h
 *
 * Copyright (C) 2002-2008 Novamente LLC
 * All Rights Reserved
 *
 * Written by Moshe Looks
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef _MOSES_SCORING_ITERATORS_H
#define _MOSES_SCORING_ITERATORS_H

namespace moses3 { namespace moses {

// Base functor for functors taking an iterator range of value type and result T
template<typename T>
struct iterator_function {
    typedef T argument_type;
    typedef T result_type;
};
        
// even_parity(x1, ..., xn) = true iff (int)x1 + ... + (int)xn is even
// where n is the arity of even_parity
struct even_parity : public iterator_function<bool>
{
    // [from, to( corresponds to the sequence of inputs of the
    // function, the result corresponds to its output
    template<typename It>
    bool operator()(It from,It to) const {
        bool parity = true;
        while (from != to)
            parity ^= *from++;
        return parity;
    }
};

// disjunction(x1, ..., xn) = true iff there exists i such that xi is true
struct disjunction : public iterator_function<bool>
{
    // [from, to( corresponds to the sequence of inputs of the
    // function, the result corresponds to its output
    template<typename It>
    bool operator()(It from,It to) const {
        while (from != to)
            if (*from++)
                return true;
        return false;
    }
};

// multiplex(a1, ..., an, d1, ..., dm) = 1 iff m = 2^n and di = 1 if i
// is the address of the string bit described by a1, ..., an.
struct multiplex  : public iterator_function<bool>
{
    multiplex(unsigned int n) : arity(n) { }
    unsigned int arity;
    // [from, to( corresponds to the sequence of inputs of the
    // function, the result corresponds to its output
    template<typename It>
    bool operator()(It from, It to) const {
        // calculate address
        unsigned int addr = 0;
        for(unsigned int i = 0; i < arity; ++i)
            if(*from++)
                addr += pow2(i);
        // return the input corresonding to that address
        return *(from+addr);
    }
};

// majority(x1, ..., xn) = 0 iff n/2 or more arguments are false
struct majority : public iterator_function<bool>
{
    majority(unsigned int n) : arity(n) { }
    unsigned int arity;
    // [from, to( corresponds to the sequence of inputs of the
    // function, the result corresponds to its output
    template<typename It>
    bool operator()(It from, It to) const {
        return std::count(from, to, true) > arity / 2;
    }
};

// simple function : f(x)_o = sum_{i={1,o}} x^i
// that is for instance:
// f(x)_3 = x+x^2+x^3
// f(x)_2 = x+x^2
// f(x)_1 = x
// f(x)_0 = 0
struct simple_symbolic_regression : public iterator_function<contin_t>
{
    simple_symbolic_regression(int o = 4) : order(o) { }
    int order;
    template<typename It>
    contin_t operator()(It from, It to) const {
        contin_t res = 0;
        dorepeat(order)
            res = (res + 1) * get_contin(*from);
        return res;
    }
};

} // ~namespace moses
} // ~namespace moses3

#endif
