/** eval-diversity.cc --- 
 *
 * Copyright (C) 2012 OpenCog Foundation
 *
 * Author: Nil Geisweiller <ngeiswei@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Tool that loads several moses outputs or tables and compute the
 * diversity of the pool of candidates.
 */

#include <boost/program_options.hpp>
#include <boost/format.hpp>

#include <moses3/util/oc_assert.h>
#include <moses3/comboreduct/table/table_io.h>
#include <moses3/learning/moses/moses/types.h>

#include "eval-diversity.h"
#include "../moses/metapopulation.h"

using namespace std;
using namespace boost::program_options;
using namespace moses3;
using namespace moses;

void log_output_error_exit(string err_msg) {
    logger().info() << "Error: " << err_msg;
    cerr << "Error: " << err_msg << endl;
    exit(1);    
}

/**
 * Display error message about not recognized diversity distance and exist
 */
void not_recognized_dst(const string& diversity_dst)
{
    stringstream ss;
    ss << diversity_dst << " is not recognized. Valid distances are "
       << p_norm << ", " << tanimoto << " and " << angular;
    log_output_error_exit(ss.str());
}

// given the sequence of distances write the results
void write_results(const eval_diversity_params& edp, const vector<float>& dsts) {
    if (edp.output_file.empty())
        ostream_results(cout, edp, dsts);
    else {
        ofstream out(edp.output_file.c_str());
        ostream_results(out, edp, dsts);
    }
}

int main(int argc, char** argv) {
    eval_diversity_params edp;
    
    // Declare the supported options.
    options_description desc("Allowed options");
    desc.add_options()
        ("help,h", "Produce help message.\n")

        ("input-file,i", value<vector<string>>(&edp.input_files),
         "DSV file containing a target feature (indicated by option -f) "
         "used to compute the distance between the other files. Can be used "
         "several times to enter several input files. "
         "The distance will be computed between each target across all files.\n")

        ("target-feature,u", value<string>(&edp.target_feature),
         "Name of the target feature to load.\n")

        ("moses-file,m", value<vector<string>>(&edp.moses_files),
         "File containing the candidates as output by moses. The output "
         "must contain the bscores (see option -t or moses). "
         "Can be used several times for several files.\n")
        
        ("output-file,o", value<string>(&edp.output_file),
         "File to write the results. If none is given it write on the stdout.\n")

        ("display-stats", value<bool>(&edp.display_stats)->default_value(true),
         "Display statistics.\n")

        ("display-values", value<bool>(&edp.display_values)->default_value(false),
         "Display the actually values (distances).\n")

        ("diversity-dst",
         value<string>(&edp.diversity_dst)->default_value(p_norm),
         str(boost::format("Set the distance between behavioral scores, "
                           "then used to determin the diversity penalty."
                           "3 distances are available: %s, %s and %s.\n")
             % p_norm % tanimoto % angular).c_str())

        ("diversity-p-norm",
         value<score_t>(&edp.diversity_p_norm)->default_value(2.0),
         "Set the parameter of the p_norm distance. A value of 1.0"
         "correspond to the Manhatan distance. A value of 2.0 corresponds to "
         "the Euclidean distance. A value of 0.0 or less correspond to the "
         "max component-wise. Any other value corresponds to the general case.\n")

        ;

        variables_map vm;
    try {
        store(parse_command_line(argc, argv, desc), vm);
    }
    catch (error& e) {
        OC_ASSERT(false, "Fatal error: invalid or duplicated argument:\n\t%s\n",
                  e.what());
    }
    notify(vm);

    if (vm.count("help") || argc == 1) {
        cout << desc << endl;
        return 1;
    }

    // set diversity parameters
    diversity_parameters diversity_params;
    // set distance
    diversity_parameters::dst_enum_t de;
    if (edp.diversity_dst == p_norm)
        de = diversity_parameters::p_norm;
    else if (edp.diversity_dst == tanimoto)
        de = diversity_parameters::tanimoto;
    else if (edp.diversity_dst == angular)
        de = diversity_parameters::angular;
    else {
        not_recognized_dst(edp.diversity_dst);
        de = diversity_parameters::p_norm; // silent compiler warning
    }
    diversity_params.set_dst(de, edp.diversity_p_norm);

    if (!edp.moses_files.empty()) {
        // load the bscores
        vector<behavioral_score> bscores;
        vector<bscored_combo_tree> bcts;
        for (const string& file : edp.moses_files) {
            ifstream in(file);
            in.exceptions(ifstream::failbit | ifstream::badbit | ifstream::eofbit);
            while (in.good()) {
                try {
                    bcts.push_back(istream_bscored_combo_tree(in));
                } catch(...) {}
            }
        }

        // compute the distances between the bscores
        vector<float> dsts;
        for (unsigned i = 0; i < bcts.size(); ++i)
            for (unsigned j = 0; j < i; ++j)
                dsts.push_back(diversity_params.dst(get_bscore(bcts[i]),
                                                    get_bscore(bcts[j])));

        // write the results
        write_results(edp, dsts);
    }

    if (!edp.input_files.empty()) {        
        // load the target features (assumed to be of float type) and
        // dump them in vectors of floats
        vector<vector<float>> targets;
        for (const string& file : edp.input_files) {
            vector<float> target;
            boost::transform(combo::loadOTable(file, edp.target_feature),
                             back_inserter(target), combo::cast_contin);
            targets.push_back(target);
        }

        // compute the distances between the targets
        vector<float> dsts;
        for (unsigned i = 0; i < targets.size(); ++i)
            for (unsigned j = 0; j < i; ++j)
                dsts.push_back(diversity_params.dst(targets[i], targets[j]));

        // write the results
        write_results(edp, dsts);        
    }
}
