/*
 * moses3/learning/moses/moses/local_moses.cc
 *
 * Copyright (C) 2002-2008 Novamente LLC
 * All Rights Reserved
 *
 * Written by Moshe Looks
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "local_moses.h"

namespace moses3 {
namespace moses {

using namespace combo;

/**
 * expand_deme -- Run one deme-creation and optimization step.
 *
 * A single step consists of representation-building, to create
 * a deme, followed by optimization, (according to the specified
 * optimizer and scoring function), and finally, a merger of
 * the unique (and possibly non-dominated) trees back into the
 * metapopulation, for potential use as exemplars for future demes.
 *
 * @param max_evals    the max evals
 *
 * @return return true if expansion has succeeded, false otherwise
 *
 */
bool expand_deme(metapopulation& mp,
                 int max_evals, time_t max_time, moses_statistics& stats)
{
    if (mp.empty())
        return true;

    // Attempt to create a non-empty representation, by looping
    // over exemplars until we find one that expands.
    // XXX When would one never expand?  Wouldn't that be a bug?
    while (1) {
        bscored_combo_tree_ptr_set_cit exemplar = mp.select_exemplar();

        // Should have found something by now.
        if (exemplar == mp.end()) {
            logger().warn(
                "WARNING: All exemplars in the metapopulation have "
                "been visited, but it was impossible to build a "
                "representation for any of them.  Perhaps the reduct "
                "effort for knob building is too high.");
            return true;
        }

        // if create_deme returned true, we are good to go.
        if (mp._dex.create_deme(*exemplar)) break;

        OC_ASSERT(false, "Exemplar failed to expand!\n");
    }

    size_t evals_this_deme = mp._dex.optimize_deme(max_evals, max_time);
    stats.n_evals += evals_this_deme;
    stats.n_expansions++;

    bool done = mp.merge_deme(mp._dex._deme, mp._dex._rep, evals_this_deme);

    if (logger().isInfoEnabled()) {
        logger().info()
           << "Expansion " << stats.n_expansions
           << " total number of evaluations so far: " << stats.n_evals;
        mp.log_best_candidates();
    }

    mp._dex.free_deme();

    // Might be empty, if the eval fails and throws an exception
    return done || mp.empty();
}

/**
 * The main function of MOSES, non-distributed version.
 *
 * This is termed "local" because it runs only on the local CPU node.
 * That is, it does not assume nor make use of any distrirubted
 * processing capabilities.  It does assume that the local environment
 * might possibly be an SMP system (i.e. generic CPU's with more-or-less
 * tightly coupled memory).
 *
 * @param mp the metapopulation
 * @param pa the parameters to run moses
 */
void local_moses(metapopulation& mp,
                 const moses_parameters& pa,
                 moses_statistics& stats)
{
    logger().info("MOSES starts, max_evals=%d max_gens=%d max_time=%d",
                  pa.max_evals, pa.max_gens, pa.max_time);

    optim_stats *os = dynamic_cast<optim_stats *> (&mp._dex._optimize);

    // Print legend for the columns of the stats.
    print_stats_header(os, mp.params.diversity.pressure > 0.0);

    struct timeval start;
    gettimeofday(&start, NULL);
    stats.elapsed_secs = 0.0;

    while ((stats.n_evals < pa.max_evals)
           && (pa.max_gens != stats.n_expansions)
           && (mp.best_score() < pa.max_score)
           && (stats.elapsed_secs < pa.max_time))
    {
        // Run a generation
        bool done = expand_deme(mp,
                                pa.max_evals - stats.n_evals, 
                                pa.max_time - stats.elapsed_secs, 
                                stats);

        struct timeval stop, elapsed;
        gettimeofday(&stop, NULL);
        timersub(&stop, &start, &elapsed);
        start = stop;
        stats.elapsed_secs += elapsed.tv_sec + 1.0e-6*elapsed.tv_usec;

        // Print stats in a way that makes them easy to graph.
        // (columns of tab-seprated numbers)
        if (logger().isInfoEnabled()) {

            stringstream ss;
            ss << "Stats: " << stats.n_expansions
               << "\t" << stats.n_evals    // number of evaluations so far
               << "\t" << ((int) stats.elapsed_secs)  // wall-clock time.
               << "\t" << mp.size()       // size of the metapopulation
               << "\t" << mp.best_score() // score of the highest-ranked exemplar.
               << "\t" << get_complexity(mp.best_composite_score()); // as above.
            if (os) {
                ss << "\t" << os->field_set_size  // number of bits in the knobs
                   << "\t" << os->nsteps  // number of iterations of optimizer
                   << "\t" << os->over_budget;  // exceeded max_evals
            }
            if (mp.params.diversity.pressure > 0.0) {
                // diversity stats over all metapopulation
                auto ds = mp.gather_diversity_stats(-1);
                ss << "\t" << ds.count // number of pairs of candidates
                   << "\t" << ds.mean  // average distance
                   << "\t" << ds.std   // average dst dev
                   << "\t" << ds.min   // min distance
                   << "\t" << ds.max;  // max distance

                // diversity stats over all best n candidates of the metapopulation
                // TODO use the option of the output
                auto best_ds = mp.gather_diversity_stats(pa.max_cnd_output);
                ss << "\t" << best_ds.count // number of pairs of candidates
                   << "\t" << best_ds.mean  // average distance
                   << "\t" << best_ds.std   // average dst dev
                   << "\t" << best_ds.min   // min distance
                   << "\t" << best_ds.max;  // max distance                
            }
            logger().info(ss.str());
        }

        // I find this particularly useful for studying diversity but
        // it could be relaxed and printed whatever
        if (logger().isDebugEnabled() and mp.params.diversity.pressure > 0.0) {
            stringstream ss;
            ss << pa.max_cnd_output << " best candidates of the metapopulation (with scores and visited status):" << std::endl;
            mp.ostream(ss, pa.max_cnd_output, true, true, true, true);
            logger().debug(ss.str());
        }
        
        // In iterative hillclimbing, it is possible (but not likely)
        // that the metapop gets empty and expand returns false.
        // Alternately, the candidate callback may urge a premature
        // termination.
        if (done) break;
    }

    logger().info("MOSES ends");
}

} // ~namespace moses
} // ~namespace moses3

