/*
 * mpi_moses.h ---
 *
 * Copyright (C) 2012 Poulin Holdings LLC
 *
 * Author: Linas Vepstas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _OPENCOG_MPI_MOSES_H
#define _OPENCOG_MPI_MOSES_H

#include <atomic>
#include <future>
#include <moses3/util/pool.h>

#include "metapopulation.h"
#include "moses_params.h"

namespace moses3 { namespace moses {

#ifdef HAVE_MPI

// MPI wrapper class.  The goal of this class is to hide all of the 
// MPI communications details into a distinct .cc file, so that none
// of the rest of the moses code is polluted with MPI header files.
// This class is effectively stateless: viz, there is no real reason
// for having a class here, other than to put all the routines in 
// a common place.
//
// It might be agood idea to turn this into a virtual base class, and
// have the distributed_moses code use the same API.  That way, we'd
// have less dispatcher code to maintain.  But this will require some
// fair amount of work and testing, and there's no pressing need to do
// this just right now. XXX TODO: do this, someday.
//
class moses_mpi_comm
{
    public:
        moses_mpi_comm();
        ~moses_mpi_comm();

        bool is_mpi_root();
        int num_workers();

        // root methods, to be used only by root node.
        void dispatch_deme(int target, const combo_tree&, int max_evals);
        int probe_for_deme();
        void recv_deme(int source, bscored_combo_tree_set&, int& n_evals);
        void send_finished(int target);

        // worker methods, to be used only by workers.
        int recv_more_work();
        void recv_exemplar(combo_tree&);
        void send_deme(const bscored_combo_tree_ptr_set&, int);

        std::atomic<size_t> sent_bytes;
        std::atomic<size_t> recv_bytes;

    protected:
        void send_tree(const combo_tree&, int target);
        void recv_tree(combo_tree&, int source);

        void send_cscore(const composite_score&, int target);
        void recv_cscore(composite_score&, int source);
};

/// mpi_moses_worker -- main loop for the worker node.
///
/// Listens for exemplars, expands them, sends back the results.
/// This is mono-threaded: i.e. both the send and receive happen
/// within the main thread.
//
void mpi_moses_worker(metapopulation& mp,
                      moses_mpi_comm& mompi);

void mpi_moses(metapopulation& mp,
               const moses_parameters& pa,
               moses_statistics& stats);


#else // HAVE_MPI

static inline void mpi_moses(metapopulation& mp,
               const moses_parameters& pa,
               moses_statistics& stats)
{
    OC_ASSERT(0, "There is no MPI support in this version of moses");
};

#endif  // HAVE_MPI

} // ~namespace moses
} // ~namespace moses3

#endif // _OPENCOG_MPI_MOSES_H
