/*
 * moses3/learning/moses/optimization/star-anneal.cc
 *
 * Copyright (C) 2002-2008 Novamente LLC
 * Copyright (C) 2012 Poulin Holdings
 * All Rights Reserved
 *
 * Written by Moshe Looks
 *            Predrag Janicic
 *            Nil Geisweiller
 *            Xiaohui Liu
 *            Linas Vepstas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <moses3/util/oc_omp.h>

#include "../moses/neighborhood_sampling.h"

#include "star-anneal.h"

namespace moses3 { namespace moses {

/////////////////////////
// Star-shaped search  //
/////////////////////////

// XXX TODO the annealing temperature control code should be ported over
// to the hill-climbing code, thus rendering the below obsolete.  The
// core problem is that the hill-climbing code is much more
// sophisticated in every way: termination conditions, exploration of
// likely instances via cross-over, sparse search when the problem
// becomes large, etc. and so the code below is esentially obsolete,
// except for the temperature-scaling idea.
//
// This ort is not urgent: it is not at all clear that simulated
// annealing (for moses) actually improves results or speeds
// convergence.  In particular, the optimal annealing schedule is
// unknown. (i.e. what the temerpatures should be at each iteration).
//
unsigned simulated_annealing::operator()(deme_t& deme,
                    const instance& init_inst,
                    const iscorer_base& iscorer,
                    unsigned max_evals,
                    time_t max_time)
{
    const field_set& fields = deme.fields();
    max_distance = opt_params.max_distance(fields);

    // @todo this should be adapted for SA
    unsigned pop_size = opt_params.pop_size(fields);
    // unsigned max_gens_total = information_theoretic_bits(deme.fields());
    size_t max_number_of_instances =
        /*(size_t)max_gens_total * */
        (size_t)pop_size;
    if (max_number_of_instances > max_evals)
        max_number_of_instances = max_evals;

    size_t current_number_of_instances = 0;

    unsigned step = 1;

    // Score the initial instance
    instance center_instance(init_inst);
    deme_inst_t scored_center_inst =
        score_instance(center_instance, iscorer);
    energy_t center_instance_energy = energy(scored_center_inst);
    double current_temp = sa_params.init_temp;

    // Logger
    {
        std::stringstream ss;
        ss << "Star search initial instance: " << fields.stream(center_instance);
        logger().debug(ss.str());
        logger().debug("Energy = %f", center_instance_energy);
    }
    // ~Logger

    do {
        unsigned current_distance = dist_temp(current_temp);

        logger().debug("Step: %u  Distance = %u, Temperature = %f",
                       step, current_distance, current_temp);

        size_t number_of_new_instances =
            sample_new_instances(sa_params.max_new_instances,
                                 current_number_of_instances,
                                 center_instance, deme,
                                 current_distance);

        // If the temperature is too high, then the distance will
        // be too large, and no instances will be found at that
        // distance.
        if (0 == number_of_new_instances) {
            current_temp = cooling_schedule((double)step
                                            * sa_params.temp_step_size);
            step++;
            continue;
         }

        // score all new instances in the deme
        OMP_ALGO::transform(next(deme.begin(), current_number_of_instances),
                            deme.end(),
                            next(deme.begin_scores(), current_number_of_instances),
                            boost::bind(boost::cref(iscorer), _1));

        // get the best instance
        deme_inst_t& best_scored_instance =
            *OMP_ALGO::max_element(next(deme.begin(), current_number_of_instances),
                                   deme.end());

        instance& best_instance = best_scored_instance.first;
        energy_t best_instance_energy = energy(best_scored_instance);

        // check if the current instance in the deme is better than
        // the center_instance
        double actual_accept_prob = sa_params.accept_prob_temp_intensity *
            accept_probability(best_instance_energy,
                               best_instance_energy, current_temp);

        if (actual_accept_prob >= randGen().randdouble()) {
            center_instance_energy = best_instance_energy;
            center_instance = best_instance;
            // Logger
            {
                std::stringstream ss;
                ss << "Center instance: "
                   << fields.stream(center_instance);
                logger().debug(ss.str());
                logger().debug("Energy = %f", center_instance_energy);
            }
            // ~Logger
        }

        current_number_of_instances += number_of_new_instances;

        current_temp = cooling_schedule((double)step
                                        * sa_params.temp_step_size);
        step++;
    } while(current_number_of_instances < max_number_of_instances &&
            current_temp >= sa_params.min_temp &&
            center_instance_energy > energy(opt_params.terminate_if_gte));

    return current_number_of_instances;
}

} // ~namespace moses
} // ~namespace moses3

