/** Counter.h --- 
 *
 * Copyright (C) 2011 OpenCog Foundation
 *
 * Author: Nil Geisweiller
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _OPENCOG_COUNTER_H
#define _OPENCOG_COUNTER_H

#include <map>
#include <initializer_list>
#include <boost/operators.hpp>
#include <boost/range/numeric.hpp>
#include <boost/range/adaptor/map.hpp>
#include <moses3/util/foreach.h>

namespace moses3 {

using boost::adaptors::map_values;

/**
 * Class that mimics python Counter container
 */

template<typename T, typename CT>
class Counter : public std::map<T, CT>,
    boost::addable<Counter<T, CT>>
{
protected:
    // this will be replaced by C++11 constructor delegation instead
    // of init
    template<typename IT>
    void init(IT from, IT to) {
        while(from != to) {
            this->operator[](*from) += 1;  // we don't use ++ to put the
                                           // least assumption on on CT
            ++from;
        }
    }

public:
    typedef std::map<T, CT> super;
    typedef typename super::value_type value_type;

    Counter() {}

    template<typename IT>
    Counter(IT from, IT to)
    {
        init(from, to);
    }

    template<typename Container>
    Counter(const Container& c)
    {
        init(c.begin(), c.end());
    }

    Counter(const std::initializer_list<value_type>& il)
    {
        for(const auto& v : il)
            this->operator[](v.first) = v.second;
    }

    /// Return the count of a key, possibly returning a default if none
    /// is present. That method different that operator[] because it
    /// doesn't insert the element if it is not present. This is very
    /// useful for multi-threading programming, by helping avoid race
    /// conditions.
    CT get(const T& key, CT c = CT()) const
    {
        typename super::const_iterator it = this->find(key);
        return it == this->cend()? c : it->second;
    }
    
    /// Return the total of all counted elements
    CT total_count() const
    {
        return boost::accumulate(*this | map_values, 0);
    }

    /// Return the element that occurs most frequently
    T most_frequent() const
    {
        T key = super::begin()->first;
        CT cnt = super::begin()->second;
        for (const auto& v : *this) {
            if (cnt < v.second)
                key = v.first;
        }
        return key;
    }

    
    // add 2 counters, for example
    // c1 = {'a':1, 'b':1}
    // c2 = {'b':1, 'c':3}
    // after
    // c1 += c2
    // now
    // c1 = {'a':1, 'b':2, 'c':3}
    Counter& operator+=(const Counter& other) {
        for (const auto& v : other)
            this->operator[](v.first) += v.second;
        return *this;
    }
    /// @todo add method to subtract, multiply, etc Counters, or
    /// scalar and Counter, etc...
};

template<typename T, typename CT>
std::ostream& operator<<(std::ostream& out, const Counter<T, CT>& c)
{
    typedef Counter<T, CT> counter_t;
    out << "{";
    for (typename counter_t::const_iterator it = c.begin(); it != c.end();) {
        out << it->first << ": " << it->second;
        ++it;
        if(it != c.end())
            out << ", ";
    }
    out << "}";
    return out;
}

} // ~namespace moses3

#endif // _OPENCOG_COUNTER_H
