/*
 * moses3/util/exceptions.cc
 *
 * Copyright (C) 2002-2007 Novamente LLC
 * All Rights Reserved
 *
 * Written by Thiago Maia <thiago@vettatech.com>
 *            Andre Senna <senna@vettalabs.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "exceptions.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <moses3/util/platform.h>
#include <moses3/util/Logger.h>

#define MAX_MSG_LENGTH 2048

using namespace moses3;

/**
 * ----------------------------------------------------------------------------
 * StandardException class
 * ----------------------------------------------------------------------------
 */
void StandardException::parseErrorMessage(const char* fmt, va_list ap, bool logError)
{
    char buf[MAX_MSG_LENGTH];

    vsnprintf(buf, sizeof(buf), fmt, ap);
    if (logError) moses3::logger().error(buf);
    setMessage(buf);
}

void StandardException::parseErrorMessage(const char *trace, const char * msg, va_list ap, bool logError)
{
    size_t tlen = 0;
    if (trace) tlen = strlen(trace);

    char * concatMsg = new char[tlen + strlen(msg) + 1];
    *concatMsg = '\0'; // empty c-string

    strcat(concatMsg, msg);
    if (trace) strcat(concatMsg, trace);

    parseErrorMessage(concatMsg, ap, logError);

    delete [] concatMsg;
}

StandardException::StandardException() throw()
{
    message = NULL;
}

// Exceptions must have a copy constructor, as otherwise the
// catcher will not be able to see the message! Ouch!
StandardException::StandardException(const StandardException& ex) throw()
{
    message = NULL;
    if (ex.message)
    {
        message = new char[strlen(ex.message) + 1];
        strcpy(message, ex.message);
    }
}

StandardException& StandardException::operator=(const StandardException& ex) throw()
{
    message = NULL;
    if (ex.message)
    {
        message = new char[strlen(ex.message) + 1];
        strcpy(message, ex.message);
    }
    return *this;
}

StandardException::~StandardException() throw()
{
    // clear memory
    if (message != NULL) {
        delete [] message;
    }
}

const char * StandardException::getMessage() const
{
    if (message == NULL) {
        return "";
    }
    return message;
}

void StandardException::setMessage(const char * msg)
{
    // clear msg
    if (message != NULL) {
        delete [] message;
    }

    message = new char[strlen(msg) + 1];
    strcpy(message, msg);
}

/**
 * ----------------------------------------------------------------------------
 * RuntimeException class
 * ----------------------------------------------------------------------------
 */
RuntimeException::RuntimeException(const char *trace, const char* fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

RuntimeException::RuntimeException() throw()
{
}

/**
 * ----------------------------------------------------------------------------
 * XMLException class
 * ----------------------------------------------------------------------------
 */
XMLException::XMLException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * IOException class
 * ----------------------------------------------------------------------------
 */
IOException::IOException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * ComboException class
 * ----------------------------------------------------------------------------
 */
ComboException::ComboException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * IndexErrorException class
 * ----------------------------------------------------------------------------
 */
IndexErrorException::IndexErrorException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * InvalidException class
 * ----------------------------------------------------------------------------
 */
InvalidParamException::InvalidParamException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap, false);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * InconsistenceException class
 * ----------------------------------------------------------------------------
 */
InconsistenceException::InconsistenceException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * FatalErrorException class
 * ----------------------------------------------------------------------------
 */
FatalErrorException::FatalErrorException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * NetworkException class
 * ----------------------------------------------------------------------------
 */
NetworkException::NetworkException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    parseErrorMessage(trace, fmt, ap);
    va_end(ap);
}

/**
 * ----------------------------------------------------------------------------
 * NotFoundException class
 * ----------------------------------------------------------------------------
 */
NotFoundException::NotFoundException(const char * trace, const char * fmt, ...) throw()
{
    va_list  ap;
    va_start(ap, fmt);
    
    char * concatMsg = new char[strlen(getMessage()) + strlen(trace) + 1];
    *concatMsg = '\0'; // empty c-string
    
    strcat(concatMsg, getMessage());
    strcat(concatMsg, trace);

    char buf[MAX_MSG_LENGTH];
    vsnprintf(buf, MAX_MSG_LENGTH, concatMsg, ap);
    setMessage(buf);

    va_end(ap);

    delete [] concatMsg;
}

/**
 * ----------------------------------------------------------------------------
 * AssertionException class
 * ----------------------------------------------------------------------------
 */
AssertionException::AssertionException(const char* fmt, ...) throw()
{
    char    buf[MAX_MSG_LENGTH];

    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);

    setMessage(buf);
    moses3::logger().error(buf);
}

AssertionException::AssertionException(const char* fmt, va_list ap) throw()
{
    char    buf[MAX_MSG_LENGTH];

    vsnprintf(buf, sizeof(buf), fmt, ap);
    setMessage(buf);
    moses3::logger().error(buf);
}

