/*
 * util/pool.h --- 
 *
 * Copyright (C) 2012 Poulin Holdings LLC
 *
 * Author: Linas Vepstas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _OPENCOG_UTIL_POOL_H
#define _OPENCOG_UTIL_POOL_H

#include <queue>
#include <mutex>
#include <condition_variable>

namespace moses3 {

/// Thread-safe blocking resource allocator.
/// If there are no resources to borrow, then the borrow() method will
/// block until one is given back.
//
// Sadly, there is nothing in boost that does this.  The boost::pool
// author clearly didn't understand the problem.  Oh well.
//
// This is implemented on top of std::queue, for no particular reason;
// pretty much any other container would do.
//
template<typename Resource>
class pool
{
    public:
        /// Fetch a resource from the pool. Block if the pool is empty.
        /// If blocked, this will unblock when a resource is put into
        /// the pool.
        Resource& borrow()
        {
            std::unique_lock<std::mutex> lock(mu);
            while (objs.size() == 0) {
                cond.wait(lock);
            }
            Resource& rv = objs.front();
            objs.pop();
            return rv;
        }

        /// Put a resource into the pool.  If the pool is empty, and
        /// other threads are blocked and waiting, this will release
        /// some other blocked thread.
        void give_back(Resource& obj)
        {
            std::lock_guard<std::mutex> lock(mu);
            objs.push(obj);
            cond.notify_one();
        }

        size_t available()
        {
            return objs.size();
        }

    private:
        std::mutex mu;
        std::condition_variable cond;
        std::queue<Resource> objs;
};


} // ~namespace moses3

#endif // _OPENCOG_UTIL_POOL_H
