/** random.h --- 
 *
 * Copyright (C) 2010 Novamente LLC
 *
 * Author: Nil Geisweiller
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _OPENCOG_RANDOM_H
#define _OPENCOG_RANDOM_H

#include <boost/numeric/conversion/cast.hpp>

#include "RandGen.h"
#include "mt19937ar.h"
#include "numeric.h"

/**
 * This file contains a collection of random generators based on RandGen
 */

namespace moses3 {

// choose uniformly randomly an element of the set s
// WARNING : it is assumed that s is non-empty
template<typename T>
T randset(const std::set<T>& s, RandGen& rng = randGen())
{
    OC_ASSERT(!s.empty());
    return *std::next(s.begin(), rng.randint(s.size()));
}

// return a random number sampled according to a Gaussian
// distribution. If the number falls out of the range of T then it is
// automatically truncated.
template<typename T>
T gaussian_rand(T mean, T std_dev, RandGen& rng = randGen())
{
    double val = mean + std_dev *
        std::sqrt(-2 * std::log(rng.randDoubleOneExcluded())) * 
        std::cos(2 * PI * rng.randDoubleOneExcluded());
    T res;
    try {
        res = boost::numeric_cast<T>(val);
    } catch(boost::numeric::positive_overflow&) {
        res = std::numeric_limits<T>::max();
    } catch(boost::numeric::negative_overflow&) {
        res = std::numeric_limits<T>::min();
    }
    return res;
}

// linear biased random bool, b in [0,1] when b tends to 1 the result
// tends to be true
static inline bool biased_randbool(float b, RandGen& rng = randGen()) {
    return b > rng.randfloat();
}

} // ~namespace moses3

#endif // _OPENCOG_RANDOM_H
