#!/usr/bin/env python
#
# reuters.py -- Create traing dataset from the Reuters corpus.
#
# Usage: reuters.py outfile <train|test> <10|90>
#    outfile specifies a filename to which data will be written
#    specify "train" to output training data
#    specify "test" to output test data
#    specify "10" to output top-10 categories
#    specify "90" to output top-90 categories
#
# Copyright (c) 2012 Poulin Holdings
#
# Use 4 whitespaces for indentation, no tab.

import collections
import os
import re
import sys

# parse input string, extracting a value for a keyword.
# Typical input strings look like this:
#    <REUTERS TOPICS="BYPASS" LEWISSPLIT="TRAIN" CGISPLIT="TRAINING-SET"  OLDID="3795" NEWID="14812">
# Thus, asking for TOPICS should return BYPASS (without the quotes)
# Assert on error.
def get_value(line, key) :
    sidx = line.find(key)
    if -1 == sidx :
        print "Aiiiii: can't find keyword ", key
        exit(1)
    sidx += len(key) + 2
    eidx = line.find("\"", sidx)
    if -1 == eidx :
        print "Aiiiii: unquoted keyword ", key
        exit(1)
    return line[sidx:eidx]


# type for dealing with multiple value tokens on a single line.
Token = collections.namedtuple('Token', ['str', 'idx'])

# Extract a category.
# A typical line looks like this:
# <PLACES><D>japan</D><D>france</D><D>japan</D><D>usa</D><D>uk</D><D>canada</D><D>west-germany</D></PLACES>
# Note japan is double-counted.
def get_next_cat(line, idx) :
    sidx = line.find("<D>", idx)
    if sidx == -1 :
        return Token ("", idx=sidx)
    sidx += 3
    eidx = line.find("</D>", sidx)
    if eidx == -1 :
        print "Ohhh noooo: parse fail! ", line
        exit(1)
    return Token(line[sidx:eidx], idx=eidx+4)

# Class holding the main results of the parse.
class Stats:
    def __init__(self) :
        # counts of the number of documents, by type
        self.train_cnt = 0
        self.test_cnt = 0
        self.other_cnt = 0

        # Categories
        self.cats = collections.Counter()
        self.train_cats = collections.Counter()
        self.test_cats = collections.Counter()

def count_words(line) :
    words = re.findall('\w+', line)
    return collections.Counter(words)


# read a Reuters SGML file, extract category and training set info.
def categorize_sgm(fileName, stats):
    print "Categorizing file: ", fileName
    infile = open(fileName)

    skipdoc = False
    intrain = False
    intest = False
    doccats = collections.Counter()

    for line in infile:
        if line.find("<REUTERS") == 0 :
            topics = get_value(line, "TOPICS")
            lsplit = get_value(line, "LEWISSPLIT")

            # Compute the modApte split:
            # Result should be 9603 train and 3299 test docs
            if lsplit == "TRAIN" and topics == "YES" :
                stats.train_cnt += 1
                intrain = True
                skipdoc = False
            elif lsplit == "TEST" and topics == "YES" :
                stats.test_cnt += 1
                intest = True
                skipdoc = False
            else :
                stats.other_cnt += 1
                skipdoc = True

            # Our work here is done
            continue

        # skip the rest of the document
        if (line.find("</REUTERS>") != 0 and skipdoc) :
            continue

        # Process the category
        # Extract the category name and increment a counter.
        if (line.find("<PLACES>") == 0 or
            line.find("<PEOPLE>") == 0 or
            line.find("<TOPICS>") == 0 or
            line.find("<EXCHANGES>") == 0 or
            line.find("<ORGS>") == 0 ) :
            idx = 0
            while 0 <= idx :
                tok = get_next_cat(line, idx)
                idx = tok.idx
                if 0 <= idx :
                    stats.cats[tok.str] += 1
                    if intrain :
                        stats.train_cats[tok.str] += 1
                    if intest :
                        stats.test_cats[tok.str] += 1
            continue

        # reset everything at the end of the document
        if line.find("</REUTERS>") == 0 :
            skipdoc = False
            intrain = False
            intest = False
            continue

    infile.close()

# read and process a Reuters SGML file
def process_sgm(fileName, outfile, cats, train):
    print "Processing file: ", fileName
    infile = open(fileName)

    skipdoc = False
    inbody = False
    # We use a counter for doccats, not a list, since, sometimes,
    # categories are double-listed in the input data.
    doccats = collections.Counter()
    wordcnt = collections.Counter()

    for line in infile:
        if line.find("<REUTERS") == 0 :
            topics = get_value(line, "TOPICS")
            lsplit = get_value(line, "LEWISSPLIT")

            # Obey the modApte split:
            # There are 9603 train and 3299 test docs
            if lsplit == "TRAIN" and topics == "YES" :
                skipdoc = not train
            elif lsplit == "TEST" and topics == "YES" :
                skipdoc = train
            else :
                skipdoc = True

            # Our work here is done
            continue

        # skip the rest of the document
        if (line.find("</REUTERS>") != 0 and skipdoc) :
            continue

        # Extract the category names
        if (line.find("<PLACES>") == 0 or
            line.find("<PEOPLE>") == 0 or
            line.find("<TOPICS>") == 0 or
            line.find("<EXCHANGES>") == 0 or
            line.find("<ORGS>") == 0 ) :
            idx = 0
            while 0 <= idx :
                tok = get_next_cat(line, idx)
                idx = tok.idx
                if 0 <= idx and tok.str in cats :
                    doccats[tok.str] += 1
            continue

        # reset everything at the end of the document
        if line.find("</REUTERS>") == 0 :
            skipdoc = False
            inbody = False
            doccats.clear()
            continue

        # ---------------------------
        # If we are here, then the document is either a train or
        # a test document; all other docs have been skipped.


        # Count the words in the body.
        if 0 <= line.find("<BODY>") :
            # Skip the rest of the document, if its no cats
            if len(doccats) == 0 :
                skipdoc = True
                continue
            idx = line.find("<BODY>") + 6
            wordcnt.clear()
            wordcnt |= count_words(line[idx:])
            inbody = True
            continue

        if 0 <= line.find("</BODY>") :
            idx = line.find("</BODY>")
            wordcnt |= count_words(line[:idx])
            # print wordcnt
            for cat in doccats :
                outfile.write(cat + ", ")
                for word in wordcnt :
                    outfile.write(word + " : " + str(wordcnt[word]) + ", ") 
                outfile.write("\n")
            inbody = False
            wordcnt.clear()
            continue

        if inbody :
            wordcnt |= count_words(line)
            continue

    infile.close()

def usage() :
    print "Usage:", sys.argv[0], "outfile <train|test> <10|90>"
    print "    outfile specifies a filename to which data will be written"
    print "    specify \"train\" to output training data"
    print "    specify \"test\" to output test data"
    print "    specify \"10\" to output top-10 categories"
    print "    specify \"90\" to output top-90 categories"

def main():
    if len(sys.argv) != 4 :
        print "Wrong number of arguments"
        usage()
        exit(1)

    # First, we categorize and count documents, categories.
    stats = Stats()
    for i in range(0,22) :
        # example filename:  reut2-014.sgm
        filename = 'reut2-%(n)03d.sgm' % {"n":i}
        categorize_sgm(filename, stats)

    print "train document cnt=", stats.train_cnt
    print "test  document cnt=", stats.test_cnt
    print "other document cnt=", stats.other_cnt

    # post-process the categories to remove those that don't
    # have both a train and a test document. & here means set
    # intersection. 
    print "total categories: ", len(stats.cats)
    stats.cats &= stats.train_cats & stats.test_cats
    print "cats after trimming: ", len(stats.cats)
    #  print "cats: ", stats.cats
    for cat in stats.cats :
        if 0 == stats.train_cats[cat] :
            print "Empty train category: ", cat, stats.cats[cat]
        if 0 == stats.test_cats[cat] :
            print "Empty test category: ", cat, stats.cats[cat]

    print "top ten cats: ", stats.cats.most_common(10)

    # Next, we generate actual output.
    outfilename = str(sys.argv[1])
    print "output file: ", outfilename
    outfile = open(outfilename, "w")

    daset = str(sys.argv[2])
    train = False
    if daset == "train" :
        train = True
    elif daset == "test" :
        train = False
    else :
        print "Unrecognized dataset type:", daset
        print "Specify \"test\" or \"train\""
        usage()
        exit(1)

    catclass = str(sys.argv[3])
    catcount = 0
    if catclass == "10" :
        catcount = 10
    elif catclass == "90" :
        catcount = 90
    else :
        print "Unrecognized category class:", catclass
        print "Specify \"10\" or \"90\""
        usage()
        exit(1)

    print "Will output", daset, "dataset, consisting of top", catcount, "categories"

    # strip out the counts, list only the top ten cats
    cats = zip(*stats.cats.most_common(catcount))[0]
    print "Will output only the following cats:", cats

    for i in range(0,22) :
        # example filename:  reut2-014.sgm
        filename = 'reut2-%(n)03d.sgm' % {"n":i}
        process_sgm(filename, outfile, cats, train)

    outfile.close()

main()

