#
# dataset.py -- POSES datafile reader library.
#
# This file defines the core datafile reader class
#
# Copyright (c) 2011, 2012 Poulin Holdings LLC
#
import csv
import math
import re     # regular expressions
import sys

from datetime import datetime
from itertools import product

# Poses stuff
from utils import ge

# Append the threshold to a variable name, using a certain format.
#
# Use the parse_var() routine below, to dundo this connection.
#
def gen_var_n_thres(var, thres):
    # return var + "_t" + str(thres)
    # ten decimal places are barely enough for normalized counts.
    return var + "_t%.10f" % thres

# Append the windows lookback to a variable name in a certain format.
def gen_var_n_win(var, lookback):
    return var + "_w" + str(lookback)

# A variable name may or may not have the threshold encoded in it.
#
# A variable name with an encoded threshold is of the form
#   "blah_t0.12345"
# with the "blah" being the variable name, and 0.12345 being the
# threshold.   This routine looks for such encoded names; if it finds
# one, then it splits it up into parts, else it returns the raw name.
# That is, it returns a triplet, the firt elt being the variable name,
# the second being Ture if there is a threshold, and the third being the
# theshold value.
#
def parse_var(var):
    # regular expression to recognize a float
    fre = "[-+]?\d*\.?\d+"
    m = re.match("((?:\w|_)+)_t(" + fre + ")$", var)
    if m :
        return m.group(1), True, float(m.group(2))
    else :
        return var, False, 0


# Strip out lines begining with comments, from out of the input file.
# This class acts as a wrapper around an iterated file handle; it
# returns the next line in a file that is not commented out.
class comment_stripper:
    def __init__(self, f):
        self.f = f
    def next(self):
        line = self.f.next()
        while (line.startswith("#") or
               line.startswith(";") or
               line.startswith("!")):
            line = self.f.next()
        return line
    def __iter__(self):
        return self


def boolStr(b):
    return "1" if b else "0"


# Fold class holds a pair of datasets, the test and the training
# datasets. It is used for k-fold cross-validation runs.
class Fold:
    def __init__(self, train, test):
        self.train = train
        self.test = test


# The main dataset structure.  Reads in the data set, computes means
# and standard deviations for each column, partitions into bins, etc.
class Dataset:

    # Returns the number of samples (rows) in this dataset
    def numberOfPoints(self):
        return len(self.strusaver)

    # Does this dataset contain the variable/feature/column named
    # 'varName'?
    def containsVar(self, varName):
        return varName in self.varnames

    # Is 'varName' an enum-valued variable?
    def isEnumVar(self, varName):
        return varName in self.enumvars

    # Return the value of variable 'var' in row 'row'
    def getVal(self, row, var):
        return self.strusaver[row].get(var, self.default)

    # Set the value of variable 'var' in row 'row' to 'newval'
    def setVal(self, row, var, newval):
        d = self.strusaver[row]
        if d.has_key(var) :
            d[var] = newval

    # Return the value of the target variable in row 'row'
    def getTargetVal(self, row):
        return self.getVal(row, self.targetVar)

    # Return the set of different possible values that the
    # variable/feature/column 'var' ranges over.  The set is made
    # unique, that is, each value appears in the set only once.
    def getRange(self, var):
        return set(self.valuemap[var])

    # Return the set of different possible values that the
    # target/output variable/feature/column ranges over.
    # The set is made unique, that is, each value appears in
    # the set only once.
    def getTargetRange(self):
        return self.getRange(self.targetVar)

    # Return the count of how many rows take the indicated value.
    def getRangeCount(self, var, val):
        return sum(1 for v in self.valuemap[var] if v == val)

    # Return the count of how many rows take the indicated value.
    def getTargetRangeCount(self, val):
        return self.getRangeCount(self.targetVar, val)

    # Compute dict of {var, list of vals}
    # This can just be thought of as the transpose of the input table,
    # flipping the rows for the columns.  The valuemap will have as key
    # the column heading, followed by a list of the values.
    def computeValuemap(self):
        before = datetime.now()
        self.valuemap = {v:[s.get(v, self.default) for s in self.strusaver]
                         for v in self.varnames}
        if self.verbose :
            after = datetime.now()
            print "# Elapsed time to computeValumap: ", after - before

    # Normalize each row to unit intensity.  That is, the values in
    # each row will be rescaled so that they add up to 1.0, exactly.
    def normalize(self):
        pv = self.targetVar
        for row in range(0, self.numberOfPoints()):
            rowsum = 0.0
            for var in self.varnames :
                if var not in self.enumvars and var != pv:
                    rowsum += self.getVal(row, var)
            if 0.0 < rowsum :
                for var in self.varnames :
                    if var not in self.enumvars and var != pv:
                        newval = self.getVal(row, var) / rowsum
                        self.setVal(row, var, newval)


    # Compute the column totals for each cohort (each target value)
    def cohortTotals(self):
        assert len(self.valuemap) != 0
        pv = self.targetVar
        numTargetVals = len(self.valuemap[pv])
        for var in self.varnames :
            if var not in self.enumvars and var != pv:

                # zero out the column sums
                colsum = {}
                for pval in self.valuemap[pv] :
                    colsum[pval] = 0.0
                # sum it up
                for row in range(0, self.numberOfPoints()):
                    pval = self.getVal(row, pv)
                    colsum[pval] += self.getVal(row, var)

                # Record the result
                self.cohort_sums[var] = colsum

    # Crazy hack to test stuff out...
    def cohortHack(self, cohort):
        assert len(self.meanmap) != 0
        if not 0 < self.numInputThr :
            return

        # Thunk the mean
        for var in self.cohort_sums.keys() :
            colsum = self.cohort_sums[var]
            vsum = colsum[cohort]
            self.meanmap[var] += vsum / len(self.valuemap[var])

        # redo the std devs.
        self.computeStds()
        for var in self.cohort_sums.keys() :
            colsum = self.cohort_sums[var]
            vsum = colsum[cohort]
            std = self.stdmap[var]
            std = std*std
            self.stdmap[var] = std + (vsum - self.meanmap[var])**2 / len(self.valuemap[var])
            self.stdmap[var] = math.sqrt(self.stdmap[var])

        # XXX Should update minmap/maxmap too.



    # Compute the mean of each variable, except for the target, for
    # which we compute the median. (Note that means are not computed
    # for the input variables, if they are not being thresholded.
    # That's because we have no use for a mean in that case).
    def computeMeans(self):
        assert len(self.valuemap) != 0
        pv = self.targetVar
        # Calculate the mean of each feature (but only if thresholding;
        # there is no nead for a mean if not thresholding)
        if 0 < self.numInputThr :
            self.meanmap = {v:sum(self.valuemap[v]) / len(self.valuemap[v])
                            for v in self.valuemap if v not in self.enumvars}
        else :
            if pv not in self.enumvars :
                self.meanmap = {pv:sum(self.valuemap[pv]) / len(self.valuemap[pv])}
        # Calculate the median (not the mean) of the target variable
        # enums don't have a median.
        if pv not in self.enumvars :
            sortTargetValues = sorted(self.valuemap[pv])
            self.targetMedian = sortTargetValues[len(sortTargetValues)/2]

    # Compute the standard deviation of each variable.
    def computeStds(self):
        if 0 >= self.numInputThr :
            return
        assert len(self.valuemap) != 0
        assert len(self.meanmap) != 0
        before = datetime.now()
        # Calculate the standard deviation of each feature
        self.stdmap = {v:math.sqrt(sum((val - self.meanmap[v])**2
                                       for val in self.valuemap[v]
                                       if v not in self.enumvars)
                                   / len(self.valuemap[v]))
                       for v in self.valuemap}
        if self.verbose :
            after = datetime.now()
            print "# Elapsed time for computeStds: ", after - before

    # Find the minimum of each variable.
    def computeMins(self):
        assert len(self.valuemap) != 0
        if 0 < self.numInputThr :
            self.minmap = {v:min(self.valuemap[v])
                           for v in self.varnames if v not in self.enumvars}
        else :
            pv = self.targetVar
            if pv not in self.enumvars :
                self.minmap[pv] = min(self.valuemap[pv])

    # Find the maximum of each variable.
    def computeMaxs(self):
        assert len(self.valuemap) != 0
        if 0 < self.numInputThr :
            self.maxmap = {v:max(self.valuemap[v])
                           for v in self.varnames if v not in self.enumvars}
        else :
            pv = self.targetVar
            if pv not in self.enumvars :
                self.maxmap[pv] = max(self.valuemap[pv])

    # Compute the list of input thresholds for each variable.
    # Threshold zero is located at the mean value of the variable.
    # If delta is zero or negative, this is the only threshold created.
    #
    # If delta is positive, additional thresholds are computed. These
    # are located in steps of the stddev times delta to each side, so
    # that the n'th threshold is at
    #      thr_n = mean + n * delta * stddev
    # for that variable.
    #
    # No more than numInputThr are created. If numInputThr is even,
    # then thr_0 is not created.  If numInputThr is zero, no thresholds
    # are computed.
    #
    # Only the threshold map is initialized here, the data are not
    # binned by this routine.  (The map is just a list of values of
    # the thresholds for each variable; there are enough thresholds to
    # cover the entire range of values for that variable.)
    def computeThresholds(self):
        if 0 == self.numInputThr :
            return
        assert len(self.valuemap) != 0
        assert len(self.meanmap) != 0
        assert len(self.stdmap) != 0
        assert len(self.minmap) != 0
        assert len(self.maxmap) != 0

        # Initialize the threshold map.
        self.thresholdmap = {v:[] for v in self.stdmap}
        # If there's an odd number of thresholds; include the center.
        # If delta isn't positive, this will be the only theshold.
        if 1 == self.numInputThr%2 :
            for v in self.stdmap :
                if v not in self.enumvars :
                    self.thresholdmap[v].insert(0, self.meanmap[v])
        if self.delta > 0:
            maxthr = 1 + (self.numInputThr - self.numInputThr%2) / 2;
            for v in self.stdmap :
                if v not in self.enumvars :
                    # Return the threshold value for the given index.
                    # idx is assumed to be a positive or negative int.
                    def threshold(idx):
                        return self.meanmap[v] + idx*self.stdmap[v]*self.delta

                    # Determine the minimum threshold
                    i = -1
                    while self.minmap[v] <= threshold(i) and -i < maxthr :
                        self.thresholdmap[v].insert(0, threshold(i))
                        i -= 1
                    # Determine the maximum threshold
                    i = 1
                    while self.maxmap[v] >= threshold(i) and i < maxthr :
                        self.thresholdmap[v].append(threshold(i))
                        i += 1

    # All-in-one wrapper for computing means, stdevs and thresholds.
    # This is meant to be called only once, immediately after reading
    # in a tabular data file.
    def computeUsefulInfo(self):
        before = datetime.now()
        self.computeValuemap()
        if self.do_normalize :
            self.normalize()
        self.cohortTotals()
        self.computeMeans()
        self.computeMins()
        self.computeMaxs()

        # Standard devs and thresholds are needed only for the boolean-
        # thresholding mode.
        if 0 < self.numInputThr :
            self.computeStds()
            # self.cohortHack(2.0) # XXX hack alert borken for everyone!
            self.computeThresholds()

        if self.verbose :
            after = datetime.now()
            print "# Total elapsed time for initial compute: ", after - before
            before = after

    # Return a pair with
    # 1) list of windowed maps {var, val}, where var is of the
    # form input_n where n is a number indicating the previous
    # instants backwards. It is
    # the equivalent of self.strusaver but windowed
    # 2) list of variable names. The equivalent of self.varnames but
    # windowed
    def windowlize(self, windowSize):
        strusaver_w = []
        varnames_w = [gen_var_n_win(v, j)
                      for v, j in product(self.varnames, xrange(windowSize))]
        pv = self.targetVar
        varnames_w.append(pv)
        for i in range(windowSize+1, self.numberOfPoints()):
            window = {gen_var_n_win(v, j) : self.strusaver[i-j-1][v]
                      for v, j in product(self.varnames, range(windowSize))
                      if v in self.strusaver[i-j-1]}
            window[pv] = self.getVal(i, pv) / self.getVal(i-1, pv)
            strusaver_w.append(window)
        return strusaver_w, varnames_w

    # Return a windowed version of self if windowSize is > 0,
    # instances in reverse order; otherwise return self.
    def windowedVersion(self, windowSize):
        if windowSize > 0:
            output = Dataset(None, self.targetVar)
            output.strusaver, output.varnames = self.windowlize(windowSize)
            output.computeUsefulInfo()
            return output
        return self

    # foldelize: partition dataset into training and testing subsets.
    #
    # This method will return a list of "folds" partitions.  Each
    # partition is a pair of datasets, nominally called the "test" and
    # the "training" datasets.  Each is a subset of the total dataset,
    # their union is the full dataset, their intersection is empty.
    # Each maintains the min/max/stddev values of the parent dataset.
    # The paritioning is done so that for folds==N the test dataset
    # will have 1/N of the total rows, and the training dataset will
    # have (N-1)/N of the total rows.   Row selection is round-robbin.
    #
    def foldelize(self, folds, windowSize):
        wv = self.windowedVersion(windowSize)
        output = []
        has_test = folds > 1
        for i in range(folds):
            train = Dataset(None, self.targetVar, self.delta,
                            self.numInputThr, self.do_normalize)
            test = Dataset(None, self.targetVar, self.delta,
                            self.numInputThr, self.do_normalize)
            train.strusaver = [wv.strusaver[j]
                               for j in xrange(len(wv.strusaver))
                               if j%folds != i] if has_test else wv.strusaver
            train.varnames = wv.varnames
            train.enumvars = wv.enumvars
            train.default = wv.default
            train.sparse = self.sparse
            train.verbose = self.verbose
            train.computeUsefulInfo()
            if has_test:
                test.verbose = self.verbose
                test.sparse = self.sparse
                test.strusaver = wv.strusaver[i::folds]
                test.varnames = wv.varnames
                test.enumvars = wv.enumvars
                test.default = wv.default
                test.computeUsefulInfo()

            # Fold just holds the pair of datasets.
            output.append(Fold(train, test))
        return output

    def varAsBin(self, var, val):
        return ge(val, self.meanmap[var])

    # Return P((var >= threshold) == b)
    def __singleProb__(self, varname, threshold, b):
        vc = sum(b == ge(s.get(varname, self.default), threshold)
                 for s in self.strusaver)
        tc = len(self.strusaver)
        return float(vc) / tc

    # Return P((var >= vthres) == vb, (targetVar >= pthres) == pb)
    def __doubleProb__(self, varname, vb, vthres, pb, pthres):
        vc = sum(vb == ge(s.get(varname, self.default), vthres) and \
                     pb == ge(s.get(self.targetVar, self.default), pthres)
                 for s in self.strusaver)
        tc = len(self.strusaver)
        return float(vc)/tc

    # Return mutual entropy (mutual information) between var and
    # targetVar. 'pthres' is the threshold used for targetVar.
    def computeVOI(self, var, pthres):
        varname, have_thresh, vthres = parse_var(var)

        if have_thresh :
            # Its a thresholded input variable.
            output = 0.0
            for pb in (False, True):
                py = self.__singleProb__(self.targetVar, pthres, pb)
                if py:
                    for vb in (False, True):
                        px = self.__singleProb__(varname, vthres, vb)
                        pxy = self.__doubleProb__(varname, vb, vthres, pb, pthres)
                        prod = px*py
                        if prod == 0.0 or pxy == 0.0:
                            continue

                        # Accumulate the mutual information (MI) between
                        # the targetVar and thegiven variable.
                        output += (pxy/prod)*math.log(pxy/prod)
            return output
        else:
            # At this time, we don't handle the non-thresholded case.
            # I guess we could handle this by calling getRange() on each
            # value, and then looping over the range (insteal of looping
            # over T/F as above) XXX Do this someday.
            output = 0.0
            return output

    # Return P((var >= threshold) == b)
    def __singleWeight__(self, varname, threshold, b):
        vc = sum(s.get(varname, self.default) * \
                     float (b == ge(s.get(varname, self.default), threshold))
                 for s in self.strusaver)
        tc = sum(s.get(varname, self.default)
                 for s in self.strusaver)
        if (tc == 0.0) :
            return 0.0
        return vc / tc

    # Return P((var >= vthres) == vb, (targetVar >= pthres) == pb)
    def __doubleWeight__(self, varname, vb, vthres, pb, pthres):
        vc = sum(s.get(varname, self.default) * \
                     s.get(self.targetVar, self.default) * \
                     float (vb == ge(s.get(varname, self.default), vthres) and \
                     pb == ge(s.get(self.targetVar, self.default), pthres))
                 for s in self.strusaver)
        tc = sum(s.get(varname, self.default) * \
                     s.get(self.targetVar, self.default)
                 for s in self.strusaver)
        if (tc == 0.0) :
            return 0.0
        return vc/tc

    # Return mutual entropy (mutual information) between var and
    # targetVar. 'pthres' is the threshold used for targetVar.
    def computeMI(self, var, pthres):
        varname, have_thresh, vthres = parse_var(var)
        if have_thresh :
            output = 0.0
            for pb in (False, True):
                py = self.__singleWeight__(self.targetVar, pthres, pb)
                if py:
                    for vb in (False, True):
                        px = self.__singleWeight__(varname, vthres, vb)
                        pxy = self.__doubleWeight__(varname, vb, vthres, pb, pthres)
                        prod = px*py
                        if prod == 0.0 or pxy == 0.0:
                            continue

                        output += (pxy/prod)*math.log(pxy/prod)
            return output
        else :
            # At this time, we only handle the thresholded variables.
            return 0

    # Declare a set of columns that will be printed out during file
    # save. If set, only these columns will be printed (and the target
    # column too). Otherwise, all columns are printed.
    def printCols(self, cols) :
        self.prtCols = cols

    # Save the (optionally) booleanized dataset in CSV or sparse format.
    # (MOSES accepts tables in either CSV or sparse format).
    def saveToFile(self, fileName, targetVar,
                    have_output_thresh, out_threshold, balance):
        before = datetime.now()

        dataset_size = 0;
        if self.sparse :
            dataset_size = self.saveAsSparse(fileName, targetVar,
                              have_output_thresh, out_threshold, balance)
        else :
            dataset_size = self.saveAsCSV(fileName, targetVar,
                              have_output_thresh, out_threshold, balance)

        if self.verbose :
            after = datetime.now()
            print "# Elapsed time to write file: ", after - before

        return dataset_size

    # Save the booleanized dataset in plain-old CSV format.
    # (MOSES accepts tables in either CSV or tab-delimited format).
    # Return the number of data lines in the dataset.
    #
    # The first column of the table will be the target variable.
    # An additional column is created for each input-variable and
    # input-threshold combination.  That is, if some given input
    # variable has N thresholds associated with it, then N columns
    # will be created for that variable.  (By default, N=1, and the
    # theshold is simply the mean of that input variable).
    #
    # If 'balance' is false, then the number of rows written will equal
    # the number of rows in the dataset (not counting the header).
    #
    # Each row consists entirely of boolean values, 0 or 1. For each
    # input-var and threshold combination, a 0 will be written if the
    # variable is less than the threshold, and a 1 if it is greater
    # than (or equal).  Likewise for the target variable: a 0 or 1 is
    # written, depending on whether the target is less than or ge than
    # the 'threshold' argument to this method.
    #
    # If 'balance' is True, then this method attempts to balance the
    # number of positive and negative samples; that is, the number of
    # of 0 and 1 values written in the first column.
    #
    # XXX FIXME: The balancing is very crude: rows are simply dropped
    # into the bit-bucket when unbalanced; thus the outpuf file will
    # consist of alternating 1's an 0's in the target column.  This is
    # just plain sloppy... it causes a lot of data loss.
    #
    # Returns the number of rows written to the file.
    #
    def saveAsCSV(self, fileName, targetVar,
                  have_output_thresh, threshold, balance):

        # varnames excluding targetVar
        if 0 < len(self.prtCols) :
            varp = [v for v in self.prtCols if v != targetVar]
        else :
            varp = [v for v in self.varnames if v != targetVar]


        if 0 < self.numInputThr :
            # gen_var_n_thres just generates some string names in the
            # format varname + "_t" + threshold value. By default,
            # there is only one threshold per variable (located at the
            # mean), but there may be more; this will generate a string
            # name for *each* threshold.
            # 'sum' here is just list concatenation.
            varp_thresholds = sum(([gen_var_n_thres(v, t) for t in self.thresholdmap[v]]
                                  for v in varp), [])

        count = 0           # number of samples (rows written)
        count_diff = 0      # difference between positive and negative samples
        with open(fileName, "w") as outfile:

            # Write header of column names.
            if 0 < self.numInputThr :
                outfile.write(self.targetVar + ","
                              + ",".join(varp_thresholds) + "\n")
            else :
                outfile.write(self.targetVar + ","
                              + ",".join(varp) + "\n")

            # Write rows
            # strusaver is a list of maps {varname, value} for
            # each sample in the dataset (i.e. each row in the file)
            for stru in self.strusaver:
                targetVal = stru.get(targetVar, self.default)
                printRow = True
                pos = True
                if have_output_thresh :
                    if ((int == type(threshold)) or
                       (float == type(threshold))):
                        # pos is true if target var is greater than the
                        # threshold.
                        pos = ge(targetVal, threshold)
                    elif str == type(threshold) :
                        pos = (targetVal == threshold)
                    else :
                        print "Fatal Error: Unexpected threshold type!"
                        sys.exit()

                    inc = 1 if pos else -1
                    targetVal = boolStr(pos)
                    # The goal of 'balance' is to have equal numbers of
                    # positive vs negative and negative values.  This
                    # algo is crude/bad, and needs to be fixed.
                    printRow = not balance or (count_diff * inc <= 0)
                    if printRow :
                        count_diff += inc

                if printRow :
                    # boolStr is string "1" is pos is true, else "0".
                    row = str(targetVal) + ","

                    if 0 < self.numInputThr :
                        # If the variable in that row is above the
                        # threshold, write 1, else write zero.
                        #
                        # 'sum' here is just list concatenation.
                        row += ",".join(
                            sum(([boolStr(ge(stru.get(v, self.default), t))
                                  for t in self.thresholdmap[v]]
                                      for v in varp), []))
                    else :
                        row += ",".join(
                            sum(([str(stru.get(v, self.default))]
                                      for v in varp), []))
                    outfile.write(row + "\n")
                    count += 1
            outfile.close()
        return count

    # Same as above, but using the sparse file format, insead of CSV.
    # This does NOT support balancing!
    def saveAsSparse(self, fileName, targetVar,
                     have_output_thresh, threshold, balance):

        # varnames excluding targetVar
        if 0 < len(self.prtCols) :
            varp = [v for v in self.prtCols if v != targetVar]
        else :
            varp = [v for v in self.varnames if v != targetVar]

        # Accumulate sum of all values appearing in a given column.
        totals = {}

        count = 0           # number of samples (rows written)
        with open(fileName, "w") as outfile:

            # Write header consisting of single target column name.
            outfile.write(self.targetVar)
            outfile.write("\n")

            # Write rows
            # strusaver is a list of maps {varname, value} for
            # each sample in the dataset (i.e. each row in the file)
            for stru in self.strusaver:

                targetVal = stru.get(targetVar, self.default)
                if have_output_thresh :
                    # pos is true if target var is greater than the
                    # threshold.
                    pos = ge(targetVal, threshold)
                    # boolStr is string "1" is pos is true, else "0".
                    targetVal = boolStr(pos)

                row = str(targetVal) + ", "

                # If the variable in that row is above the threshold
                # write the variable name :1, else do nothing.
                for v in varp :
                    if 0 < self.numInputThr :
                        for t in self.thresholdmap[v] :
                            pos = ge(stru.get(v, self.default), t)
                            if pos :
                                # gen_var_n_thres generates string names
                                # w/format varname + "_t" + threshold value.
                                feat = gen_var_n_thres(v, t)
                                row += feat + " : 1, "
                                # Keep a running total
                                if feat in totals :
                                    totals[feat] += 1
                                else :
                                    totals[feat] = 1
                    else :
                        val = stru.get(v, self.default)
                        if val != self.default :
                            row += v + " : " + str(val) +", "

                        # Keep a running total
                        if v in totals :
                            totals[v] += val
                        else :
                            totals[v] = val
                outfile.write(row + "\n")
                count += 1

            # Print totals of values appearing in each column.
            # The totals are preceeded by a # to comment them out
            # from the dataset (i.e. they're meant for human
            # readability and debug only.)
            row = "#\n# Totals:: , "
            for v in totals.keys() :
                val = totals[v]
                if val != self.default :
                    row += v + " : " + str(val) + ", "
            outfile.write(row + "\n#\n")

            # Print out the cohort sums too.
            for pval in sorted(self.cohort_sums.keys()) :
                sums = self.cohort_sums[pval]
                row = "# Cohort:: " + str(pval) + ", "
                for v in sums.keys() :
                    val = sums[v]
                    if val != self.default :
                        row += str(v) + " : " + str(val) + ", "
                outfile.write(row + "\n")
            outfile.close()
            
        return count


    # Return true if the file appears to be in sparse CSV format.
    # SparseCSV can be recognized by having a colon, surrounded by
    # spaces, in the second non-comment row. (The first non-comment
    # row contains column titles, and thus no colons)
    def sniffSparseCSV(self, fileName):
        infile = open(fileName)
        line = infile.readline()
        # Skip over comment lines
        while (line.startswith("#") or
               line.startswith(";") or
               line.startswith("!")):
            line = infile.readline()
        line = infile.readline()
        while (line.startswith("#") or
               line.startswith(";") or
               line.startswith("!")):
            line = infile.readline()
        infile.close()
        return -1 != line.find(" : ")

    # Read a basic CSV file.
    # 'Basic' here means the standard CSV format, all features present,
    # and optional column titles. This is not for the sparse format.
    #
    # Columns may contain float pt. numeric values, or enums.
    def readSimpleCSV(self, fileName):
        infile = open(fileName)

        # Files may be comma or tab delimited. However, the
        # DictReader can't guess this while reading, it must
        # sniff this out first. So read some lines, get dialect,
        # rewind, and read again.
        strip = comment_stripper(infile)
        dialect = csv.Sniffer().sniff(strip.next())
        infile.seek(0)
        table = csv.DictReader(comment_stripper(infile), None,
                            None, None, dialect)

        # strusaver is a list of maps {varname, value} for each sample
        # (i.e. each row in the file). Only the non-zero values are
        # noted.  More precisely, only the non-self.default values are
        # noted.

        targetBase = self.targetVar

        non_zero_vals = 0
        zero_vals = 0
        enums = set()
        rmcol = set()
        for row in table:         # row is actually a dict
            kvp = {}              # key-value pairs
            for col in row:       # col is a string column name
                try:
                    # float point values get stored by column name
                    val = float(row[col])
                    if (val != self.default) :
                        kvp[col] = val
                        non_zero_vals += 1
                    else :
                        zero_vals += 1
                except ValueError, e:
                    # non-float-point values must be enums.
                    self.enumvars.add(col)
                    kvp[col] = row[col]
                except TypeError, e:
                    print ("Fatal Error: Missing value in file \""
                        + fileName + "\". There are probably more " 
                        + "column labels than values in this file!")
                    sys.exit()

            # strusaver is a list of maps {varname, value} for each
            # sample (i.e. each row in the file).
            self.strusaver.append(kvp)
        infile.close()

        if 0 == len(self.varnames) :
            self.varnames = table.fieldnames
        else :
            # Verify that the fieldnames match!
            if set(self.varnames) != set(table.fieldnames) :
                print ("Fatal Error: the input filenames do not have "
                    + "the same column headers!")
                sys.exit()

        if self.verbose and 0 < len(self.enumvars):
            print "# Enum-valued columns: ", self.enumvars

        # If most values are default, then the file is sparse...
        # This flag is checked during file output.
        if 5 * non_zero_vals < zero_vals :
            self.sparse = True

    # Read files in sparseCSV format.  This format consists of rows
    # in the form val1, val2, val3, key1 : kval1, key2 : kval2
    # The initial columns are the "fixed" columns; here there are
    # three fixed columns. These are followed by colon-separated
    # key-value pairs.  All entries must be comma-spearated; tab
    # and whitespace-delimited files are not supported.  The key-value
    # pairs must have whitespace on both sides of the colon.
    # The first line is assumed to be column labels for the fixed
    # columns.
    def readSparseCSV(self, fileName):
        self.sparse = True
        haveTitles = False
        titles = []
        fieldnames = []
        infile = open(fileName)
        for line in infile:
            # Skip over comment lines
            if (line.startswith("#") or
                   line.startswith(";") or
                   line.startswith("!")):
                continue
            entries = line.split(",")

            # Grab the column titles from the first row.
            if not haveTitles :
                for entry in entries :
                    entry = entry.strip()
                    titles.append(entry)
                haveTitles = True
                continue

            # Unpack elements of the row.
            colnum = 0
            kvp = {}                # key-value pairs
            for entry in entries :
                entry = entry.strip()

                # Ignore whitespace crap at end of line.
                if len(entry) == 0:
                    continue

                # Look for a fixed entry
                if -1 == entry.find(" : ") :
                    # Use the column titles for the fixed entries.
                    col = titles[colnum]
                    fieldnames.append(col)
                    try :
                        kvp[col] = float(entry)
                    except ValueError, e:
                        # non-float-point values must be enums.
                        kvp[col] = entry
                        self.enumvars.add(col)
                    colnum += 1
                else :
                    pair = entry.split(":")
                    try :
                        feat = pair[0].strip()
                        kvp[feat] = float(pair[1])
                        fieldnames.append(feat)
                    except ValueError, e:
                        print ("Fatal error: Bad value \"", pair[1],
                               "\" for key \"", pair[0],
                               "\" while reading dataset ", fileName, e)
                        sys.exit()

            # strusaver is a list of maps {varname, value} for each
            # sample (i.e. each row in the file).
            self.strusaver.append(kvp)
        infile.close()

        if 0 == len(self.varnames) :
            # varnames are unique, set() makes them unique
            self.varnames = list(set(fieldnames))
        else :
            # Verify that the fieldnames match!
            if set(self.varnames) != set(fieldnames) :
                print ("Fatal Error: the input filenames do not have "
                    + "the same column headers!")
                sys.exit()

        if self.verbose :
           print "Read", len(self.varnames), " distinct features"

    # Dataset constructor.
    # 'fileName' is the file holding the data.
    # 'targetVar' is the name of the dependent variable.
    # 'delta' will be used to space locations of variable thresholds.
    # 'nIn' will be the max number of input thresholds
    def __init__(self, fileNames, targetVar, delta = -1, nIn = 1, norm = False):
        self.verbose = False    # Print debug output
        self.verbose = True    # Print debug output
        self.sparse = False     # Use sparse file format for I/O
        self.targetVar = targetVar
        self.targetMedian = 0   # Median of targt variable
        self.varnames = []      # list of variables, including the target
        self.prtCols = []       # list of columns to output
        self.enumvars = set()   # variables that are enum-valued.
        self.strusaver = []     # list of {varname, value} maps.
        self.valuemap = {}      # mapping variables to list of values
        self.meanmap = {}       # mapping variables to mean of that variable.
        self.stdmap = {}        # mapping variables to standard deviation
        self.minmap = {}        # mapping variables to min value
        self.maxmap = {}        # mapping variables to max value
        self.delta = delta      # delta used to determine the input thresholds
        self.numInputThr = nIn  # max number of input thresholds
        self.thresholdmap = {}  # mapping variables to list of thresholds
        self.do_normalize = norm # whether or not to normalize rows.
        self.cohort_sums = {}   # solumn sums, over cohorts

        # Default value, useful for sparse representations.
        self.default = 0.0

        if not fileNames:
            return

        if list == type(fileNames) :
            for fi in fileNames :
                before = datetime.now()
                if self.sniffSparseCSV(fi) :
                    self.readSparseCSV(fi)
                else :
                    self.readSimpleCSV(fi)
                if self.verbose :
                    after = datetime.now()
                    print "# Elapsed time to read \"" + fi + "\"", after - before

        elif str == type(fileNames) :
            before = datetime.now()
            if self.sniffSparseCSV(fileNames) :
                self.readSparseCSV(fileNames)
            else :
                self.readSimpleCSV(fileNames)
            if self.verbose :
                after = datetime.now()
                print "# Elapsed time to read file: ", after - before

        else :
            print "Fatal Error: confused by dataset type", type(fileNames)
            sys.exit()

        # Check for a common error; this avoids cryptic errors later.
        if self.numberOfPoints() < 2:
            print ("Fatal Error: Datafile \"" + str(fileNames) +
                   "\" must contain at least two rows! ")
            sys.exit()

        # Make sure the target var is present in the dataset.
        # Must do this before the call to computeUsefulInfo.
        if targetVar and not self.containsVar(targetVar):
            print ("Fatal Error: Target variable \"" + targetVar +
                "\" not found in file \"" + str(fileNames) +
                "\".  Did you specify the right " \
                "target variable with option -p?")
            sys.exit()

        if targetVar :
            self.computeUsefulInfo()

    # Import a case file -- d is a single dictionary line.
    # XXX TODO add support for enum values
    def importCase(self, d):

        # Case files won't have a target var, so we need to add this.
        d[self.targetVar] = 0.0
        self.strusaver = [d]

        self.varnames = d.keys()
        self.computeUsefulInfo()

# =========================== END OF FILE =====================
