#
# levels.py -- POSES thresholding code.
#
# This file defines the level threshold code.
#
# Copyright (c) 2011, 2012 Poulin Holdings LLC
#
import pickle
import sys

from utils import ge

# TargetLevels -- class holding target feature level specifications.
#
# This class holds the level (or interval) specifications for the
# target variable (the target feature).  Each interval is specified by
# its two endpoints; each endpoint is called a "threshold".
# Note that the number of thresholds equals the numberofLevels-1.
# So, e.g. one threshold splits the number-line into two intervals,
# the left and right inteval.
#
# This class stores the smallest and largest possible values of the
# target variable, and the thresholds splitting up this range.  Each
# interval is given a label; the label is just the numeric value of the
# location of the center of the interval.
#
# Because enum-valued targets behave a lot like intervals/levels, this
# class is also used to manage a collection of enums.
#
class TargetLevels:
    # Append a pickled copy of self to the indicated (open) filehandle.
    def write(self, levelFile):
        levelFile.write("<levels>\n")
        pickle.dump(self, levelFile)

    # Given a numeric value, return the label for the interval in
    # which the value would sit.
    def val2label(self, val):
        for i in range(self.numberOfLevels - 1):
            if ge(val, self.__intervals__[i][0]) \
                    and not ge(val, self.__intervals__[i][1]):
                return self.labels[i]
        if ge(val, self.__intervals__[self.numberOfLevels - 1][0]):
            return self.labels[self.numberOfLevels - 1]

    # Given some partially-specified info from the user, fill out
    # the thresholds and intervals  arrays & etc. Make sure the data
    # is complete, consistent.
    def __set_thresholds__(self, thresholds, lowerNeutral, upperNeutral,
                           dataset):

        # No thresholding to be done...
        if 1 >= self.numberOfLevels :
            return

        self.thresholds=[]
        if len(thresholds) == self.numberOfLevels - 1:
            self.thresholds = thresholds
        else:
            # If we are here, then len(thresholds) < numberOfLevels - 1,
            # which was guaranteed because we checked before calling
            # this method (by the check_thresholds() function in
            # pm_impl.py.) So here, we add the missing thresholds.
            #
            if self.numberOfLevels == 2:
                # Use the mean (not the median) as the threshold.
                self.thresholds.append(dataset.meanmap[dataset.targetVar])
            elif self.numberOfLevels == 3:
                if len(thresholds) == 0:
                    threshold = dataset.targetMedian
                    self.__set_thresholds_LU__(threshold,
                                               lowerNeutral, upperNeutral)
                else:
                    assert len(thresholds) == 1
                    self.__set_thresholds_LU__(thresholds[0],
                                               lowerNeutral, upperNeutral)
            else:
                # Create an evenly spaced set of thresholds.
                assert len(thresholds) == 0
                step = self.get_step()
                self.thresholds=[self.minval + i*step
                                 for i in range(1, self.numberOfLevels)]

        # However they were specified, make sure they're in sorted
        # order.
        self.thresholds.sort()

    # Special case of set_thresholds when there are 3 levels and one
    # threshold, defined with lowerNeutral and upperNeutral.
    def __set_thresholds_LU__(self, threshold, lowerNeutral, upperNeutral):
        self.thresholds = []
        self.thresholds.append(threshold * (1.0 - lowerNeutral))
        self.thresholds.append(threshold * (1.0 + upperNeutral))

        # Avoid insane thresholds: they really should lie within
        # the range of the data.
        if self.thresholds[0] <= self.minval:
            print ("Error: The lower threshold (" +
                str(self.thresholds[0]) +
                ") is set lower than the smallest value (" +
                str(self.minval) +
                ") in the dataset. Use the -l option to set a more " +
                "reasonable lower threshold.")
            sys.exit()
        if self.thresholds[1] >= self.maxval:
            print ("Error: The upper threshold (" +
                str(self.thresholds[1]) +
                ") is set higher than the largest value (" +
                str(self.maxval) +
                ") in the dataset. Use the -u option to set a more " +
                "reasonable upper threshold.")
            sys.exit()

    def __set_intervals__(self):
        lbounds = [self.minval] + self.thresholds
        ubounds = self.thresholds + [self.maxval]
        self.__intervals__ = [(l, u) for l, u in zip(lbounds, ubounds)]

    # Create an ASCII label for each level.  The label value is set
    # to the float point value that is exactly in the middle of the
    # level; viz. the average of the two endpoints of the interval.
    def __set_labels__(self):
        self.labels = [str((l+u)/2) for l, u in self.__intervals__]

    # Given a string label, return the correspoinding index into the
    # array.
    def label_to_idx(self, label):
        return self.labels.index(label)

    # Return a step size so that the target variable range is split
    # into N-1 intervals. i.e. so that there are N thresholds.
    def get_step(self):
        return (self.maxval - self.minval) / (self.numberOfLevels - 1)

    def __init__(self, numberOfLevels, thresholds,
                 lowerNeutral, upperNeutral, dataset):

        self.targetIsEnum = dataset.isEnumVar(dataset.targetVar)

        self.numberOfLevels = numberOfLevels
        self.thresholds = []

        if not self.targetIsEnum :
            # If the target variable is an ordinary floating-point
            # number, then set up the thresholds appropriately.

            self.minval = dataset.minmap[dataset.targetVar]
            self.maxval = dataset.maxmap[dataset.targetVar]

            try:
                flts = [float(t) for t in thresholds]
            except:
                print("Fatal Error: invalid floating-point value in "
                    "thresholds: ", thresholds)
                sys.exit()

            self.__set_thresholds__(flts, lowerNeutral, upperNeutral, dataset)
            self.__intervals__ = []
            self.__set_intervals__()
            self.labels = []
            self.__set_labels__()

            # Some checks -- These should have already passed earlier
            # during options parsing.
            if 1 < self.numberOfLevels :
                assert len(self.thresholds) == self.numberOfLevels - 1
                assert len(self.labels) == self.numberOfLevels
        else:
            # If the target variable is an enumerated set, then the
            # "levels" are just the enum values.
            self.labels = dataset.getTargetRange()

            # If the user asked for more than one enum, but failed to
            # specify them by listing them, then keep only the most
            # frequent ones.
            # thresholds will be non-zero in length only if the user
            # specified thresholds with the -t flag; else the length is
            # zero.
            if ((0 == len(thresholds)) and
                (1 < self.numberOfLevels) and
                (self.numberOfLevels < len(self.labels))) :
                cnts = {v:dataset.getTargetRangeCount(v) for v in dataset.getTargetRange()}
                tops = sorted(cnts, key = lambda x:x[1])
                tops.reverse()
                del tops[self.numberOfLevels:]
                self.labels = tops
            elif 0 < len(thresholds) :
                self.labels = thresholds


# Cheap backward-compatibility hack.
# The pickled levels used to live in a module called "shared.py"
# Now they live here, in levels.py. This confused pickle. Bummer!
# To de-confuse pickle, we twiddle the module name in the pickle data.
# Its easy: its in the first line of data, as ascii text. Go for it.
class module_name_thunker:
    # f is an already-open file object handle
    def __init__(self, f) :
        self.done = False
        self.f = f
    # overloaded system file readline() method
    def readline(self, size=None) :
        if None == size :
            ln = self.f.readline()
        else :
            ln = self.f.readline(size)

        # Be transparent if done
        if self.done :
            return ln
        self.done = True

        # Lie about the module name
        if -1 == ln.find("shared") :
            return ln
        return "levels\n"

    # overloaded system file read() method
    def read(self, size) :
        return self.f.read(size)

# XXX old model files expect a class named "Levels"; the name of this class
# is now "TargetLevels".  That's the problem with pickling... lack of
# backwards compatibility.  XXX fixme. really.
# Whatever. We could handle this in the thunker above, if it was really
# urgent...
Levels = TargetLevels

def readTargetLevels(levelFile):
    return pickle.load(module_name_thunker(levelFile))

# =========================== END OF FILE =====================
