#
# model.py -- POSES Combo Model
#
# This file defines the POSES Combo model of a dataset.
#
# Copyright (c) 2011, 2012 Poulin Holdings LLC
#
import operator
import string
import sys

from collections import Counter
from itertools import product

# poses stuff
from utils import ge
from levels import readTargetLevels
from dataset import parse_var

class Node:

    # Evaluate a node
    # v2v is a dictionary of key-value pairs, mapping variable to value
    def evaluate(self, v2v):
        # modern-day literals are preceeded by a dollar sign. e.g. $varname
        # Yea olden dayes literals employed the hash mark e.g. #varname
        if self.opName[0] == "$" or self.opName[0] == "#": # positive literal
            var, have_thresh, threshold = parse_var(self.opName[1:])
            if have_thresh :
                self.otype = "bool"
                return ge(v2v.get(var, 0), threshold)
            else :
                return v2v.get(var, 0)

        if self.opName[0] == "!": # negative literal
            var, have_thresh, threshold = parse_var(self.opName[2:])
            if have_thresh :
                self.otype = "bool"
                return not ge(v2v.get(var, 0), threshold)
            else :
                return not v2v.get(var, 0)

        if self.opName == "true":
            self.otype = "bool"
            return True
        if self.opName == "false":
            self.otype = "bool"
            return False

        # If we are here, and there are no leaves, then the opName
        # is an enum value, i.e. justsome string.  Just return that.
        if 0 == len(self.leaves) :
            self.otype = "enum"
            return self.opName

        # If we are here, then we have an operator with 1 or more args.
        output = self.leaves[0].evaluate(v2v)
        if self.opName == "not":
            self.otype = "bool"
            return (not output)

        # Greater than. The arg must be float point, the returned
        # value is boolean.
        if self.opName == "0<":
            self.otype = "bool"
            return (0.0 < output)

        # Boolean ops.
        if self.opName == "and":
            self.otype = "bool"
            if not output :
                return False
            for child in self.leaves[1:]:
                if not child.evaluate(v2v) :
                    return False
            return True

        if self.opName == "or":
            self.otype = "bool"
            if output :
                return True
            for child in self.leaves[1:]:
                if child.evaluate(v2v) :
                    return True
            return False

        # conditional consists of a list of interleaved booleans
        # and values.  The value of the conditional is the value
        # that immediately follows the first true boolean. If
        # no bools are true, then the value of the cond is the last
        # child. (Thus, conds always have an odd number of children:
        # they are basically if..elif..elif..else.. chains.)
        if self.opName == "cond":
            if output :
                val = self.leaves[1].evaluate(v2v)
                self.otype = self.leaves[1].otype
                return val

            idx = 2
            while idx < len(self.leaves) :

                output = self.leaves[idx].evaluate(v2v)
                if self.leaves[idx].otype != "bool" :
                    self.otype = self.leaves[idx].otype
                    return output
                if output :
                    val = self.leaves[idx+1].evaluate(v2v)
                    self.otype = self.leaves[idx+1].otype
                    return val
                idx += 2

            print "Fatal error: malformed cond statement"
            sys.exit()

        print ("Fatal error: unsupported combo function: \"" +
            self.opName + "\"\n")
        sys.exit()

    def __init__(self, opName, level):
        self.opName = opName
        # type of the resulting value, after evaluation
        self.otype = "unknown"
        self.level = level
        self.leaves = []


# A "model" is a combo expression that "models" some dataset, together
# with a confidence weight in that model.  Each model is the result of
# a MOSES learning session.  See the moses documentation for more info
# about the combo language, and how it works.
#
# Aside from storing the model, the main focus of this class it the
# ability to evaluate that model on a given set of inputs.  That is,
# given a set of variable-name and variable-value pairs, this evaluates
# the combo program on that input, returning a result.  Both the
# inputs, and the result, may be boolean, floating point, or
# enum-valued.
class Model:

    # Evaluate the model for a given set of variable-value pairs.
    #
    # v2v is a dictionary containing key-value pairs, with the key being
    # a "variable" (feature) name.  v2v == "variable to value".
    def evaluate(self, v2v):
        return self.nodes[0].evaluate(v2v)

    # Return the output type of the model.  This is a string, taking
    # values of "bool", "contin", "enum" or "unknown".  The type will
    # be set after the first evaluation.
    def otype(self) :
        return self.nodes[0].otype

    # Returns the number of consecutive whitespaces from the begining
    # of the line, or -1 if there are none.  The indentation corresponds
    # to the depth of the combo tree (i.e. the number of open
    # parenthesis in the combo program).
    def computeDepth(self,line):
        return len(line) - len(line.lstrip())

    def mkSpaces(self, spaces):
        return " " * spaces

    # Turn a combo expression with parenthesis into an indented
    # expression, for instance
    #
    #    and($1 $2)
    #
    # becomes
    #
    #    and
    #     $1
    #     $2
    #
    # This is just a half-way step for parsing the combo program, and
    # turning it into a tree of Nodes.  There is no particularly good
    # reason for doing this; it sort of does-but-doesn't make combo
    # program parsing easier.
    #
    # Modern-day literals are preceeded by a dollar sign. e.g. $varname
    # Yea olden dayes litorals employed the hash mark e.g. #varname
    def deParenthize(self, modelLine):
        output = ""
        spaces = 0
        buf = ""
        for c in modelLine:
            if c == "(":
                output += self.mkSpaces(spaces)+buf+"\n"
                buf = ""
                spaces += 1
                continue
            if c == ")":
                output += self.mkSpaces(spaces)+buf+"\n"
                buf = ""
                spaces -= 1
                continue
            if c == " ":
                if len(buf)>0:
                    output += self.mkSpaces(spaces)+buf+"\n"
                    buf = ""
                continue
            buf += c
        return output

    # Read a single model, contained in the modelLine sting.
    def __init__(self, modelLine):
        # set of variables + thresholds names
        self.var_n_thres_names = set()
        modelPart = modelLine.partition(" ")
        # get the weight (i.e. confidence) of the model
        self.weight = float(modelPart[0])
        # get the combo program
        self.model = modelPart[2]

        vertical = self.deParenthize(self.model)
        lines = vertical.split("\n")
        self.nodes = []
        for line in lines:
            level = self.computeDepth(line)
            node_str = line.strip()
            if node_str == "":
                continue
            node = Node(node_str, level)

            # fill varnames
            # Modern-day literals are preceeded by a dollar sign. e.g. $varname
            # Yea olden dayes litorals employed the hash mark e.g. #varname
            if node_str[0] == '$' or node_str[0] == '#':
                self.var_n_thres_names.add(node_str[1:])
            elif node_str[1] == '$' or node_str[1] == '#':
                self.var_n_thres_names.add(node_str[2:])

            i = len(self.nodes)
            self.nodes.append(node)
            while i >= 0:
                if self.nodes[i].level == level-1:
                    self.nodes[i].leaves.append(node)
                    break
                i -= 1


# Write models, learned by moses, in an "ensemble" format
# A collection of moses models is an "Ensemble"
def writeEnsemble(models, modelFile, targetVar, have_output_thresh, threshold):
    modelFile.write("<" + Ensemble.tag + ">\n")
    if have_output_thresh :
        modelFile.write(targetVar + "->" + str(threshold) + "\n")
    else :
        modelFile.write(targetVar + "->" + Ensemble.no_thresh + "\n")
    for m in models:
        if m.strip() == "":
            continue
        modelFile.write(m + "\n")
    modelFile.write("</" + Ensemble.tag + ">\n")


# Class containing an ensemble of Models predicting the same target
# (or the thresholded binary value of the target).  A single model
# is a combo expression plus a confidence weight in that expression.
# The evaluate method evaluates them all and returns the majority vote.
class Ensemble:

    # Used for XML like read/write
    tag = "ensemble"
    no_thresh = "__NO_THRESHOLD__"

    # Evaluate the ensemble for a given set of variable-value pairs.
    # v2v is a dictionary containing key-value pairs, with the key being
    # a "variable" (feature) name.  v2v == "variable to value".
    def evaluateStru(self, v2v):
        outs = [m.evaluate(v2v) for m in self.models]
        weights = [m.weight for m in self.models]

        # In principle, we could verify that all models have the same
        # otype...
        self.otype = self.models[0].otype()

        if self.otype == "bool" :
            # Just a simple dot-product of weights and outputs
            w_hits = sum(o * w for o,w in zip(outs, weights))
            w_total = sum(weights)

            # Confidence is a crude measure in how much we trust this
            # result.  Range of 0.0 to 1.0
            self.confidence = 2.0 * (w_hits / w_total - 0.5)
            if self.confidence < 0.0 :
                self.confidence = -self.confidence

            # Majority vote: if more than half agree, then its true.
            return w_hits > 0.5 * w_total

        elif self.otype == "enum" :
            # Another majority vote; return the enum value that
            # is elected by the most models.
            cnt = Counter()
            for o,w in zip(outs, weights) :
               cnt[o] += w

            # The top vote-getter is the winner.
            winner = cnt.most_common(1)[0]
            self.confidence = winner[1] / sum(weights)

            return winner[0]

        else :
            print ("Fatal Error: Ensemble: unsupported output type \"" +
                self.otype + "\"\n")
            sys.exit()


    # Compute the VOI for the ensemble. The variables with same name
    # but different thresholds are considered independent (which
    # translates by the fact that they are summed via Counter). This
    # assumption is certainly not true but probably acceptable.
    #
    # XXX huh? Why counter? the {} creates a dict which prevents
    # double-counting, so its pointless to have a counter.  Surely,
    # this is either pointless or its broken to use counter here.
    # Verfiy and fix this ...
    def computeVOI(self, dataset):
        c = Counter({parse_var(v)[0]:dataset.computeVOI(v, self.threshold)
                     for v in self.var_n_thres_names})
        return c

    def computeMI(self, dataset):
        c = Counter({parse_var(v)[0]:dataset.computeMI(v, self.threshold)
                     for v in self.var_n_thres_names})
        return c

    # Read ensemble from modelFile to initialize self.
    # modelFile is assumed to be opened already, and positioned
    # at the first line of an ensemble.  Upon exit, the file is
    # positioned at the end of the ensemble.
    def __init__(self, modelFile):
        self.models = []
        self.var_n_thres_names = set() # set of variables + thresholds names
        while True:
            line = modelFile.readline()
            if line is None: break
            if "->" in line:
                tvar = line.split("->")
                self.targetVar = tvar[0]
                self.name = string.strip(tvar[1])
                self.have_output_thresh = False
                self.have_output_enum = False
                if -1 == tvar[1].find(self.no_thresh) :
                    try :
                        self.threshold = float(tvar[1])
                        self.have_output_thresh = True
                    except :
                        self.have_output_enum = True
                continue
            if "</" + Ensemble.tag + ">" in line:
                break
            if "" == line.strip() :
                continue
            model = Model(line)
            self.var_n_thres_names |= model.var_n_thres_names
            self.models.append(model)

        # The weight of an ensemble of models is the average of the
        # weights of these models.
        self.weight = 0
        if (0 < len(self.models)) :
            self.weight = sum(m.weight for m in self.models)
            self.weight /= len(self.models)

        # Type of the ensemble output. Set during evaluation.
        self.otype = "unknown"

        # Confidence in the ensemble output. Set during evaluation.
        self.confidence = 0.0

# Class containing an ensemble of ensembles of binary classifiers
# predicting whether targetVar is above a certain threshold (varying
# for each classifier). That way one can do multi-level classification.
class PowerEnsemble:

    def computeMatrix(self, dataset):

        if 1 < self.output_levels.numberOfLevels :
            # If we are here, its a traditional, thresholded
            # output.
            CM = ConfusionMatrix(self.output_levels.labels)
            for i in xrange(dataset.numberOfPoints()):
                expected = self.output_levels.val2label(dataset.getTargetVal(i))
                v2v = dataset.strusaver[i]
                modelled = self.evaluateStru(v2v)
                CM.account(expected, modelled)
            return CM

        else :
            # If we are here, its an enumerated output.
            enums = dataset.getTargetRange()
            CM = ConfusionMatrix(enums)
            for i in xrange(dataset.numberOfPoints()):
                expected = dataset.getTargetVal(i)
                v2v = dataset.strusaver[i]
                modelled = self.evaluateStru(v2v)
                CM.account(expected, modelled)
            return CM

    # temporarily disabled because scipy dependence
    # def computePearson(self, dataset):
    #     expected_data = [dataset.getExpected_idx(i)
    #                      for i in xrange(dataset.numberOfPoints())]
    #     model_data = [self.evaluateStru(dataset.strusaver[i])
    #                   for i in xrange(dataset.numberOfPoints())]
    #     return pearsonr(expected_data, model_data)[0]
    # -------------------------------------------------------

    # Returns an index, using all models with various thresholds and
    # picking up the one with highest agreement.
    #
    # v2v is a dictionary containing key-value pairs, with the key being
    # a "variable" (feature) name.  v2v == "variable to value".
    def evaluateStru(self, v2v):

        outs = [e.evaluateStru(v2v) for e in self.ensembles]
        weights = [e.weight for e in self.ensembles]
        confids = [e.confidence for e in self.ensembles]

        # In principle, we should verify that all ensembles have the
        # same otype ...
        otype = self.ensembles[0].otype
        bool_enum = self.ensembles[0].have_output_enum
        if (otype == "bool") and not bool_enum :
            # The voting algo below is as clear as mud; it is not easy
            # to understand what it does. Bear with.  The vote is 
            # represented by V_j, indicating the 'beleif' that level j
            # is the correct level.  Thus, the V_j that is largest
            # corresponds to the most likely level j.  If two or more
            # levels j get the same vote, then the middle-most j is
            # picked out of all of these.
            #
            # The vote V_j is given by
            #
            # V_j = Sum_{i=0}^{j-1} w_i*E_i(w2c) + Sum_{i=j}^{N-1} w_i*!E_i(w2c)
            #
            # where N is the number of ensembles of binary classifiers 
            # (thus, there are N+1 levels) and w_i is the weight of
            # classifier i.  E_i(w2c) is the output of the classifier
            # (model) i for data input w2c.  
            #
            # Recall that a model (classifier) i returns 0 if the
            # expected output is below the threshold, and returns 1 if
            # the expected output is above the threshold.  If the models
            # are perfect and consistent, we expect a rising esequence:
            #
            #   E_i E_i E_i E_i  E_i
            #   i=0 i=1 i=2 i=3 i=...  j=level
            #   --- --- --- --- -----  -------
            #    0   0   0   0   ...     0
            #    1   0   0   0   ...     1
            #    1   1   0   0   ...     2
            #    1   1   1   0   ...     3
            #    1   1   1   1   ...     4
            #
            # If the classifiers are not self-consistent, then other
            # pattens may occur, which is why voting is needed.
            #
            # We then return self.labels[argmax_j V_j]
            #
            V = [0] * self.output_levels.numberOfLevels
            i = 0
            for r,w in zip (outs, weights) :
                # r is boolean-valued. ^ is bit-wise xor
                R = [(r^(i>=j)) * w for j in range(len(V))]
                V = [a+b for a, b in zip(V, R)] # vector addition
                i+= 1
            # Return the median of the indices pointing to the max of V
            m = max(V)
            indices = [j for j in range(len(V)) if V[j] == m]
            idx = indices[len(indices)/2]

            # Now take a weighted reciprocal mean of the uncertainty.
            # (i.e. of 1-confidence).  The point here is that if all\
            # of the classifiers agree, then the confidence is greater
            # than just the average confidence; uncertainty decreases
            # when all of them agree.
            i = 0
            Wtot = 0
            U = [0] * self.output_levels.numberOfLevels
            C = [0] * self.output_levels.numberOfLevels
            for r,w,c in zip (outs, weights, confids) :
                Wtot += w

                # C accumulates the arithmetic mean.
                R = [(r^(i>=j)) * c *w  for j in range(len(V))]
                C = [a+b for a, b in zip(C, R)] # vector addition

                # U accumulated the reciprocal mean
                if c > 0.9999 :   # avoid divide-by-zero below.
                    c = 0.9999
                u = w / (1-c)
                # r is boolean-valued. ^ is bit-wise xor
                R = [(r^(i>=j)) * u for j in range(len(V))]
                U = [a+b for a, b in zip(U, R)] # vector addition
                i+= 1

            Cavg = C[idx] / Wtot
            Cmax = max(confids)
            uncert = Wtot / U[idx]
            Crecip = 1-uncert

            # If the individual confidence is low, and all classifiers
            # agree, then use the reciprocal mean as the confidence.
            # Else, if the individual confidence is high, and the
            # classifiers disagree, then confidence drops down in
            # linear proportion to the disagreement.
            # Note: this is ad-hoc, and not anchored by any great theory
            # or anything.  And is almost surely discontinuous.
            if Crecip > Cmax :
                self.confidence = Crecip
            else :
                self.confidence = Cavg

            # The label is a numeric value, exactly at the midpoint
            # of a target value interval.
            return self.output_levels.labels[idx]

        # the code below handles the case where the model is a
        # collection of ensembles, and each ensemble predicts a single
        # enumerated outcome as holding true, or false.  Thus, the most
        # likely enum is the one we have the most confidence in.
        elif (otype == "bool") and bool_enum :
            # If all are false, then pick the one with the lowest
            # confidence of being false.  Else, if any are true,
            # then pick the one with the highest confidence.
            winner = ""
            loser = ""
            got_one = False
            lowest = 999
            highest = -444
            names = [e.name for e in self.ensembles]
            for r,w,c,n in zip (outs, weights, confids, names) :
                # print "duuude r=", r, "w=", w, "c=", c, "n=", n
                got_one = got_one or r
                wc = w*c
                if wc < lowest :
                    lowest = wc
                    loser = n
                if highest < wc :
                    highest = wc
                    winner = n
            # If any of the ensembles claimed a match, then return
            # that one. If none claimed it, then return the least
            # confident "no".
            if got_one :
                self.confidence = highest
                return winner
            self.confidence = 1 - lowest
            return loser

        # otype "enum" happens when the moses model itself is a big
        # cond statement, so that evaluating the moses model gives an
        # enum as an output, directly.
        elif otype == "enum" :

            # Sanity check: enum targets shouldn't have levels
            assert(1 == self.output_levels.numberOfLevels)
            # A majority vote; return the enum value that
            # is elected by the most models.
            cnt = Counter()
            for o,cfd in zip(outs, confids) :
               cnt[o] += cfd

            # The top vote-getter is the winner.
            winner = cnt.most_common(1)[0]
            self.confidence = winner[1] / len(confids)

            return winner[0]

        else :
            print ("Fatal Error: PowerEnsemble: unsupported output type \"" +
                otype + "\"\n")
            sys.exit()


    def computeVOI(self, dataset):
        w2voi = sum((e.computeVOI(dataset) for e in self.ensembles), Counter())
        w2voi_l = sorted(w2voi.items(), key = lambda x:x[1])
        w2voi_l.reverse()
        return w2voi_l

    def computeMI(self, dataset):
        w2voi = sum((e.computeMI(dataset) for e in self.ensembles), Counter())
        w2voi_l = sorted(w2voi.items(), key = lambda x:x[1])
        w2voi_l.reverse()
        return w2voi_l

    def __init__(self, modelFileName):

        # self.ensembles will contain a list of Ensembles
        self.ensembles = []
        self.windowSize = 0
        modelFile = open(modelFileName)
        while True:
            line = modelFile.readline()
            # if "<WindowSize>" in line:
            #     self.windowSize=readWindowSize(modelFile)

            # After Ensemble(modelFile) returns, the current line will
            # be positioned just after the ensemble.
            if "<" + Ensemble.tag + ">" in line:
                self.ensembles.append(Ensemble(modelFile))

            # self.output_levels contains a pickled copy of the
            # TargetLevels class (which contains the thresholding info
            # for the target feature).
            # Assumed to be the last in the file.
            if "<levels>" in line:
                self.output_levels = readTargetLevels(modelFile)
                break

        # If the first ensemble has an output threshold, its assumed
        # that they all do.
        if self.ensembles[0].have_output_thresh :
            self.ensembles.sort(key=operator.attrgetter('threshold'))

        # Confidence of the prediction made
        self.confidence = 0.0


# ConfusionMatrix: maintain a simple matrix of bin-counts.
#
# In the ctor, "levels" specifies the levels into which the target
# feature has been partitioned.  Then, given a classifier and some
# inputs, the classifier will place each result into one of the bins.
# The confusion matrix just counts how often the the classifier placed
# its computed result in some bin, as compared to the expected bin.
class ConfusionMatrix:

    # Add one measurement to the confusion matrix.
    def account(self, expected, computed):
        self.data[expected][computed] += 1

    # Accumulate other CM into this one.
    # Column labels must be consistent.
    def add(self, other):
        # If self is empty, then grab labels from other
        if 0 == len(self.levelLabels) :
            self.__init__(other.levelLabels)

        if len(self.levelLabels) != len(other.levelLabels) :
            print "Fatal Error: cannot accumulate mistmatch matrixes"
            sys.exit()
        for expected, computed in product(self.levelLabels, self.levelLabels) :
            self.data[expected][computed] += other.data[expected][computed]

    # Accumulate other CM into this one.
    # level labels need not be consistent; however, the overall
    # dimension of the matrix must be the same
    def accum(self, other):
        # If self is empty, then grab labels from other
        if 0 == len(self.levelLabels) :
            self.__init__(other.levelLabels)

        if len(self.levelLabels) != len(other.levelLabels) :
            print "Fatal Error: cannot accumulate mistmatch matrixes"
            sys.exit()
        for sexpected, oexpected in zip(self.levelLabels, other.levelLabels) :
            for scomputed, ocomputed in zip(self.levelLabels, other.levelLabels) :
                self.data[sexpected][scomputed] += other.data[oexpected][ocomputed]

    # Return the accuracy of the confusion matrix.
    # Entries on the diagonal are perfect scores, all others are not.
    def accuracy(self):
        ll = self.levelLabels
        total = sum(self.data[e][c] for e, c in product(ll, ll))
        hits = sum(self.data[e][c] for e, c in product(ll, ll) if e == c)
        return float(hits)/total

    # Return the precision of the confusion matrix.
    # Works only for 2x2 matrices
    def precision(self):
        ll = self.levelLabels
        if (2 != len(ll)) :
            print "Fatal error: precision requires 2 x 2 matrix"
            sys.exit()
        true_pos = self.data[ll[1]][ll[1]]
        false_pos = self.data[ll[0]][ll[1]]
        tp_fp = float (true_pos + false_pos)
        precision = 0
        if 0 < tp_fp :
            precision = float(true_pos) / tp_fp
        return precision

    # Return the recall of the confusion matrix.
    # Works only for 2x2 matrices
    def recall(self):
        ll = self.levelLabels
        if (2 != len(ll)) :
            print "Fatal error: recall requires 2 x 2 matrix"
            sys.exit()
        true_pos = self.data[ll[1]][ll[1]]
        false_neg = self.data[ll[1]][ll[0]]
        tp_fn = float (true_pos + false_neg)
        recall = float(true_pos) / tp_fn
        return recall

    # Return the FPR (false positive rate) of the confusion matrix.
    # Works only for 2x2 matrices
    def fpr(self):
        ll = self.levelLabels
        if (2 != len(ll)) :
            print "Fatal error: FPR requires 2 x 2 matrix"
            sys.exit()
        false_pos = self.data[ll[0]][ll[1]]
        true_neg = self.data[ll[0]][ll[0]]
        fp_tn = float (false_pos + true_neg)
        fpr = float(false_pos) / fp_tn
        return fpr

    # Return the f_one measure
    def f_one(self):
        pre = self.precision()
        rec = self.recall()
        f = 0.0
        if 0.0 < (pre + rec) :
            f = 2.0 * pre * rec / (pre + rec)
        return f

    # Return the f_two measure
    def f_two(self):
        pre = self.precision()
        rec = self.recall()
        f = 0.0
        if 0.0 < (pre + rec) :
            f = 5.0 * pre * rec / (4.0*pre + rec)
        return f

    # Print the CM to stdout
    def display(self):
        ll = self.levelLabels
        print "Classifier results in columns:"
        print "\t".join(str(e) for e in ll)
        print "Expected results in rows."
        for e in ll:
            print "\t".join(str(self.data[e][c]) for c in ll)
        total = sum(self.data[e][c] for e, c in product(ll, ll))
        hits = sum(self.data[e][c] for e, c in product(ll, ll) if e == c)
        print ("Accuracy : " + str(float(hits)/total) +
            " (" + str(hits) + " correct out of " + str(total) +
            " total)")

        # For 2x2 classifiers, also print out the recall and precision
        if (2 == len(ll)) :
            true_pos = self.data[ll[1]][ll[1]]
            false_pos = self.data[ll[0]][ll[1]]
            false_neg = self.data[ll[1]][ll[0]]
            true_neg = self.data[ll[0]][ll[0]]
            tp_fp = float (true_pos + false_pos)
            tp_fn = float (true_pos + false_neg)
            fp_tn = float (false_pos + true_neg)
            precision = 0
            if 0 < tp_fp :
                precision = float(true_pos) / tp_fp
            recall = float(true_pos) / tp_fn
            fpr = float(false_pos) / fp_tn
            fone = 0.0
            ftwo = 0.0
            if 0.0 < precision + recall :
                fone = 2.0 * precision * recall / (precision + recall)
                ftwo = 5.0 * precision * recall / (4.0*precision + recall)
            print ("Precision: " + str(precision) +
                " (" + str(true_pos) + " correct out of " + str(tp_fp) +
                " total)")
            print ("Recall   : " + str(recall) +
                " (" + str(true_pos) + " correct out of " + str(tp_fn) +
                " total)")
            print ("FP Rate  : " + str(fpr) +
                " (" + str(false_pos) + " false pos out of " + str(fp_tn) +
                " total)")
            print "F_1 Score: " + str(fone)
            print "F_2 Score: " + str(ftwo)

        print "----"
        sys.stdout.flush()

    # The 'labels' arg is just a list of labels for each row and column
    # in the matrix.
    def __init__(self, labels = []):
        self.levelLabels = labels
        ll = self.levelLabels
        self.data = {expected:{computed:0 for computed in ll} for expected in ll}

# =========================== END OF FILE =====================
