#
# pm_impl.py -- Main POSES executable
#
# Copyright (c) 2011, 2012 Poulin Holdings LLC
#
import os
import subprocess
import sys

from datetime import datetime, timedelta
from optparse import OptionParser


# poses stuff
from dataset import Dataset
from shared import (check_file, check_files, find_file, get_exe_paths,
                   runMoses, loadCaseFile)
from model import PowerEnsemble, ConfusionMatrix
from levels import  TargetLevels

# Use 4 whitespaces for indentation, no tab.

# This version of POSES only works with moses version 3.0.0 or newer.
# This function returns if the moses version is acceptable, else it
# exits.
def check_moses_version():
    cmd = find_file("moses")
    if (cmd == None) :
        print "Fatal Error: unable to locate the moses executable"
        paths = get_exe_paths("moses")
        for path in paths:
            print "\tSearched for " + path
        sys.exit()
    cmd += " --version"

    # One common failure is not finding shared lib libmoses.so
    verno = "1.0.0"
    try :
        verno = subprocess.check_output(cmd, shell=True)
    except subprocess.CalledProcessError, e:
        # moses version 3.2.1 and earlier return 1 on sucess :-(
        verno = e.output
        if (e.returncode != 1) :
            print >>sys.stderr, "Execution failed:", e
            sys.exit()

    vstr = verno.split(" ")[1]
    maj = vstr.split(".")[0]
    # moses version 2 doesn't have a version string, and so this
    # int conversion will throw an error. So catch it.
    try:
        imaj = int(maj)
    except:
        print "Fatal Error: Can't find moses version; found " + verno
        sys.exit()
    if (3 != imaj) :
        print "Fatal Error: moses version 3.0 or newer required; found " + verno
        sys.exit()
    return

def check_mpi_count(mpiProcs):
    if mpiProcs == 1 or mpiProcs < 0 :
        print "Fatal Error: MPI operation requies at least two processes."
        sys.exit()

def check_loadCaseFile(dataFileName, t = None):
    stru = ""
    try:
        stru = loadCaseFile(dataFileName, t)
    except Exception, err:
        print "Fatal Error: Unable to read the case file", dataFileName
        print err
        print "Is the file missing or corrupt?"
        sys.exit()
    return stru

# Make sure that the number of rows in the dataset are not insane.
def check_dataset_size(dataFileNames, dataset):
    if dataset.numberOfPoints() < 2:
        if 1 == len(dataFileNames) :
            print ("Fatal Error: Datafile " + dataFileNames[0] +
                " must contain at least two rows! ")
        else :
            print ("Fatal Error: The datafiles " + str(dataFileNames) +
                " must contain at least two rows! ")
        sys.exit()

# The target or "parent" variable is the dependent variable for
# regression. Make sure its in the model file.
def check_targetVar_from_model(modelFileName, model, targetVar):
    if model.targetVar != targetVar:
        print ("Fatal Error: The model stored in " + modelFileName +
            " is not targeted to the target variable " + targetVar)
        sys.exit()

def check_targetVar_from_power_ensemble(modelFileName, pe, targetVar):
    for e in pe.ensembles:
        check_targetVar_from_model(modelFileName, e, targetVar)

def check_testCaseFiles(testCaseFileNames):
    if testCaseFileNames==None or len(testCaseFileNames)==0:
        print "Fatal Error: You need at least one test case file, see option -c"
        sys.exit()

def check_neutralBound(lowerNeutral, upperNeutral):
    if lowerNeutral<0.0 or lowerNeutral>1.0:
        print ("Fatal Error: Invalid lower bound for neutral range.  " +
            "The lower bound should lie between 0.0 and 1.0; but " +
            str(lowerNeutral) + " was specified.")
        sys.exit()
    if upperNeutral<0.0:
        print ("Fatal Error: Negative upper bound for neutral range.  " +
            "The upper bound should not be negative; but " +
            str(upperNeutral) + " was specified.")
        sys.exit()
    if lowerNeutral > upperNeutral:
        print ("Fatal Error: Lower bound for neutral range must be " +
            "less than the upper bound, but lower=" +
            str(lowerNeutral) + " and upper=" +
            str(upperNeutral) + " was specified.")
        sys.exit()

# Make sure the Window size is in a valid range.
def check_windowSize(dataset, windowSize):
    if windowSize < 0:
        print "Fatal Error: Negative window size"
        sys.exit()
    if windowSize > dataset.numberOfPoints():
        print "Fatal Error: Window size larger than number of data points"
        sys.exit()

def check_thresholds(dataset, thresholds, numberOfLevels):
    if dataset.isEnumVar(dataset.targetVar) :
        return

    if 1 < numberOfLevels and len(thresholds) != 0 and len(thresholds) != numberOfLevels - 1:
        print "Fatal Error: The number of thresholds specified "\
            "does not match the number of levels.  " \
            "There were " + str(len(thresholds)) + " indicated, " \
            "and " + str(numberOfLevels-1) + " requested."
        sys.exit()

# Run moses on the training dataset, and create a model.
# This includes reading in the training data, partitioning it into
# binary features,  running moses, and writing out the resulting model.
def doTraining(options):
    before = datetime.now()
    check_mpi_count(options.mpiNodes)
    # Read the training file.  If options.delta is non-zero, then
    # compute bin boundaries, for binning the independent features.
    check_files("Training file", options.dataFileNames)
    if options.numInputThrsh > 1 and options.delta == -1:
        options.delta = 1
    dataset = Dataset(options.dataFileNames, options.targetVar,
        options.delta, options.numInputThrsh, options.normalize)
    check_dataset_size(options.dataFileNames, dataset)

    # Windowing is actually lagging some of the columns, for time-series
    # analysis.  So lag as specified.  If there's no "window", the
    # dataset is not touched.
    check_windowSize(dataset, options.windowSize)
    windowedDataset = dataset.windowedVersion(options.windowSize)

    # Split up the range of the target feature into a bunch of
    # intervals.  Each interval is called a "level", and a "threshold"
    # separates two levels.  Thus, there will be numberOfLevels
    # intervals in total, and numberOfLevels-1 thesholds.
    check_neutralBound(options.lowerNeutral, options.upperNeutral)
    check_thresholds(dataset, options.thresholds, options.numberOfLevels)
    levels = TargetLevels(options.numberOfLevels, options.thresholds,
                          options.lowerNeutral, options.upperNeutral,
                          windowedDataset)

    # Do the actual training.
    runMoses(windowedDataset, options, options.modelFileName, levels)
    after = datetime.now()
    print "# Total train time:", after - before

    # Read the model file, from scratch, and report on it.
    pEnsemble = PowerEnsemble(options.modelFileName)
    trainMatrix = pEnsemble.computeMatrix(windowedDataset)

    print "Train matrix for", options.dataFileNames, ":"
    trainMatrix.display()

    return trainMatrix

# Perform self-cross-validation on the training set.
#
# Much like doTraining above, except that the dataset is split up into
# multiple "folds", one of which is set aside for testing, and the
# remainder used for training.  The accuracy (precision, recall, etc)
# is judged by applying the model learned from the training set, applied
# to the test set.
#
# Specifying 5 folds normally requires 5 distinct training sessions,
# unless the -g flag is used to reduce this number.
def doValidation(options):
    before = datetime.now()
    check_mpi_count(options.mpiNodes)
    check_files("Data file", options.dataFileNames)
    if options.numInputThrsh > 1 and options.delta == -1:
        options.delta = 1;
    dataset = Dataset(options.dataFileNames, options.targetVar,
        options.delta, options.numInputThrsh, options.normalize)
    check_windowSize(dataset, options.windowSize)

    check_neutralBound(options.lowerNeutral, options.upperNeutral)
    check_thresholds(dataset, options.thresholds, options.numberOfLevels)
    levels = TargetLevels(options.numberOfLevels, options.thresholds,
                          options.lowerNeutral, options.upperNeutral,
                          dataset.windowedVersion(options.windowSize))

    # -f flag.
    has_test = options.numberOfFolds > 1
    if options.numberOfFolds < 1 :
        options.numberOfFolds = 1

    # -g flag
    if options.foldSessions < 1 :
        options.foldSessions = options.numberOfFolds

    # Initialize a bincount matrix to all zeros.
    trainMatrix = ConfusionMatrix()
    trainPearson = 0
    if has_test:
        testMatrix = ConfusionMatrix()
        testPearson = 0

    after = datetime.now()
    print "# Accuracy: file read elapsed wallclock time: ", after - before

    # Each fold is a pair of datasets, the "test" and the "train" datasets.
    # the "test" dataset contains 1/fold of the total rows, the "train"
    # contains the rest.
    fold_idx = 0
    for fold in dataset.foldelize(options.numberOfFolds, options.windowSize):
        temporaryModelFileName = "tmp_fold_" + str(fold_idx) + ".model"
        # run moses on the train subset
        runMoses(fold.train, options, temporaryModelFileName, levels, fold_idx)
        fold_idx += 1
        pEnsemble = PowerEnsemble(temporaryModelFileName)
        trainMatrix.add(pEnsemble.computeMatrix(fold.train))

        # TODO reenable when scipy dependency requirement is met
        # trainPearson += pEnsemble.computePearson(fold.train)
        if has_test:
            testMatrix.add(pEnsemble.computeMatrix(fold.test))
            # testPearson += pEnsemble.computePearson(fold.test)

        if fold_idx >= options.foldSessions :
            break

    print "Train matrix:"
    trainMatrix.display()
    # print "Pearson:", trainPearson / options.numberOfFolds
    if has_test:
        print "Test matrix:"
        testMatrix.display()
        # print "Pearson:", testPearson / options.numberOfFolds
        return testMatrix

    return trainMatrix

# Given a previously computed model of training dataset, take a single
# line of input data, and infer a value for the target variable from
# that input.
def doInference(options):
    check_file("Model file", options.modelFileName)
    pe = PowerEnsemble(options.modelFileName)
    check_targetVar_from_power_ensemble(
            options.modelFileName, pe, options.targetVar)

    check_testCaseFiles(options.testCaseFileNames)

    if options.windowSize > 0:
        if options.windowSize != len(options.testCaseFileNames):
            print ("Fatal Error: Expecting ", options.windowSize,
                " test case files (as many as the window size)")
            sys.exit()
        fileno = 0
        check_files("Test case file", options.testCaseFileNames)
        for testCaseFileName in options.testCaseFileNames:
            check_loadCaseFile(testCaseFileName,fileno)
            fileno += 1
        print pe.evaluateStru(stru)
    else:
        check_file("Test case file", options.testCaseFileNames[0])
        stru = check_loadCaseFile(options.testCaseFileNames[0])
        print pe.evaluateStru(stru)
        print "Confidence:", pe.confidence


# Given a previously computed model for the training dataset, read
# a "test" file, containing multiple lines of both data and target
# variable values, and compute an accuracy score for the entire file.
# i.e. provide an accuracy score for the model.
#
# XXX unfinished. Problem for enums: the reuters dataset usually has 
# exactly the same bag of words tagged with multiple categories.  Thus,
# there is no one right answer; there are many right answers, and we
# should be able to identify all of them. Unfortunately, the dataset is
# not structured in this way (i.e. to assign one-of-n tags) so this
# whole tagging issue is confusing.  This needs to be reviewed from the
# ground-up.
def doScore(options):
    check_file("Model file", options.modelFileName)
    pe = PowerEnsemble(options.modelFileName)
    check_targetVar_from_power_ensemble(
            options.modelFileName, pe, options.targetVar)

    check_files("Test file", options.dataFileNames)
    dataset = Dataset(options.dataFileNames, options.targetVar,
        -1, 0, options.normalize)

    for row in dataset.strusaver :
        print "Predict:", pe.evaluateStru(row),  \
              "Confidence:", pe.confidence,  \
              "Actual:", row.get(options.targetVar)


# VOI -- Variables of Interest
# Compute mutual information between model, and dataset.
def doVOI(options):

    # Read the model file (a collection of ensembles)
    check_file("Model file", options.modelFileName)
    pe = PowerEnsemble(options.modelFileName)
    check_targetVar_from_power_ensemble(options.modelFileName, pe, options.targetVar)

    check_files("Training file", options.dataFileNames)
    dataset = Dataset(options.dataFileNames, options.targetVar)
    check_dataset_size(options.dataFileNames, dataset)
    check_windowSize(dataset, options.windowSize)

    w2voi = pe.computeVOI(dataset.windowedVersion(options.windowSize))

    outfile = open(options.outputFileName, "w")
    for t in w2voi:
        outfile.write("voi("+t[0]+"="+str(t[1])+")\n")
    outfile.close()


# doAltVOI -- Alternate VOI calculation
# This computes a weighted variant of the counting mutual info, above.
# This does not work very well, in the end, because the feature
# selection algo chose thresholds such that all probabilities would be
# simple fractions (so, for simple 2-clustering, the probs are just
# 0, 0.5 or 1.0).  This means that the MI calculations simplify in
# various ways (mostly sums of log2) and so the result is boring (and
# bad).  This maight be more useful with other feature selection algos??
#
# XXX This is a candidate for removal if any future cleanup is done.
# Be sure to remove the correspnding code in shared.py also.
def doAltVOI(options):

    # Read the model file (a collection of ensembles)
    check_file("Model file", options.modelFileName)
    pe = PowerEnsemble(options.modelFileName)
    check_targetVar_from_power_ensemble(options.modelFileName, pe, options.targetVar)

    check_files("Training file", options.dataFileNames)
    dataset = Dataset(options.dataFileNames, options.targetVar)
    check_windowSize(dataset, options.windowSize)

    w2voi = pe.computeVOI(dataset.windowedVersion(options.windowSize))

    outfile = open(options.outputFileName, "w")
    for t in w2voi:
        outfile.write("voi("+t[0]+"="+str(t[1])+")\n")
    outfile.close()

    # Read the model file (a collection of ensembles)
    cvoi = pe.computeMI(dataset.windowedVersion(options.windowSize))

    outfile = open(options.outputFileName, "w")
    for t in cvoi:
        outfile.write("alt("+t[0]+"="+str(t[1])+")\n")
    outfile.close()


# Read in multiple data files, verify that the headers are consistent,
# and output one big data file (possible changing format from CSV to
# sparse, or v.v.)
def doFilter(options):
    check_files("Input file", options.dataFileNames)
    print "Filtering:", options.dataFileNames
    #if options.numInputThrsh > 1 and options.delta == -1:
    #    options.delta = 1
    dataset = Dataset(options.dataFileNames, options.targetVar,
        -1, 0, options.normalize)
    check_dataset_size(options.dataFileNames, dataset)

    # filter out only certain columns, mask the rest.
    if len(options.showCols) > 0 :
        dataset.printCols(options.showCols)
        print "# Keep columns:", options.showCols

    dataset.saveToFile(options.outputFileName,
                       options.targetVar, False, [], False)

# Run the training command on a batch of files, each individually.
def doBatch(options):
    dataFiles = options.dataFileNames
    modelFile = options.modelFileName
    summaryCM = ConfusionMatrix()

    # Run one training session for each input file
    for file in dataFiles :
        options.dataFileNames = []
        options.dataFileNames.append(file)
        options.modelFileName = file + modelFile
        CM = doTraining(options)
        summaryCM.accum(CM)

    # Restore orig values, in case we are called again.
    options.dataFileNames = dataFiles
    options.modelFileName = modelFile

    print "Summary accuracy results:"
    summaryCM.display()
    return summaryCM

# Run a series of performance measurements, changing an adjustable
# parameter each time through the loop.
def doPerf(options):
    print "Perf: fitness_evals time accuracy recall FPR precision f_one f_two"
    saveMod = options.modelFileName
    for i in range(0,5) :
        options.modelFileName = saveMod + str(i)
        before = datetime.now()
        CM = doValidation(options)
        # CM = doTraining(options)
        # CM = doBatch(options)
        acc = CM.accuracy()
        prec = CM.precision()
        rec = CM.recall()
        fpr = CM.fpr()
        fone = CM.f_one()
        ftwo = CM.f_two()
        after = datetime.now()
        elapsed = (after-before).total_seconds()
        print "Perf:", options.fitnessEvals, elapsed, acc, rec, fpr, prec, fone, ftwo
        options.fitnessEvals *= 2

# Generate a graph of the receiver operating curve.
def doROC(options):
    print "ROC: recall_thresh recall fpr precision accuracy f_one f_two time"
    saveOpts = options.mosesOptions
    saveMod = options.modelFileName
    num_pts = 10
    for i in range(0,num_pts) :
        min_recall = (float(i) + 0.5) / num_pts
        # -Q is the hardness of going below the -q bound
        # -A is the max allowed score, always 1.0 for precision, recall
        options.mosesOptions = saveOpts + " -Hprerec -Q2.0 -A1.0"
        options.mosesOptions += " -q " + str(min_recall)
        options.modelFileName = saveMod + "-q" + str(min_recall)
        before = datetime.now()
        # CM = doTraining(options)
        CM = doValidation(options)
        acc = CM.accuracy()
        prec = CM.precision()
        rec = CM.recall()
        fpr = CM.fpr()
        fone = CM.f_one()
        ftwo = CM.f_two()
        after = datetime.now()
        elapsed = (after-before).total_seconds()
        print "ROC:", min_recall, rec, fpr, prec, acc, fone, ftwo, elapsed


def main():

    usage = ("Usage: %prog command [options]\n"
             "Where command=accuracy|moses|inference|voivar|altvoi|filter")
    # The 'make dist' will substitute for the underbar-REVNO-underbar
    parser = OptionParser(usage, version="poulin-moses-1.r"+"436")

    defaultDataFile = "synthetic.dat"
    parser.add_option("-d", "--data-file",
             dest = "dataFileNames", action="append",
             default = [],
             help = "File containing the data. [default: " + defaultDataFile + "]")

    parser.add_option("-m", "--model-file",
             dest = "modelFileName", default = "test.model",
             help = "File where to write the model. [default: %default]")

    parser.add_option("-p", "--target-variable",
             default = "SP_Change", dest = "targetVar",
             help = "Name of the target variable (variable to " \
                 "predict) in the training file. [default: %default]")

    parser.add_option("-e", "--fitness-evaluations", type="int",
                      default=1000, dest="fitnessEvals",
                      help="The maximum number of fitness evaluations "
                      "that MOSES will perform. [default: %default]")

    parser.add_option("-n", "--number-of-levels", type="int",
                      default=3, dest="numberOfLevels",
                 help="When thresholding the output feature, use this " \
                 "number of levels, or intervals, for the classifier. " \
                 "Note: there is always one more level than there are " \
                 "thresholds.  [default: %default]")

    parser.add_option("-l", "--lower-neutral", type="float",
                      default=0.2, dest="lowerNeutral",
             help="If the number of output levels is 3, this specifies " \
            "the lower threshold as t_md * (1-LOWERNEUTRAL), where " \
            "t_md is the binary classification threshold (the median " \
            "of targetVar, by default). [default: %default]")

    parser.add_option("-u", "--upper-neutral", type="float",
                      default=0.2, dest="upperNeutral",
             help="If the number of output levels is 3, this option " \
             "specifies the upper threshold as t_md*(1+UPPERNEUTRAL), " \
             "where t_md is the binary classification threshold (the " \
             "median of targetVar by default). [default: %default]")

    parser.add_option("-t", "--threshold", action="append",
                      dest="thresholds",
             help="Define an output threshold for the classifier. "\
            "That is, the classifier will return true whenever the "\
            "targetVar >= threshold.   If 'n' levels were specified "\
            "with the -n option, and 'n' is greater than 3, then the "\
            "-t option must be used n-1 times.")

    parser.add_option("-s", "--number-of-thresholds", type = "int",
                      default = 1, dest="numInputThrsh",
                 help="When thresholding input values, create no more " \
                 "than this number of new binary features. " \
                 "[default: %default]")

    parser.add_option("-D", "--delta", type = "float", default = -1.0,
             help = "By default, each continuous-valued input " \
             "feature 'f' converted into a single boolean-valued " \
             "feature 'f_0', taking on values of T or F, indicating " \
             "whether or not 'f' is above or below it's mean value " \
             "'t_0'.  That is, the binary feature 'f_0' indicates "\
             "whether 'f' lies in the interval (mean(f), +inf) or " \
             "not.  By specifying a DELTA > 0  with this option, " \
             "additional new binary features will be created, in " \
             "multiples DELTA of the standard deviation of 'f'.  " \
             "That is, each new feature 'f_n' will be a binary " \
             "indicator of whether 'f' lies above or below (to the " \
             "right or left) of the threshold " \
             "t_n =  mean(f) + DELTA * n * std(f).  "\
             "Here, 'std' is the standard deviation of 'f'. "\
             "The number of new features (the number of values of 'n') " \
             "will be chosen so that the entire range of values for " \
             "that variable is covered.  [default = %default (i.e. disabled)]")

    parser.add_option("-w", "--window-size", type="int",
                      default=0, dest="windowSize",
                      help="If >0 then poses is used in time series mode and defines the lookback window to consider for predicting targetVar. [default: %default]")

    parser.add_option("-f", "--folds", type="int",
            default=5, dest="numberOfFolds",
            help="Specifies how many cross-validation folds will be " \
            "created. This option applies to the 'accuracy' command. " \
            "The dataset is partitioned into N=numberOfFolds parts; "\
            "N-1 of these are used for training, while the last part " \
            "is used for testing (validation).  The number of " \
            "folds must be greater than 1, else cross-validation is " \
            "disabled.  [default: %default]")

    parser.add_option("-g", "--fold-sessions", type="int",
            default=-1, dest="foldSessions",
            help="Specifies how many of the cross-validation folds " \
            "will be used for validation.  When a dataset is partitioned " \
            "into N folds, one may run as many as N training sessions " \
            "for cross-validation.  However, as each training session " \
            "can take a very large amount of time, it can be useful " \
            "to train for less than all n folds.  This option " \
            "specifies how many training sessions should be run. " \
            "Valid values range from 1 to the number of folds. " \
            "The default is to run a training session for each fold.")
          
    parser.add_option("-c", "--test-case-file", action="append",
            dest="testCaseFileNames",
            help="Test case file (or files in the case of time series). " \
            "Required by the \"inference\" command.")

    parser.add_option("-o", "--output-file", default="test.voi",
            dest="outputFileName",
            help="File where to dump the results, used by the commands \
                  voivar and filter. [default: %default]")

    parser.add_option("-b", "--balance", action="store_true", default=False,
                      dest="balance",
                      help="If enabled, then the dataset is balanced \
                           (so that there are the same number of \
                           positive and negative samples) before passed \
                           to MOSES.")

    parser.add_option("-W", "--weighted", action="store_true", default=False,
                      dest="weighted",
                      help="If enabled, then the models are weighted \
                            according to their predictive strength.  \
                            Otherwise, they are uniformly weighted.")

    parser.add_option("-M", "--moses-opt",
                      default="", dest="mosesOptions",
              help="MOSES has a large number of options that control " \
              "its behavior and performance.  See the moses man page " \
              "for details.  To pass multiple moses options, they must " \
              "be surrounded by double quotes.  So, for example, " \
              "-M \"-v12 -Z1\" [default: \"\"]")

    parser.add_option("-F", "--feature-selection-opt",
                      default="", dest="FSOptions",
                      help="Specify arguments to the 'feature-selection' \
                            program (a part of the MOSES suite). The \
                            feature-selection step is used to \
                            pre-process the data.  To pass multiple \
                            options, place them between double quotes, \
                            such as \"-c10 -m10000\".  The complete set \
                            of options is documented on the \
                            feature-selection man page. \
                            [default: \"\"]")

    parser.add_option("-V", "--verbose", action="store_true", default=False,
                      dest="verbose",
                      help="If enabled, write the feature selection and \
                           moses log files to the current working \
                           directory (cwd).")

    parser.add_option("--mpi", type="int",
             dest = "mpiNodes", default = "0",
             help = "Number of MPI nodes to distribute processing to. \
                     [default: MPI disabled]")

    parser.add_option("--filt", action="append",
                      dest="showCols", default = [],
             help="Specify one or more columns that will not be " \
                  "discarded by the filter command. Any columns not "\
                  "specified will be discarded.  This option may be "\
                  "used multiple times.  Alternately, column names "\
                  "may be white-space separated. By default, all "\
                  "columns are kept.")

    parser.add_option("--normalize", action="store_true", default=False,
                      dest="normalize",
             help="If enabled, then the data in each row will be " \
                  "rescaled so that the sum-total of the columns " \
                  "equals one.  This is done before thresholding.")

    parser.add_option("--restarts", type="int",
             dest = "restarts", default = "1",
             help = "Number of MOSES runs to perform during model " \
                    "build.  Each run restarts MOSES from the " \
                    "begining, using a different random number seed. " \
                    "[default: 1]")

    (options, args) = parser.parse_args()

    if 0 == len(options.dataFileNames) :
        options.dataFileNames.append(defaultDataFile)
     
    if options.thresholds == None:
        options.thresholds=[]

    if 1 > options.numberOfLevels :
        options.numberOfLevels = 1

    if len(args) < 1:
        parser.error("incorrect number of arguments. Use --help to get more information")

    # Assume that all arguments after the initial command are data file
    # names. This allows, for example, "poses moses -d *.dat" to specify
    # multiple files by wild-card.
    if 1 < len(args):
        for i in range(1,len(args)) :
            options.dataFileNames.append(args[i])

    # Allow user to say --filt "A B C" to specify multiple columns.
    if 0 < len(options.showCols) :
       lst = []
       for i in range(0, len(options.showCols)) :
           s = options.showCols[i].split()
           for j in range(0, len(s)) :
              lst.append(s[j])
       options.showCols = lst

    check_moses_version()

    argstr = ""
    for arg in sys.argv :
        argstr += arg + " "
    print "# Command line: ", argstr
    print "# Version: poulin-moses-1.r"+"436"

    command=args[0]
    if command=="accuracy":
        doValidation(options)
        sys.exit()

    if command=="moses":
        doTraining(options)
        sys.exit()

    if command=="inference":
        doInference(options)
        sys.exit()

    if command=="voivar":
        doVOI(options)
        sys.exit()

    # Experimental commands, not for general consumption (yet).
    if command=="altvoi":
        doAltVOI(options)
        sys.exit()

    if command=="batch":
        doBatch(options)
        sys.exit()

    if command=="filter":
        doFilter(options)
        sys.exit()

    if command=="perf":
        doPerf(options)
        sys.exit()

    if command=="roc":
        doROC(options)
        sys.exit()

    if command=="score":
        doScore(options)
        sys.exit()

    parser.error("unrecognized command")

