#
# shared.py -- Poulin-MOSES core library.
#
# This file defines all the core routines for mangling data.
# The main routine (main entry point) is elsewhere.
#
# Copyright (c) 2011, 2012 Poulin Holdings LLC
#
import os
import sys

from datetime import datetime
from itertools import product
from tempfile import mkstemp

# POSES imports
from model import writeEnsemble


# Verify that a file exists; this cleans up otherwise cryptic error
# messages that the user would have to read.
def check_file(errmsg, fileName):
    if not os.path.exists(fileName):
        print "Error: " + errmsg + " \"" + fileName + "\" not found"
        sys.exit()

# As above, takes multiple files.
def check_files(errmsg, fileNames):
    not_found = False
    for fileName in fileNames :
        if not os.path.exists(fileName):
            print "Error: " + errmsg + " \"" + fileName + "\" not found"
            not_found = True
    if not_found :
        sys.exit()


# Create a list of possible locations of binary executables.
def get_exe_paths(fileName):
    paths = []
    paths.append(os.path.join("/usr/local/bin", fileName))
    paths.append(os.path.join("/usr/bin", fileName))
    paths.append(os.path.join(sys.path[0], fileName))
    paths.append(os.path.join(sys.path[0], "bin", fileName))
    # Reset sys.path[0] to work around weird cxfreeze bug.
    altbase = os.path.dirname(sys.path[0])
    paths.append(os.path.join(altbase, fileName))
    paths.append(os.path.join(altbase, "bin", fileName))
    return paths


# Find the fully-qualified path to a file.  Right now, this is
# implemented so it only works for the binaries, searching in the
# current and the bin directory... This could be fancied up.
def find_file(fileName):
    paths = get_exe_paths(fileName)
    for path in paths:
        if os.path.exists(path):
            return path
    return None


def levelsFileName(prefixFileName):
    return prefixFileName+"_levels"


# Read file containing key : value pairs.  Return dict containing same.
#
# File is assumed to  contain one key : value pair per line.  Values are
# assumed to be floats.  Comments are not allowed.
def loadCaseFile(fileName, t = None):
    w2c = {}
    infile = open(fileName)
    suffix = ""
    if t != None:
        suffix += "_w" + str(t)
    for line in infile:
        sline = line.strip()
        if (0 < len(sline)
            and not line.startswith("#")
            and not line.startswith("!")
            and not line.startswith(";")) :
                cols = sline.split(":")
                w2c[cols[0].strip() + suffix] = float(cols[1].strip())
    infile.close()
    return w2c

# Run the feature selection command on the indicated input files,
# generating the output files, using the indicated options.
#
def featureSelectionCmd(DataMosesFileName, FSOptions, outputFileName,
                        targetVar, verbose):
    cmd = find_file("feature-selection")
    cmd += " -i" + DataMosesFileName
    cmd += " -o" + outputFileName
    if verbose :
        cmd += " -F feature_selection_" + DataMosesFileName + ".log"
    else :
        cmd += " -l NONE"
    cmd += " -u" + targetVar
    cmd += " " + FSOptions
    return cmd

def overrideOpt(mosesOptions, opt) :
    s = " " + opt
    return (0 == mosesOptions.find(opt)) or (-1 < mosesOptions.find(s))

def mosesCmd(fitnessEvals, DataMosesFileName, mosesOptions, tmppath,
             targetVar, verbose, mpiNodes):
    cmd = ""
    if (0 < mpiNodes) :
        mpiCmd = ""
        try:
            mpiCmd = os.environ['MPIRUN']
        except KeyError:
            mpiCmd = "mpirun"
        mpiHosts = ""
        try:
            mpiHosts = os.environ['MPI_HOSTS']
        except KeyError:
            mpiHosts = "~/mpd.hosts"
        cmd += mpiCmd + " -n " + str(mpiNodes)
        cmd += " --hostfile " + mpiHosts + " "
        mosesOptions += " --mpi=1"

    cmd += find_file("moses")

    # Options that cannot be overwritten.
    # See moses man page for their explanations.
    cmd += " -S1 -x0 -t0"

    if verbose :
        cmd += " -f moses_" + DataMosesFileName + ".log"
    else :
        cmd += " -l NONE"
    cmd += " -u" + targetVar
    cmd += " -i" + DataMosesFileName
    cmd += " -m" + str(fitnessEvals)
    cmd += " -o" + tmppath
    cmd += " -W1"         # output result with labels instead of place holders
    cmd += " -n sin -n log -n exp -n div"   # Standard contin options
    # cmd += " -s1 -d1"     # Enable caching

    # User may over-ride the -H option, e.g. by saying -Hrecall
    if not overrideOpt(mosesOptions, "-H") :
        cmd += " -Hit"

    # backwards-compat
    if not overrideOpt(mosesOptions, "-Z") :
        cmd += " -Z1"         # Enable cross-over

    cmd += " " + mosesOptions
    return cmd


# Determine the weight (i.e. confidence) of a model given its score.
# Keep in mind, the score is negative, and the closer to zero it gets,
# the better the score is.  i.e. the Score is the number of incorrect
# answers given on the training set.
#
# This assigns a confidence of 0.0 to anything with a score that gets
# less than half the answer correct, and a confidence of 1.0 to anything
# with a perfect score.
def score2confidence(score, dataset_size):
    return max(0.5 + score/float(dataset_size), 0) * 2.0


# Takes a model line, that is the score followed by the combo program
# and return the weight followed by the combo program.
def translateModel(modelLine, dataset_size, weighted):
    modelPart = modelLine.partition(" ")
    weight = score2confidence(float(modelPart[0]), dataset_size) if weighted else 1
    return str(weight) + modelPart[1] + modelPart[2]


# Run MOSES just once.  Return a list of models so obtained.
def runMosesOnce(options, mosesDataFileName, dataset_size) :

    # tmpfile to which the results will be written.
    # tmppath = os.path.join(os.getcwd(), "tmp.txt")
    # tmppath = mkstemp(suffix='.txt', dir=os.getcwd())[1]
    tmppath = mkstemp(suffix='.txt')[1]

    # Invokke the MOSES binary.
    MOSESCmd = mosesCmd(options.fitnessEvals,
                        mosesDataFileName,
                        options.mosesOptions, tmppath,
                        options.targetVar, options.verbose,
                        options.mpiNodes)
    print "#", MOSESCmd
    sys.stdout.flush()

    # measure wallcock time.
    before = datetime.now()
    os.system(MOSESCmd)
    after = datetime.now()
    print "# Training: elapsed wallclock time: ", after - before

    # If moses crashed, then there's no tmp file ...
    check_file("Moses failed to create output file; did moses crash? :", tmppath)
    infile = open(tmppath)
    models = [translateModel(m, dataset_size, options.weighted)
              for m in infile.read().split("\n") if m]
    infile.close()
    os.remove(tmppath)
    return models


# Run feature-selection+ MOSES just once.
# Return a list of models so obtained.
def runFSMOnce(options, mosesDataFileName, dataset_size) :

    # Perform feature-selection
    enableFS = options.FSOptions
    if enableFS:
        FSCmd = featureSelectionCmd(mosesDataFileName, options.FSOptions,
                                    mosesDataFileName + "_filtered",
                                    options.targetVar, options.verbose)
        print "#", FSCmd
        sys.stdout.flush()
        before = datetime.now()
        os.system(FSCmd)
        after = datetime.now()
        print "# Feature selection: elapsed wallclock time: ", after - before
    else:
        print "# Skipped feature selection"
        sys.stdout.flush()

    # If only one rand seed, then just do it.
    if 1 == options.restarts :
        models = runMosesOnce(options,
                            mosesDataFileName
                            + ("_filtered" if enableFS else ""),
                            dataset_size)
        return models

    # Re-run moses multiple times, each time starting with a
    # different random number seed.
    models = []
    saveOpts = options.mosesOptions
    for seed in range(0, options.restarts) :
        options.mosesOptions = saveOpts + " -r " + str(seed)
        mods = runMosesOnce(options,
                            mosesDataFileName
                            + ("_filtered" if enableFS else ""),
                            dataset_size)
        models.extend(mods)
    options.mosesOptions = saveOpts

    return models

# Run feature-selection and moses, and write the resulting ensemble
# to the model file.
def runWriteEnsemble(dataset, mosesDataFileName, options, 
                     modelFile, have_output_thresh, label) :

    # The saveToFile() method applies a boolean filter to the data,
    # replacing data values by boolean 0,1 values indicating if the
    # variable is below/above the threshold value(s).
    # (But it only does this if have_output_thresh is set).
    dataset_size = dataset.saveToFile(mosesDataFileName,
                                      options.targetVar,
                                      have_output_thresh, label,
                                      options.balance)
    models = runFSMOnce(options, mosesDataFileName, dataset_size)

    writeEnsemble(models, modelFile, options.targetVar,
                  have_output_thresh, label)



# Run MOSES for all thresholds and produce as many models as
# thresholds, denoted in modelFileNames.
def runMoses(dataset, options, modelFileName, output_levels, fold_idx = -1):

    # Generate unique data file names mostly to aid with deubgging,
    # other than that, there is no real reason to pick distinct file
    # names. (well, if we run multiple threads, we'd need this...)
    fold_suffix = ""
    if fold_idx >= 0:
        fold_suffix = "_fold_" + str(fold_idx)

    have_output_thresh = (1 != output_levels.numberOfLevels)

    modelFile = open(modelFileName,"w")

    if not have_output_thresh :
        mosesDataFileName = "csv" + fold_suffix
        runWriteEnsemble(dataset, mosesDataFileName, options,
                         modelFile, False, 0)

    else :
        if output_levels.targetIsEnum :
            # For each target-enum value
            for label in output_levels.labels :
                fileName = "csv" + fold_suffix + "_e_" + label
                runWriteEnsemble(dataset, fileName, options,
                                 modelFile, True, label)
        else :
            # For each target-variable threshold...
            for threshold in output_levels.thresholds :
                fileName = "csv" + fold_suffix + "_t" + str(threshold)
                runWriteEnsemble(dataset, fileName, options,
                                 modelFile, True, threshold)

    output_levels.write(modelFile)
    modelFile.close()


# =========================== END OF FILE =====================
