/*
 * moses3/learning/moses/eda/replacement.h
 *
 * Copyright (C) 2002-2008 Novamente LLC
 * All Rights Reserved
 *
 * Written by Moshe Looks
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef _EDA_REPLACE_H
#define _EDA_REPLACE_H

#include <moses3/util/exceptions.h>
#include <moses3/util/lazy_random_selector.h>
#include <moses3/util/oc_omp.h>

#include "using.h"
#include "../representation/field_set.h"
#include "scoring.h"

namespace moses3 { 
namespace moses {

// Replace the lowest-scored individuals in the current population,
// specified by the iteratators Dst, and replace them unconditionally
// with the individuals given by NewInst.  The container holding the
// Dst iterators will be partially sorted.
struct replace_the_worst
{
    template<typename NewInst, typename Dst>
    void operator()(NewInst from, NewInst to, Dst from_dst, Dst to_dst) const
    {
        OC_ASSERT(distance(from, to) <= distance(from_dst, to_dst),
                  "ERROR: The number of individuals to replace in a population\n"
                  "is larger than the population itself.");

        // OMP_ALGO here means that this is done in parallel.
        // nth_element is a partial sort, putting the weakest individuals
        // at the head. The new individuals then replace these by simply
        // getting copied over.
        OMP_ALGO::nth_element(from_dst, from_dst + distance(from, to), to_dst);
        copy(from, to, from_dst);
    }
};

// Replace the most similar individual, where similarity is determined by
// the hamming distance.
//
// TODO: I think it might be a little more efficent to use the
// hamming_distance as a sort comparison operator, and hand off the whole
// thing to std:nth_element, and let that class figure out who is close or
// not.  This avoids the use of doubly-nested loops, and multiple redundant
// evaluations of the hamming distance, right?
//
struct rtr_replacement
{
    rtr_replacement(const field_set& fs, int ws, RandGen& _rng = randGen())
        : window_size(ws), _fields(&fs), rng(_rng) { }

    template<typename NewInst, typename Dst>
    void operator()(NewInst from, NewInst to, Dst from_dst, Dst to_dst) const
    {
        OC_ASSERT(window_size <= distance(from_dst, to_dst),
                  "windows size greater than distance from_dst -> to_dst.");

        for (;from != to; ++from)
            operator()(*from, from_dst, to_dst);
    }

    template<typename Dst, typename ScoreT>
    void operator()(const scored_instance<ScoreT>& inst,
                    Dst from_dst, Dst to_dst) const
    {
        lazy_random_selector select(distance(from_dst, to_dst), rng);
        Dst closest = from_dst + select();
        int closest_distance = _fields->hamming_distance(inst, *closest);
        dorepeat(window_size - 1) {
            Dst candidate = from_dst + select();
            int distance = _fields->hamming_distance(inst, *candidate);
            if (distance < closest_distance) {
                closest_distance = distance;
                closest = candidate;
            }
        }
        if (*closest < inst)
            *closest = inst;
    }

    int window_size;
protected:
    const field_set* _fields;
    RandGen& rng;
};

} // ~namespace moses
} // ~namespace moses3

#endif
