/*
 * moses3/util/Config.h
 *
 * Copyright (C) 2008 by OpenCog Foundation
 * All Rights Reserved
 *
 * Written by Gustavo Gama <gama@vettalabs.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _OPENCOG_CONFIG_H
#define _OPENCOG_CONFIG_H

#include <string>
#include <map>

namespace moses3
{

class Config;

typedef Config* ConfigFactory(void);


class Config
{

protected:

    const std::string* DEFAULT()
    {
        static const std::string defaultConfig[] = {
            "SERVER_PORT",           "17001",
            "LOG_FILE",              "moses3_server.log",
            "LOG_LEVEL",             "info",
            "ANSI_ENABLED",          "false",
            "BACK_TRACE_LOG_LEVEL",  "warn",
            "LOG_TO_STDOUT",         "true",
            "SERVER_CYCLE_DURATION", "100",     // in milliseconds
            "EXTERNAL_TICK_MODE",    "false",
            "STARTING_STI_FUNDS",    "10000",
            "STARTING_LTI_FUNDS",    "10000",
            "STI_FUNDS_BUFFER",      "10000",
            "LTI_FUNDS_BUFFER",      "10000",
            "MIN_STI",               "-400",
            "PROMPT",                "moses3> ",
            "ANSI_PROMPT",           "moses3> ",
            "MODULES",               "libbuiltinreqs.so",
            "SCM_PRELOAD",           ""
            "",                      ""
        };
        return defaultConfig;
    }

    std::string emptyString;
    std::map<std::string, std::string> table;
    std::string _path_where_found;

public:
    // constructor and destructor
    ~Config();
    Config();

    // Returns a new Config instance
    static Config* createInstance(void);

    // reset configuration to default
    virtual void reset();

    /// Load passed file and redefines values for parameters.
    void load(const char* config_file, bool resetFirst = true);

    const std::string& path_where_found() { return _path_where_found; }

    // Checks whether a parameter exists
    const bool has(const std::string &parameter_name) const;

    // Set the value of a given parameter.
    void set(const std::string &parameter_name, const std::string &parameter_value);

    // Return current value of a given parameter.
    const std::string& get(const std::string &parameter_name) const;
    const std::string& operator[](const std::string &name) const;

    // Return current value of a given parameter as an integer
    int get_int(const std::string &parameter_name) const;

    // Return current value of a given parameter as an long
    long get_long(const std::string &parameter_name) const;

    // Return current value of a given parameter as a double
    double get_double(const std::string &parameter_name) const;
    //
    // Return current value of a given parameter as a boolean
    bool get_bool(const std::string &parameter_name) const;

    // Dump all configuration parameters to a string
    std::string to_string() const;
};

// singleton instance (following meyer's design pattern)
// Nil: if overwrite is true then the static variable instance
//      is changed with the createInstance provided
//      it is a temporary dirty hack 
Config& config(ConfigFactory* = Config::createInstance,
               bool overwrite = false);

} // namespace moses3

#endif // _OPENCOG_CONFIG_H
