/** MannWhitneyU.h --- 
 *
 * Copyright (C) 2012 Nil Geisweiller
 *
 * Author: Nil Geisweiller <nilg@desktop>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _OPENCOG_MANNWHITNEYU_H
#define _OPENCOG_MANNWHITNEYU_H

#include <boost/range/numeric.hpp>
#include <boost/range/adaptor/map.hpp>

#include "Counter.h"
#include "ranking.h"

namespace moses3 {

using boost::adaptors::map_values;

/**
 * return the Mann-Whitney U (U_2 to be precise) given 2 distributions
 * described by Counters. If n2, the size of c2, is not provided it is
 * automatically calculated.
 */
template<typename Key, typename FloatT>
FloatT MannWhitneyU(const Counter<Key, FloatT>& c1,
                    const Counter<Key, FloatT>& c2,
                    FloatT n2 = -1) {
    typedef Counter<Key, FloatT> counter_t;
    counter_t c = c1 + c2;
    counter_t r = ranking(c);
    FloatT sum2 = 0;
    foreach(const auto& v, c2)
        sum2 += r[v.first];
    if (n2 < 0)
        n2 = boost::accumulate(c2 | map_values, 0);
    return sum2 - n2*(n2+1)/2;
}

/**
 * return the standardized Mann-Whitney U given 2 distributions
 * described by Counters
 */
template<typename Key, typename FloatT>
FloatT standardizedMannWhitneyU(const Counter<Key, FloatT>& c1,
                                const Counter<Key, FloatT>& c2,
                                FloatT n1 = -1, FloatT n2 = -1) {
    if(n1 < 0)
        n1 = boost::accumulate(c1 | map_values, 0);
    if(n2 < 0)
        n2 = boost::accumulate(c2 | map_values, 0);
    FloatT U = MannWhitneyU(c1, c2, n2),
        mU = n1*n2/2,
        sU = sqrt(mU*(n1+n2+1)/6);
    return (U - mU) / sU;
}

} // ~namespace moses3

#endif // _OPENCOG_MANNWHITNEYU_H
